#include "zeusmp.def"
c=======================================================================
c
c    \\\\\\\\\\      B E G I N   S U B R O U T I N E      //////////
c    //////////                 A V I S C                 \\\\\\\\\\
c
c                            Developed by
c                Laboratory of Computational Astrophysics
c               University of Illinois at Urbana-Champaign
c
c=======================================================================
c
       subroutine avisc (ibeg,iend,jbeg,jend,kbeg,kend,dvdxmn
     &                  ,w1,w2,w3,u1,u2,u3,s1,s2,s3)
c
c    mln:zeus3d.viscous <------------- artificial viscosity source terms
c                                                        ?????????, 19??
c
c    written by: Mike Norman
c    modified 1: June, 1988 by Jim Stone; incorporated into ZEUS2D
c    modified 2: February, 1990 by David Clarke; incorporated into
c                ZEUS3D
c    modified 3: June, 1992 by David Clarke; expunged "ISMIN", thereby
c                decreasing execution time by 30%.
c    modified 4: Oct., 1994 by Robert Fiedler to run in parallel on SGIs
c    modified 5: Totally rewritten 2/27/96 by RAF for ZEUS-MP.
c
c  PURPOSE: Computes the artificial viscosity source terms in the
c  momentum and energy equations.  i.e., it computes
c
c             dv / dt = -DIV(Q) / rho             for v1, v2, and v3
c      and    de / dt = -Q * delta(v) / delta(x)
c
c  This routine uses the von Neumann-Richtmyer form of the artificial
c  viscosity.  This means that geometric terms are not included in
c  DIV(Q).
c
c
c  LOCAL VARIABLES:
c    qqs        diagonal elements of viscous tensor.  Thus, this is a
c               linear treatment of the artificial viscosity.
c    dvelb      v1(i+1,j,k)-v1(i,j,k) for i-sweep
c               v2(i,j+1,k)-v2(i,j,k) for j-sweep
c               v3(i,j,k+1)-v3(i,j,k) for k-sweep
c    dvela      min ( zro, dvelb ) - ensures that only compressional
c               waves (shocks) are affected.
c    dvdxmn     min ( ( delta(v1) / delta(x1) ),
c                     ( delta(v2) / delta(x2) ),
c                     ( delta(v3) / delta(x3) ) )
c    w1,w2,w3   velocity values prior to viscosity update.
c    u1,u2,u3   velocity values after    viscosity update.
c    s1,s2,s3   updated momentum densities for transport step.
c
c BOUNDARY VALUES USED:
c
c  Macro defined  var   ii    oi    ij    oj    ik    ok
c  -------------  ---  ----  ----  ----  ----  ----  ----
c                  d   is-1        js-1        ks-1
c                  u1  is-1  ie+1
c                  u2              js-1  je+1
c                  u3                          ks-1  ke+1
c
c  EXTERNALS:
c
c-----------------------------------------------------------------------
c
       implicit NONE
#include "param.h"
#include "root.h"
#include "field.h"
#include "grid.h"
#include "scratch.h"
c
       integer i,j,k, ibeg,iend,jbeg,jend,kbeg,kend
       REAL    q1,q3, g3i
       REAL    qqs,qqsm, dvelas,dvelasm, dvdxmn
       REAL    w1(in,jn,kn),w2(in,jn,kn),w3(in,jn,kn)
     &        ,u1(in,jn,kn),u2(in,jn,kn),u3(in,jn,kn)
     &        ,s1(in,jn,kn),s2(in,jn,kn),s3(in,jn,kn)
#ifndef UNROLL_I
#define UNROLL_I
#endif
c
c-----------------------------------------------------------------------
c
c      Start von Neumann-Richtmyer artificial viscosity update.
c
#ifdef SUB_AV
       q1     = avisc_dt * qcon
#else
       q1     = dt * qcon
#endif /* SUB_AV */
c
c Do all the work in one big loop so that d and e will stay in cache.
c
         do 160 k=kbeg,kend
           do 150 j=jbeg,jend
CDIR$ UNROLL (UNROLL_I)
             do 140 i=ibeg,iend
c
c      Do v1.
c
                 dvelasm = min ( zro, w1(i  ,j,k) - w1  (i-1,j,k) )
                 dvelas  = min ( zro, w1(i+1,j,k) - w1  (i  ,j,k) )
                 qqsm    = q1 * d(i-1,j,k) * dvelasm * dvelasm
                 qqs     = q1 * d(i  ,j,k) * dvelas  * dvelas
                 q3      = dvelas  * dx1ai(i)
                 dvdxmn  = min ( dvdxmn, q3 )
#ifndef ISO
                 e  (i,j,k) = e(i,j,k) - q3 * qqs
#endif /* not ISO */
                 u1 (i,j,k) = w1(i,j,k)
     1                      - ( qqs  - qqsm   ) * dx1bi(i)
     2                      * 2.0 / ( d(i-1,j,k) + d(i,j,k) )
c
c      Do v2.
c
                 dvelasm = min ( zro, w2(i,j  ,k) - w2  (i,j-1,k) )
                 dvelas  = min ( zro, w2(i,j+1,k) - w2  (i,j  ,k) )
                 qqsm    = q1 * d(i,j-1,k) * dvelasm  * dvelasm
                 qqs     = q1 * d(i,j  ,k) * dvelas   * dvelas
                 q3      = dvelas   * dx2ai(j) * g2bi(i)
                 dvdxmn  = min ( dvdxmn, q3 )
#ifndef ISO
                 e  (i,j,k) = e(i,j,k) - q3 * qqs
#endif /* not ISO */
                 u2 (i,j,k) = w2(i,j,k)
     1                      - ( qqs   - qqsm    ) * dx2bi(j) * g2bi(i)
     2                      * 2.0 / ( d(i,j-1,k) + d(i,j,k) )
c
c      Do v3.
c
                 g3i     = g31bi(i) * g32bi(j)
                 dvelasm = min ( zro, w3(i,j,k  ) - w3  (i,j,k-1) )
                 dvelas  = min ( zro, w3(i,j,k+1) - w3  (i,j,k  ) )
                 qqsm    = q1 * d(i,j,k-1) * dvelasm  * dvelasm
                 qqs     = q1 * d(i,j,k  ) * dvelas   * dvelas
                 q3      = dvelas   * dx3ai(k) * g3i
                 dvdxmn  = min ( dvdxmn, q3 )
#ifndef ISO
                 e  (i,j,k) = e(i,j,k) - q3 * qqs
#endif /* not ISO */
                 u3 (i,j,k) = w3(i,j,k)
     1                      - ( qqs   - qqsm    ) * dx3bi(k) * g3i
     2                      *  2.0 / ( d(i,j,k-1) + d(i,j,k) )
c
c Save the updated momentum densities for the transport step, since 
c everything should be in the cache already.
c
               s1(i,j,k) = u1(i,j,k) * 0.5 * (d(i-1,j  ,k  ) + d(i,j,k))
               s2(i,j,k) = u2(i,j,k) * 0.5 * (d(i  ,j-1,k  ) + d(i,j,k))
     1                   * g2b(i)
               s3(i,j,k) = u3(i,j,k) * 0.5 * (d(i  ,j  ,k-1) + d(i,j,k))
     1                   * g31b(i) * g32b(j)
140          continue
150        continue
160      continue
c
       return
       end
c
c=======================================================================
c
c    \\\\\\\\\\        E N D   S U B R O U T I N E        //////////
c    //////////                 A V I S C                 \\\\\\\\\\
c
c=======================================================================
c
c
