#include "zeusmp.def"
c
c=======================================================================
c
c    \\\\\\\\\\      B E G I N   S U B R O U T I N E      //////////
c    //////////               A V I S C _ D
c
c                            Developed by
c                Laboratory of Computational Astrophysics
c               University of Illinois at Urbana-Champaign
c
c=======================================================================
c
       subroutine avisc_d
     1            (v1old,v2old,v3old,v1new,v2new,v3new,s1,s2,s3)
c
c  PURPOSE
c    Driver for artificial viscosity module.  Updates velocity
c    components and material energy.
c
c  AUTHOR
c    Robert A. Fiedler
c
c  LAST MODIFIED
c    01/21/97
c
c  INPUT
c    v*old     Velocity components before viscous update.
c
c  OUPUT
c    v*new     Velocity components after  viscous update.
c    s*        Momentum components for use in the transport step.
c
c  EXTERNALS:
c    BVALV1  , BVALV2  , BVALV3
c    BVALD
c    AVISC
c
c.......................................................................
c
       implicit NONE
#include "mod_names.h"
#include "param.h"
#include "root.h"
#include "field.h"
#include "bndry.h"
#include "grid.h"
#include "mpicom.h"
#ifdef PROFILING
#ifdef MPE_USED
#else
#include "clocks.h"
#endif /* MPE_USED */
#endif /* PROFILING */
c
       REAL v1old(in,jn,kn),v2old(in,jn,kn),v3old(in,jn,kn)
     1     ,v1new(in,jn,kn),v2new(in,jn,kn),v3new(in,jn,kn)
     2     ,s1   (in,jn,kn),s2   (in,jn,kn),s3   (in,jn,kn)
c
       integer i, k1, k2, imax, jmax, kmax
       REAL    dvdxmn
c
c      External statements
c
       external      avisc
     1             , bvalv1  , bvalv2  , bvalv3  , bvald
c
c-----------------------------------------------------------------------
c
c Divide the computational volume into three equal pieces.  We must
c have at least 5 active zones in the 3-direction.
c
       k1 = int( real( ke - ks + 1 ) / 3.0 ) + ks
       k2 = int( real( ke - ks + 1 ) / 3.0 ) + k1
c
#ifdef DEBUG
       write(*,"(/'AVISC_D: ',i2,' calling avisc.')") myid
#endif /* DEBUG */
c
c Save the largest velocity gradient for computing the viscous
c time step.
c
       dvdxmn = 0.0
c
c i boundaries
c
c    1) Post sends and receives.
c
       nreq = 0
       nsub = nsub + 1
       call bvald  (1,0,0,0,0,0,d    )
       call bvalv1 (1,1,0,0,0,0,v1old)

#ifdef PROFILING
#ifdef MPE_USED
       call MPE ...
#else
       call clocks (cputime, wclock)
       wal0 = wal1
       cpu0 = cpu1
       wal1 = real(wclock)
       cpu1 = real(cputime)
       timers(nsubstps,1,2,2) = timers(nsubstps,1,2,2) + wal1  - wal0
       timers(nsubstps,1,2,1) = timers(nsubstps,1,2,1) + cpu1  - cpu0
#endif /* MPE_USED */
#endif /* PROFILING */

c
c    2) Do first portion of the interior points.
c
       call avisc (is+1,ie-1,js+1,je-1,ks+1,k1  ,dvdxmn
     &            ,v1old,v2old,v3old,v1new,v2new,v3new,s1,s2,s3)
c
c       subroutine avisc (ibeg,iend,jbeg,jend,kbeg,kend,dvdxmn
c     &                  ,w1,w2,w3,u1,u2,u3,s1,s2,s3)
c
cc    w1,w2,w3   velocity values prior to viscosity update.
cc    u1,u2,u3   velocity values after    viscosity update.
cc    s1,s2,s3   updated momentum densities for transport step.
c

#ifdef PROFILING
#ifdef MPE_USED
       call MPE ...
#else
       call clocks (cputime, wclock)
       wal0 = wal1
       cpu0 = cpu1
       wal1 = real(wclock)
       cpu1 = real(cputime)
       timers(2,1,2,2) = timers(2,1,2,2) + wal1  - wal0
       timers(2,1,2,1) = timers(2,1,2,1) + cpu1  - cpu0
#endif /* MPE_USED */
#endif /* PROFILING */

#ifdef MPI_USED
c
c    3) Wait for communications to complete.
c
       if (nreq .ne. 0) call MPI_WAITALL ( nreq, req, stat, ierr )

#ifdef PROFILING
#ifdef MPE_USED
       call MPE ...
#else
       call clocks (cputime, wclock)
       wal0 = wal1
       cpu0 = cpu1
       wal1 = real(wclock)
       cpu1 = real(cputime)
       timers(2,1,1,2) = timers(2,1,1,2) + wal1  - wal0
       timers(2,1,1,1) = timers(2,1,1,1) + cpu1  - cpu0
#endif /* MPE_USED */
#endif /* PROFILING */

#endif
c......................................................................
c
c j boundaries
c
c    1) Post sends and receives.
c
       nreq = 0
       nsub = nsub + 1
       call bvald  (0,0,1,0,0,0,d    )
       call bvalv2 (0,0,1,1,0,0,v2old)

#ifdef PROFILING
#ifdef MPE_USED
       call MPE ...
#else
       call clocks (cputime, wclock)
       wal0 = wal1
       cpu0 = cpu1
       wal1 = real(wclock)
       cpu1 = real(cputime)
       timers(nsubstps,2,2,2) = timers(nsubstps,2,2,2) + wal1  - wal0
       timers(nsubstps,2,2,1) = timers(nsubstps,2,2,1) + cpu1  - cpu0
#endif /* MPE_USED */
#endif /* PROFILING */

c
c    2) Do second portion of the interior points, and some on borders.
c
       call avisc (is  ,is  ,js+1,je-1,ks+1,k1  ,dvdxmn
     &            ,v1old,v2old,v3old,v1new,v2new,v3new,s1,s2,s3)
       call avisc (ie  ,ie  ,js+1,je-1,ks+1,k1  ,dvdxmn
     &            ,v1old,v2old,v3old,v1new,v2new,v3new,s1,s2,s3)
       call avisc (is  ,ie  ,js+1,je-1,k1+1,k2  ,dvdxmn
     &            ,v1old,v2old,v3old,v1new,v2new,v3new,s1,s2,s3)

#ifdef PROFILING
#ifdef MPE_USED
       call MPE ...
#else
       call clocks (cputime, wclock)
       wal0 = wal1
       cpu0 = cpu1
       wal1 = real(wclock)
       cpu1 = real(cputime)
       timers(2,2,2,2) = timers(2,2,2,2) + wal1  - wal0
       timers(2,2,2,1) = timers(2,2,2,1) + cpu1  - cpu0
#endif /* MPE_USED */
#endif /* PROFILING */

#ifdef MPI_USED
c
c    3) Wait for communications to complete.
c
       if (nreq .ne. 0) call MPI_WAITALL ( nreq, req, stat, ierr )

#ifdef PROFILING
#ifdef MPE_USED
       call MPE ...
#else
       call clocks (cputime, wclock)
       wal0 = wal1
       cpu0 = cpu1
       wal1 = real(wclock)
       cpu1 = real(cputime)
       timers(2,2,1,2) = timers(2,2,1,2) + wal1  - wal0
       timers(2,2,1,1) = timers(2,2,1,1) + cpu1  - cpu0
#endif /* MPE_USED */
#endif /* PROFILING */

#endif
c......................................................................
c
c k boundaries
c
c    1) Post sends and receives.
c
       nreq = 0
       nsub = nsub + 1
       call bvald  (0,0,0,0,1,0,d    )
       call bvalv3 (0,0,0,0,1,1,v3old)

#ifdef PROFILING
#ifdef MPE_USED
       call MPE ...
#else
       call clocks (cputime, wclock)
       wal0 = wal1
       cpu0 = cpu1
       wal1 = real(wclock)
       cpu1 = real(cputime)
       timers(nsubstps,3,2,2) = timers(nsubstps,3,2,2) + wal1  - wal0
       timers(nsubstps,3,2,1) = timers(nsubstps,3,2,1) + cpu1  - cpu0
#endif /* MPE_USED */
#endif /* PROFILING */

c
c    2) Do last portion of the interior points, and some on borders.
c
       call avisc (is  ,ie  ,js  ,js  ,ks+1,k2  ,dvdxmn
     &            ,v1old,v2old,v3old,v1new,v2new,v3new,s1,s2,s3)
       call avisc (is  ,ie  ,je  ,je  ,ks+1,k2  ,dvdxmn
     &            ,v1old,v2old,v3old,v1new,v2new,v3new,s1,s2,s3)
       call avisc (is  ,ie  ,js  ,je  ,k2+1,ke-1,dvdxmn
     &            ,v1old,v2old,v3old,v1new,v2new,v3new,s1,s2,s3)
c
c      Mark the boundary values out of date.
c
       do 20 i = 1,6
#ifndef ISO
         bvstat(i,2) = 0      !  e
#endif
         bvstat(i,3) = 0      !  v1
         bvstat(i,4) = 0      !  v2
         bvstat(i,5) = 0      !  v3
20     continue

#ifdef PROFILING
#ifdef MPE_USED
       call MPE ...
#else
       call clocks (cputime, wclock)
       wal0 = wal1
       cpu0 = cpu1
       wal1 = real(wclock)
       cpu1 = real(cputime)
       timers(2,3,2,2) = timers(2,3,2,2) + wal1  - wal0
       timers(2,3,2,1) = timers(2,3,2,1) + cpu1  - cpu0
#endif /* MPE_USED */
#endif /* PROFILING */

#ifdef MPI_USED
c
c    3) Wait for communications to complete.
c
       if (nreq .ne. 0) call MPI_WAITALL ( nreq, req, stat, ierr )

#ifdef PROFILING
#ifdef MPE_USED
       call MPE ...
#else
       call clocks (cputime, wclock)
       wal0 = wal1
       cpu0 = cpu1
       wal1 = real(wclock)
       cpu1 = real(cputime)
       timers(2,3,1,2) = timers(2,3,1,2) + wal1  - wal0
       timers(2,3,1,1) = timers(2,3,1,1) + cpu1  - cpu0
#endif /* MPE_USED */
#endif /* PROFILING */

#endif
c......................................................................
c
c Do the remaining border zones.
c
       call avisc (is  ,ie  ,js  ,je  ,ks  ,ks  ,dvdxmn
     &            ,v1old,v2old,v3old,v1new,v2new,v3new,s1,s2,s3)
       call avisc (is  ,ie  ,js  ,je  ,ke  ,ke  ,dvdxmn
     &            ,v1old,v2old,v3old,v1new,v2new,v3new,s1,s2,s3)
c
c......................................................................
c
c  Compute the viscous timestep.  Note that the minimum dv/dx is found
c  since it is less than 0.  Thus the minimum dv/dx gives the maximum
c  absolute value.  We'll do a global min on this later, in nudt.
c
       dtqqi2 = ( 4.0 * qcon * dvdxmn )**2

#ifdef PROFILING
#ifdef MPE_USED
       call MPE ...
#else
       call clocks (cputime, wclock)
       wal0 = wal1
       cpu0 = cpu1
       wal1 = real(wclock)
       cpu1 = real(cputime)
       timers(2,4,2,2) = timers(2,4,2,2) + wal1  - wal0
       timers(2,4,2,1) = timers(2,4,2,1) + cpu1  - cpu0
#endif /* MPE_USED */
#endif /* PROFILING */

       return
       end
c
c=======================================================================
c
c    \\\\\\\\\\        E N D   S U B R O U T I N E        //////////
c    //////////               A V I S C _ D               \\\\\\\\\\
c
c=======================================================================
c
