#include "zeusmp.def"

#ifdef PNEWT 
#define FORCE_LAW pnewt
#else /* PNEWT */
#define FORCE_LAW forces
#endif /* !PNEWT */

c
c=======================================================================
c
c    \\\\\\\\\\      B E G I N   S U B R O U T I N E      //////////
c    //////////              F O R C E S _ D              \\\\\\\\\\
c
c                            Developed by
c                Laboratory of Computational Astrophysics
c               University of Illinois at Urbana-Champaign
c
c=======================================================================
c
       subroutine forces_d
     1 (v1old, v2old, v3old, v1new, v2new, v3new)
c
c  PURPOSE
c    Driver for forces module.  Updates the velocities due to the
c    following body forces:
c
c    1) Thermal pressure gradient
c    2) Self-gravity
c    3) Rotational Pseudo-Forces
c    4) Gravitational point mass
c    5) Radiation force
c    6) Magnetic pressure
c
c  AUTHOR
c    Robert A. Fiedler
c
c  LAST MODIFIED
c    02/17/97, MHD on 3 Mar 98 by M-M Mac Low
c
c  INPUT
c    v*old    velocity components before acceleration
c
c  OUTPUT
c    v*new    velocity components after update
c
c  EXTERNALS:
c    BVALV1  , BVALV2  , BVALV3
c    BVALD   , BVALE   , BVALER
c    FORCE_LAW
c
c.......................................................................
c
       implicit NONE
#include "mod_names.h"
#include "param.h"
#include "grid.h"

       REAL v1old(in,jn,kn), v2old(in,jn,kn), v3old(in,jn,kn)
     1    , v1new(in,jn,kn), v2new(in,jn,kn), v3new(in,jn,kn)

#ifndef NO_FORCES

#include "root.h"
#include "field.h"
#include "bndry.h"
#include "scratch.h"
#include "mpicom.h"
#if defined RAD || defined RAD_EXP
#include "radexp.h"
#endif /* RAD || RAD_EXP */
#ifdef PROFILING
#ifdef MPE_USED
#else
#include "clocks.h"
#endif /* MPE_USED */
#endif /* PROFILING */
c
       integer i, k1, k2, imax, jmax, kmax
c
c      External statements
c
       external      FORCE_LAW
     1             , bvalv1  , bvalv2  , bvalv3  , bvale   , bvald

#if defined RAD || defined RAD_EXP
     3             , bvalers
#endif /* RAD || RAD_EXP */
c
c-----------------------------------------------------------------------
c
c Divide the computational volume into three equal pieces.  We must
c have at least 5 active zones in the 3-direction.
c
       k1 = int( real( ke - ks + 1 ) / 3.0 ) + ks
       k2 = int( real( ke - ks + 1 ) / 3.0 ) + k1
c
c
#ifdef DEBUG
       write(*,"(/'FORCES_D: ',i2,' calling FORCE_LAW.')") myid
#endif /* DEBUG */

#ifdef PROFILING
#ifdef MPE_USED
       call MPE ...
#else
       call clocks (cputime, wclock)
       wal1 = real(wclock)
       cpu1 = real(cputime)
#endif /* MPE_USED */
#endif /* PROFILING */

c......................................................................
c
c i boundaries
c
c    1) Post sends and receives.  We never use d(i+1) in any of the
c       source term substeps, so we need to pass only 1 "m" layer
c       of density boundary values.  We also need only one
c       "m" layer of energy values.  If we need the velocity, we
c       need just one layer, but both "m" and "p".  This should
c       minimize communication for any physics/geometry.  By
c       exchanging the i, j, and k boundary data in three separate 
c       steps, we ensure that the corner and edge data are correctly
c       updated.
c
       nreq = 0
       nsub = nsub + 1
       call bvald  (1,0,0,0,0,0,d )
#ifndef ISO
       call bvale  (1,0,0,0,0,0,e )
#endif /* not ISO */
#ifdef TOTAL_ENERGY
       imax = ie - 1
       call bvalv1 (1,1,0,0,0,0,v1old)
#else  /* TOTAL_ENERGY */
       imax = ie
#endif /* TOTAL_ENERGY */
#if defined TOTAL_ENERGY || defined RTP
       jmax = je - 1
       call bvalv2 (1,0,0,0,0,0,v2old)
       call bvalv3 (1,0,0,0,0,0,v3old)
#else  /* TOTAL_ENERGY || RTP */
       jmax = je
#endif /* TOTAL_ENERGY || RTP */
C#if defined GRAV || defined GRAV_FFT
C       call bvalgp (1,1,0,0,0,0,gp)
C#endif /* GRAV || GRAV_FFT */
#if defined RAD || defined RAD_EXP
       call bvalers (1,0,0,0,0,0,er)
#endif /* RAD || RAD_EXP */

#ifdef PROFILING
#ifdef MPE_USED
       call MPE ...
#else
       call clocks (cputime, wclock)
       wal0 = wal1
       cpu0 = cpu1
       wal1 = real(wclock)
       cpu1 = real(cputime)
       timers(nsubstps,1,2,2) = timers(nsubstps,1,2,2) + wal1  - wal0
       timers(nsubstps,1,2,1) = timers(nsubstps,1,2,1) + cpu1  - cpu0
#endif /* MPE_USED */
#endif /* PROFILING */

c
c    2) Do first portion of the interior points.
c
       call FORCE_LAW (is+1,imax,js+1,jmax,ks+1,k1
     &             ,v1old,v2old,v3old,v1new,v2new,v3new)
c
c       subroutine FORCE_LAW (ibeg,iend,jbeg,jend,kbeg,kend
c     &                   ,u1,u2,u3,w1,w2,w3)
cc
cc Arrays u1 , u2 , u3  hold the old velocity values, while
cc        w1 , w2 , w3  receive the updated values.
c

#ifdef PROFILING
#ifdef MPE_USED
       call MPE ...
#else
       call clocks (cputime, wclock)
       wal0 = wal1
       cpu0 = cpu1
       wal1 = real(wclock)
       cpu1 = real(cputime)
       timers(1,1,2,2) = timers(1,1,2,2) + wal1  - wal0
       timers(1,1,2,1) = timers(1,1,2,1) + cpu1  - cpu0
#endif /* MPE_USED */
#endif /* PROFILING */

#ifdef MPI_USED
c
c    3) Wait for communications to complete.
c
       if (nreq .ne. 0) call MPI_WAITALL ( nreq, req, stat, ierr )

#ifdef PROFILING
#ifdef MPE_USED
       call MPE ...
#else
       call clocks (cputime, wclock)
       wal0 = wal1
       cpu0 = cpu1
       wal1 = real(wclock)
       cpu1 = real(cputime)
       timers(1,1,1,2) = timers(1,1,1,2) + wal1  - wal0
       timers(1,1,1,1) = timers(1,1,1,1) + cpu1  - cpu0
#endif /* MPE_USED */
#endif /* PROFILING */

#endif /* MPI_USED */
c......................................................................
c
c j boundaries
c
c    1) Post sends and receives.  We need d(j+1) only for RTP.
c
       nreq = 0
       nsub = nsub + 1
#ifdef RTP
       call bvald  (0,0,1,1,0,0,d )
#else /* RTP */
       call bvald  (0,0,1,0,0,0,d )
#endif /* RTP */
#ifndef ISO
       call bvale  (0,0,1,0,0,0,e )
#endif /* not ISO */
#ifdef TOTAL_ENERGY
       call bvalv1 (0,0,1,0,0,0,v1old)
#endif /* TOTAL_ENERGY */
#ifdef TOTAL_ENERGY
       call bvalv2 (0,0,1,1,0,0,v2old)
#else /* TOTAL_ENERGY */
#ifdef RTP
       call bvalv2 (0,0,0,1,0,0,v2old)
#endif /* RTP */
#endif /* TOTAL_ENERGY */
#if defined TOTAL_ENERGY || defined ZRP || defined RTP
       call bvalv3 (0,0,1,0,0,0,v3old)
#endif /* TOTAL_ENERGY || ZRP || RTP */
C#if defined GRAV || defined GRAV_FFT
C       call bvalgp (0,0,1,1,0,0,gp)
C#endif /* GRAV || GRAV_FFT*/
#if defined RAD || defined RAD_EXP
       call bvalers (0,0,1,0,0,0,er)
#endif /* RAD || RAD_EXP */

#ifdef PROFILING
#ifdef MPE_USED
       call MPE ...
#else
       call clocks (cputime, wclock)
       wal0 = wal1
       cpu0 = cpu1
       wal1 = real(wclock)
       cpu1 = real(cputime)
       timers(nsubstps,2,2,2) = timers(nsubstps,2,2,2) + wal1  - wal0
       timers(nsubstps,2,2,1) = timers(nsubstps,2,2,1) + cpu1  - cpu0
#endif /* MPE_USED */
#endif /* PROFILING */

c
c    2) Do middle portion of the interior points, plus some on borders.
c
       call FORCE_LAW (is  ,is  ,js+1,jmax,ks+1,k1
     &             ,v1old,v2old,v3old,v1new,v2new,v3new)
#if defined TOTAL_ENERGY
       call FORCE_LAW (ie  ,ie  ,js+1,jmax,ks+1,k1
     &             ,v1old,v2old,v3old,v1new,v2new,v3new)
#endif /* TOTAL_ENERGY */
       call FORCE_LAW (is  ,ie  ,js+1,jmax,k1+1,k2
     &             ,v1old,v2old,v3old,v1new,v2new,v3new)

#ifdef PROFILING
#ifdef MPE_USED
       call MPE ...
#else
       call clocks (cputime, wclock)
       wal0 = wal1
       cpu0 = cpu1
       wal1 = real(wclock)
       cpu1 = real(cputime)
       timers(1,2,2,2) = timers(1,2,2,2) + wal1  - wal0
       timers(1,2,2,1) = timers(1,2,2,1) + cpu1  - cpu0
#endif /* MPE_USED */
#endif /* PROFILING */

#ifdef MPI_USED
c
c    3) Wait for communications to complete.
c
       if (nreq .ne. 0) call MPI_WAITALL ( nreq, req, stat, ierr )

#ifdef PROFILING
#ifdef MPE_USED
       call MPE ...
#else
       call clocks (cputime, wclock)
       wal0 = wal1
       cpu0 = cpu1
       wal1 = real(wclock)
       cpu1 = real(cputime)
       timers(1,2,1,2) = timers(1,2,1,2) + wal1  - wal0
       timers(1,2,1,1) = timers(1,2,1,1) + cpu1  - cpu0
#endif /* MPE_USED */
#endif /* PROFILING */

#endif /* MPI_USED */
c......................................................................
c
c k boundaries
c
c    1) Post sends and receives.  We need d(k+1) for ZRP or RTP.
c
       nreq = 0
       nsub = nsub + 1
#if defined ZRP || defined RTP
       call bvald  (0,0,0,0,1,1,d )
#else /* ZRP || RTP */
       call bvald  (0,0,0,0,1,0,d )
#endif /* ZRP || RTP */
#ifndef ISO
       call bvale  (0,0,0,0,1,0,e )
#endif /* not ISO */
#ifdef TOTAL_ENERGY
       call bvalv1 (0,0,0,0,1,0,v1old)
#endif /* TOTAL_ENERGY */
#ifdef TOTAL_ENERGY
       call bvalv2 (0,0,0,0,1,0,v2old)
#endif /* TOTAL_ENERGY */
#if defined TOTAL_ENERGY
       kmax = ke - 1
       call bvalv3 (0,0,0,0,1,1,v3old)
#else  /* TOTAL_ENERGY */
#if defined ZRP || defined RTP
       kmax = ke - 1
       call bvalv3 (0,0,0,0,0,1,v3old)
#else
       kmax = ke
#endif /* ZRP || RTP */
#endif /* TOTAL_ENERGY */
C#if defined GRAV || defined GRAV_FFT
C       call bvalgp (0,0,0,0,1,1,gp)
C#endif /* GRAV || GRAV_FFT*/
#if defined RAD || defined RAD_EXP
       call bvalers (0,0,0,0,1,0,er)
#endif /* RAD || RAD_EXP */

#ifdef PROFILING
#ifdef MPE_USED
       call MPE ...
#else
       call clocks (cputime, wclock)
       wal0 = wal1
       cpu0 = cpu1
       wal1 = real(wclock)
       cpu1 = real(cputime)
       timers(nsubstps,3,2,2) = timers(nsubstps,3,2,2) + wal1  - wal0
       timers(nsubstps,3,2,1) = timers(nsubstps,3,2,1) + cpu1  - cpu0
#endif /* MPE_USED */
#endif /* PROFILING */

c
c    2) Do last portion of the interior points, plus some on borders.
c
       call FORCE_LAW (is  ,ie  ,js  ,js  ,ks+1,k2
     &             ,v1old,v2old,v3old,v1new,v2new,v3new)
#if defined TOTAL_ENERGY || defined RTP
       call FORCE_LAW (is  ,ie  ,je  ,je  ,ks+1,k2
     &             ,v1old,v2old,v3old,v1new,v2new,v3new)
#endif /* TOTAL_ENERGY || RTP */
       call FORCE_LAW (is  ,ie  ,js  ,je  ,k2+1,kmax
     &             ,v1old,v2old,v3old,v1new,v2new,v3new)
c
c      Mark the velocity boundary values out of date.
c
       do 10 i = 1,6
         bvstat(i,3) = 0      !  v1
         bvstat(i,4) = 0      !  v2
         bvstat(i,5) = 0      !  v3
10     continue

#ifdef PROFILING
#ifdef MPE_USED
       call MPE ...
#else
       call clocks (cputime, wclock)
       wal0 = wal1
       cpu0 = cpu1
       wal1 = real(wclock)
       cpu1 = real(cputime)
       timers(1,3,2,2) = timers(1,3,2,2) + wal1  - wal0
       timers(1,3,2,1) = timers(1,3,2,1) + cpu1  - cpu0
#endif /* MPE_USED */
#endif /* PROFILING */

#ifdef MPI_USED
c
c    3) Wait for communications to complete. 
c
       if (nreq .ne. 0) call MPI_WAITALL ( nreq, req, stat, ierr )

#ifdef PROFILING
#ifdef MPE_USED
       call MPE ...
#else
       call clocks (cputime, wclock)
       wal0 = wal1
       cpu0 = cpu1
       wal1 = real(wclock)
       cpu1 = real(cputime)
       timers(1,3,1,2) = timers(1,3,1,2) + wal1  - wal0
       timers(1,3,1,1) = timers(1,3,1,1) + cpu1  - cpu0
#endif /* MPE_USED */
#endif /* PROFILING */

#endif /* MPI_USED */
c......................................................................
c
c Finally, do the remaining border zones.
c
       call FORCE_LAW (is  ,ie  ,js  ,je  ,ks  ,ks  
     &             ,v1old,v2old,v3old,v1new,v2new,v3new)
#if defined TOTAL_ENERGY || defined ZRP || defined RTP
       call FORCE_LAW (is  ,ie  ,js  ,je  ,ke  ,ke  
     &             ,v1old,v2old,v3old,v1new,v2new,v3new)
#endif /* TOTAL_ENERGY || ZRP || RTP */

#ifdef PROFILING
#ifdef MPE_USED
       call MPE ...
#else
       call clocks (cputime, wclock)
       wal0 = wal1
       cpu0 = cpu1
       wal1 = real(wclock)
       cpu1 = real(cputime)
       timers(1,4,2,2) = timers(1,4,2,2) + wal1  - wal0
       timers(1,4,2,1) = timers(1,4,2,1) + cpu1  - cpu0
#endif /* MPE_USED */
#endif /* PROFILING */

#else  /* not NO_FORCES */
       integer i,j,k
c
c-----------------------------------------------------------------------
c
c Just copy the old velocities to the new for all active zones.
c
       do k=ks,ke
         do j=js,je
           do i=is,ie
             v1new(i,j,k) = v1old(i,j,k)
             v2new(i,j,k) = v2old(i,j,k)
             v3new(i,j,k) = v3old(i,j,k)
           enddo ! i
         enddo ! j
       enddo ! k
#endif /* not NO_FORCES */

       return
       end
c
c=======================================================================
c
c    \\\\\\\\\\        E N D   S U B R O U T I N E        //////////
c    //////////             F O R C E S _ D               \\\\\\\\\\
c
c=======================================================================
c
