#include "zeusmp.def"
c=======================================================================
c
c    \\\\\\\\\\      B E G I N   S U B R O U T I N E      //////////
c    //////////                   J E T                   \\\\\\\\\\
c
c                            Developed by
c                Laboratory of Computational Astrophysics
c               University of Illinois at Urbana-Champaign
c
c=======================================================================
c
       subroutine jet
c
c    dac:zeus3d.jetinit <------------------- initialises jet calculation
c                                                              may, 1990
c
c    written by: David Clarke
c    modified 1: December, 1990 by Chris Loken and David Clarke; added
c                poloidal magnetic field generator and made compatable
c                with XYZ and ZRP coordinates in 2-D.
c    modified 2: December, 1999 by PS Li; for ZeusMP
c
c  PURPOSE:  This subroutine initialises the inflow boundary conditions
c  and the grid for a jet (either propagating or equilibrium) in one of
c  the following geometries:
c
c  3-D cylindrical jet in Cartesian   coordinates  -  set XYZ
c  3-D cylindrical jet in cylindrical coordinates  -  set ZRP
c
c  For a 3-D jet, the cross-section of the inlet is as circular as 
c  possible given rectangular zones.  A uniform atmosphere is also 
c  initialised.  The jet will enter the grid from the inner i-boundary.
c  The jet axis will pass through the origin, and so the user must 
c  ensure that the origin lies on the inner i boundary.  Note that for 
c  XYZ geometry, the jet will extend above and below the origin, whereas
c  for ZRP geometry, the jet will only lie above the origin (positive
c  values of r only).
c
c  Time-varying inflow boundary conditions for the velocity components
c  are possible by setting the WIGGLE, which effectively applies a 
c  helical precession to the orifice.  Input data for the precession are
c  read from namelist "pgen" in this subroutine and placed in the common
c  blocks.
c
c  LOCAL VARIABLES:
c
c  EXTERNALS:
c    OVERLAP
c
c-----------------------------------------------------------------------
c
       implicit NONE
#include "param.h"
#include "field.h"
#include "bndry.h"
#include "grid.h"
#include "root.h"
#include "scratch.h"
#if defined RAD || defined RAD_EXP
#include "radiation.h"
#include "opac.h"
#include "radexp.h"
#include "cons.h"
#endif /* RAD || RAD_EXP */
#include "mod_names.h"
#include "mpicom.h"
c
       integer       i       , j       , jp1     , k       , kp1
     1             , pore    , jlm1    , jlp1    , jum1
     2             , jup1    , klm1    , klp1    , kum1    , kup1
     3             , jave    , kave
     4             , jjetmn  , jjetmx  , kjetmn  , kjetmx 
       REAL          mjet    , prat    , drat    , b1jet   , d0
     1             , e0      , v10     , v20     , v30     , b10
     2             , b20     , b30     , rjet    , ojet    , x10
     3             , x20     , x30
       REAL          djet    , ejet    , v1jet   , v2jet   , v3jet
     1             , pjet    , p0      , xsarea  , rsq     , rinsq
     2             , routsq  , frac    , cofrac  , width   , fact
c
       integer       jin     (ijkn), jout    (ijkn), kin     (ijkn)
     1             , kout    (ijkn)
c
       equivalence   ( jin     , w1da     ), ( jout    , w1db     )
     1             , ( kin     , w1dc     ), ( kout    , w1dd     )
c
c      External statements
c
       REAL        overlap
       external    overlap
c
c-----------------------------------------------------------------------
c
c  Input parameters for the jet material:
c
c  hydrodynamical variables:
c    rjet      radius of jet (in coordinate units, default = 1.0)
c    drat      ratio between jet density and ambient medium density
c              effective at the in-flow boundary (default = 0.1)
c    prat      ratio between jet pressure and ambient medium pressure
c              effective at the in-flow boundary (default = 1.0)
c    mjet      Mach number of the jet with respect to its own sound
c              speed at the jet inlet (default = 6.0).
c    ojet      initial angular velocity of jet about axis (solid body
c              rotation, default = 0.0)
c    pore      =1 => propagating jet
c              =2 => equilibrium jet
c    b1jet     initial magnetic field in 1-direction (equilibrium jet
c              only, default = 0.0)
c
c  Input parameters for the ambient medium:
c
c    d0      density (default = 1.0)
c    e0      specific internal energy ( = e/d , default = 0.9 which
c              gives ambient sound speed = 1.0)
c    p0      pressure ( default = 0.6 )
c    v10     velocity in 1 direction (default = 0.0)
c    v20     velocity in 2 direction (default = 0.0)
c    v30     velocity in 3 direction (default = 0.0)
c    b10     magnetic field in 1 direction (default = 0.0)
c    b20     magnetic field in 2 direction (default = 0.0)
c    b30     magnetic field in 3 direction (default = 0.0)
c
       namelist / pgen     /
     1               rjet    , x10     , x20     , x30     , drat
     2             , prat    , mjet    , ojet    , p0      , mjet
     2             , pore    , b1jet   , d0      , e0      , v10   
     3             , v20     , v30     , b10     , b20     , b30     
c
c      Default values
c
       rjet    = 1.0
       drat    = 0.1
       prat    = 1.0
       mjet    = 6.0
       ojet    = 0.0
       pore    = 1
       b1jet   = 0.0
       d0      = 1.0
       e0      = 0.9
       p0      = 0.6
       x10     = 0.0
       x20     = 0.0
       x30     = 0.0
       v10     = 0.0
       v20     = 0.0
       v30     = 0.0
       b10     = 0.0
       b20     = 0.0
       b30     = 0.0
c
       if(myid .eq. 0) then
         read (1, pgen)
         write (2, pgen)
#ifdef MPI_USED
         buf_in( 1) = rjet
         buf_in( 2) = drat 
         buf_in( 3) = prat 
         buf_in( 4) = mjet
         buf_in( 5) = ojet
         buf_in( 6) = b1jet  
         buf_in( 7) = d0
         buf_in( 8) = e0
         buf_in( 9) = p0  
         buf_in(10) = x10 
         buf_in(11) = x20 
         buf_in(12) = x30 
         buf_in(13) = v10 
         buf_in(14) = v20 
         buf_in(15) = v30 
         buf_in(16) = b10 
         buf_in(17) = b20 
         buf_in(18) = b30 

#if defined RAD || defined RAD_EXP
         buf_in(19) = er0
         buf_in(20) = er1
#endif /* RAD || RAD_EXP */
         ibuf_in(1) = pore
       endif
#if defined RAD || defined RAD_EXP
       call MPI_BCAST( buf_in, 20, MPI_FLOAT
     &               , 0, comm3d, ierr )
#else /* RAD || RAD_EXP */
       call MPI_BCAST( buf_in, 18, MPI_FLOAT
     &               , 0, comm3d, ierr )
#endif /* RAD || RAD_EXP */
       call MPI_BCAST( ibuf_in, 1, MPI_INTEGER
     &               , 0, comm3d, ierr )
       if (myid .ne. 0) then
         rjet    = buf_in( 1)
         drat    = buf_in( 2)
         prat    = buf_in( 3)
         mjet    = buf_in( 4)
         ojet    = buf_in( 5)
         b1jet   = buf_in( 6)
         d0      = buf_in( 7)
         e0      = buf_in( 8)
         p0      = buf_in( 9)
         x10     = buf_in(10)
         x20     = buf_in(11)
         x30     = buf_in(12)
         v10     = buf_in(13)
         v20     = buf_in(14)
         v30     = buf_in(15)
         b10     = buf_in(16)
         b20     = buf_in(17)
         b30     = buf_in(18)
#if defined RAD || defined RAD_EXP
         er0     = buf_in(19)
         er1     = buf_in(20)
#endif /* RAD || RAD_EXP */
         pore    = ibuf_in(1)
#endif /* MPI_USED */
       endif

       if (e0 .ne. 0.0) p0 = e0 * gamm1
       if (e0 .eq. 0.0) e0 = p0 / gamm1
#if defined RAD || defined RAD_EXP
       if (er0 .eq. 0.0) er0 = rad_con * (gamm1*e0*mmw*mh/(d0*boltz))**4
#endif /* RAD || RAD_EXP */
c
c      Set up atmosphere
c
       do 30 k=ks-2,ke+3
         do 20 j=js-2,je+3
           do 10 i=is-2,ie+3
             d (i,j,k) = d0
             v1(i,j,k) = v10
             v2(i,j,k) = v20
             v3(i,j,k) = v30
#ifndef ISO
             e (i,j,k) = e0
#endif
#if defined RAD || defined RAD_EXP
             er(i,j,k) = er0
#endif /* RAD || RAD_EXP */
#ifdef MHD
             b1(i,j,k) = b10 * g2ai(i) * g31ai(i)
             b2(i,j,k) = b20           * g32ai(j)
             b3(i,j,k) = b30 * g2bi(i)
#endif MHD
 10        continue
 20      continue
 30    continue
c
c      Determine the extent of the jet in j (and k for XYZ).
c
#ifdef XYZ
       kjetmn = ks
       do 40 k=ks,ke+1
         if ( x3a(k) .lt. -rjet ) kjetmn = k
40     continue
       kjetmx = ke
       do 50 k=ke+1,ks,-1
         if ( x3a(k) .ge.  rjet ) kjetmx = k
50     continue
       klm1 = kjetmn - 1
       klp1 = kjetmn + 1
       kum1 = kjetmx - 1
       kup1 = kjetmx + 1
       kave = ( kjetmx + kjetmn ) / 2
c
       jjetmn = js
       do 60 j=js,je+1
         if ( x2a(j) .lt. -rjet ) jjetmn = j
60     continue
       jjetmx = je+1
       do 70 j=je+1,js,-1
         if ( x2a(j) .ge.  rjet ) jjetmx = j
70     continue
       jlm1 = jjetmn - 1
       jlp1 = jjetmn + 1
       jum1 = jjetmx - 1
       jup1 = jjetmx + 1
       jave = ( jjetmx + jjetmn ) / 2
#endif XYZ
#ifdef ZRP
       kjetmn = ks
       kjetmx = ke
       klm1 = kjetmn - 1
       klp1 = kjetmn + 1
       kum1 = kjetmx - 1
       kup1 = kjetmx + 1
       kave = ( kjetmx + kjetmn ) / 2
       jjetmn = js
       jjetmx = je
       do 70 j=je,js,-1
         if ( x2a(j) .ge.  rjet ) jjetmx = j
70     continue
       jlm1 = jjetmn - 1
       jlp1 = jjetmn + 1
       jum1 = jjetmx - 1
       jup1 = jjetmx + 1
       jave = ( jjetmx + jjetmn ) / 2
#endif              
c
c      Determine the corner of each zone which lies closest to the
c  origin (jin,kin) and furthest from the origin (jout,kout).
c
       do 80 j=jjetmn,jjetmx
         jp1 = j + 1
         if ( abs(x2a(j)) .lt. abs(x2a(jp1)) ) then
           jin (j) = j
           jout(j) = jp1
         else
           jin (j) = jp1
           jout(j) = j
         endif
80     continue
c
#ifdef XYZ
       do 90 k=kjetmn,kjetmx
         kp1 = k + 1
         if ( abs(x3a(k)) .lt. abs(x3a(kp1)) ) then
           kin (k) = k
           kout(k) = kp1
         else
           kin (k) = kp1
           kout(k) = k
         endif
90     continue
#endif XYZ
c
c      Determine other jet parameters
c
       width = rjet
       djet  = drat * d0
       pjet  = prat * p0
       ejet  = pjet / gamm1
       v1jet = mjet * sqrt ( gamma * pjet / djet )
c
c-----------------------------------------------------------------------
c---------------- SET HYDRODYNAMICAL BOUNDARY VARIABLES ----------------
c-----------------------------------------------------------------------
c
       xsarea = 0.0

       do 130 k=kjetmn,kjetmx
         v2jet = ojet * x3b(k)
         do 120 j=jjetmn,jjetmx
           v3jet  =-ojet * x2b(j)
           rsq    = rjet**2
#ifdef XYZ
           rinsq  = x2a(jin (j))**2 + x3a(kin (k))**2
           routsq = x2a(jout(j))**2 + x3a(kout(k))**2
#endif
#ifdef ZRP
           rinsq  = x2a(jin (j))**2
           routsq = x2a(jout(j))**2
#endif
c
c  Case:  Zones are completely contained by the inlet cross section.
c
           if ((routsq .le. rsq) .and. (x1a(is) .eq. 0.0)) then
             niib   (j,k) = 3
             diib (j,k,1) = djet
             diib (j,k,2) = djet
             v1iib(j,k,1) = v1jet
             v1iib(j,k,2) = v1jet
#ifndef ISO
             eiib (j,k,1) = ejet
             eiib (j,k,2) = ejet
#endif
             xsarea      = xsarea
     1                   + dx2a(j) * g2b (is-1)
     2                   * dx3a(k) * g31b(is-1) * g32b(j)
             if (pore .eq. 2) then
               do 100 i=is,ie
                 d (i,j,k) = djet
                 v1(i,j,k) = v1jet
                 v2(i,j,k) = v2jet
                 v3(i,j,k) = v3jet
#ifndef ISO
                 e (i,j,k) = ejet
#endif
#ifdef MHD
                 b1(i,j,k) = b1jet
#endif MHD
100            continue
             endif
           endif
c
c  Case:  Zones are partially contained by the inlet cross section .
c
           if ( (rinsq .lt. rsq) .and. (routsq .gt. rsq) .and.
     1          (x1a(is) .eq. 0.0) ) then
#ifdef XYZ
             frac = overlap ( 2, rjet, x10, x20         , x30
     1                         , zro, x2a(jin (j)), x3a(kin (k))
     2                         , zro, x2a(jout(j)), x3a(kout(k)) )
#endif XYZ
#ifdef ZRP
             frac = ( rsq - rinsq ) / ( routsq - rinsq )
#endif ZRP
             cofrac      = 1.0 - frac
             niib   (j,k) = 3
             diib (j,k,1) = djet * frac
             diib (j,k,2) = djet * frac
             v1iib(j,k,1) = v1jet
             v1iib(j,k,2) = v1jet
#ifndef ISO
             eiib (j,k,1) = ejet * frac
             eiib (j,k,2) = ejet * frac
#endif -ISO
             xsarea     = xsarea
     1                  + dx2a(j) * g2b (is-1)
     2                  * dx3a(k) * g31b(is-1) * g32b(j) * frac
             if (pore .eq. 2) then
               do 110 i=is,ie
                 v1(i,j,k) = v1jet         *   frac * djet / d(i,j,k)
     1                     + v1(i,j,k)     * cofrac
                 v2(i,j,k) = ojet * x3b(k) *   frac * djet / d(i,j,k)
     1                     + v2(i,j,k)     * cofrac
                 v3(i,j,k) =-frac * djet / d(i,j,k) * ojet * x2b(j)
     1                     + d(i,j,k) * v3(i,j,k) * cofrac
                 d (i,j,k) = djet  * frac + d (i,j,k) * cofrac
#ifndef ISO
                 e (i,j,k) = ejet  * frac + e (i,j,k) * cofrac
#endif -ISO
#ifdef MHD
                 b1(i,j,k) = b1jet * frac + b1(i,j,k) * cofrac
#endif MHD
110            continue
             endif
           endif
120      continue
130    continue
c
       return
       end
c
c=======================================================================
c
c    \\\\\\\\\\        E N D   S U B R O U T I N E        //////////
c    //////////               J E T I N I T               \\\\\\\\\\
c
c=======================================================================
c
c
#include "zeusmp.def"
c=======================================================================
c
c    \\\\\\\\\\        B E G I N   F U N C T I O N        //////////
c    //////////               O V E R L A P               \\\\\\\\\\
c
c=======================================================================
c
#if defined SINGLE_PRECISON || defined UNICOS
       real function overlap ( ishp, rad, x0, y0, z0, xin, yin
#else
       real*8 function overlap ( ishp, rad, x0, y0, z0, xin, yin
#endif
     1                       , zin, xout, yout, zout )
c
c    dac:zeus3d.overlap <--------- overlap of region over Cartesian zone
c                                                            april, 1990
c
c    written by: David Clarke
c    modified 1:
c
c  PURPOSE:  Determines the fraction of a Cartesian zone that overlaps
c  the specified geometrical region (sphere or right cylinder).  This
c  is done by dividing the zone into 20**3 "subzones", and finding the
c  fraction of subzone centres lying inside the surface of the region.
c
c  INPUT VARIABLES:
c
c    ishp            =1 => sphere
c                    =2 => right cylinder
c    rad             radius of region
c    x0,y0,z0        coordinates of centre of curvature.
c    xin,yin,zin     coordinates of zone corner known to lie inside
c                    region.
c    xout,yout,zout  coordinates of zone corner diametrically opposed to
c                    zone corner known to lie inside region.
c
c  OUTPUT VARIABLES:
c
c  LOCAL VARIABLES:
c
c  EXTERNALS: [NONE]
c
c-----------------------------------------------------------------------
c
       implicit NONE
#include "param.h"
c
       integer       i       , j       , k       , nx      , ny
     1             , nz      , ishp
c       REAL        overlap
       REAL        rad     , x0      , y0      , z0      , xin
     1             , yin     , zin     , xout    , yout    , zout
     2             , delx    , dely    , delz    , r       , fact
     3             , scount
c
       REAL        xsq     (  20), ysq     (  20), zsq     (  20)
     1             , count   (  20)
c
c-----------------------------------------------------------------------
c
#ifdef XYZ
c      Number of subzones in the x-direction is "nx", etc. for "ny" and
c  "nz".  Increment between subzones in x-direction is "delx", etc. for
c  "dely" and "delz".
c
       nx   = 20
       ny   = 20
       nz   = 20
       delx = ( xout - xin ) /  real( nx )
       dely = ( yout - yin ) /  real( ny )
       delz = ( zout - zin ) /  real( nz )
c
c      Set up subgrid inside zone.
c
       do 10 i=1,nx
         xsq  (i) = ( xin + ( 0.5 +  real(i-1) ) * delx - x0 )**2
         count(i) = 0.0
10     continue
       do 20 j=1,ny
         ysq  (j) = ( yin + ( 0.5 +  real(j-1) ) * dely - y0 )**2
20     continue
       do 30 k=1,nz
         zsq  (k) = ( zin + ( 0.5 +  real(k-1) ) * delz - z0 )**2
30     continue
c
c      Count the number of subzones lying inside the surface of the
c  region which passes through the zone.
c
       fact  = 1.0
       if (ishp .eq. 2) fact = 0.0
       do 60 k=1,nz
         do 50 j=1,ny
           do 40 i=1,nx
             r = sqrt ( fact * xsq(i) + ysq(j) + zsq(k) )
             if (r .le. rad) count(i) = count(i) + 1.0
40         continue
50       continue
60     continue
       scount = 0.0
       do 70 i=1,nx
         scount = scount + count(i)
70     continue
       scount =   max ( one, scount )
c
c      Set the fraction of the zone which overlaps the region.
c
       overlap = scount /  real( nx * ny * nz )
#else  /* XYZ */
       overlap = 1.0
#endif /* XYZ */
c
       return
       end
c
c=======================================================================
c
c    \\\\\\\\\\          E N D   F U N C T I O N          //////////
c    //////////               O V E R L A P               \\\\\\\\\\
c
c=======================================================================
c
c
