#include "zeusmp.def"
c
c=======================================================================
c
c    \\\\\\\\\\      B E G I N   S U B R O U T I N E      //////////
c    //////////              L O R E N T Z _ D            \\\\\\\\\\
c
c                            Developed by
c                Laboratory of Computational Astrophysics
c               University of Illinois at Urbana-Champaign
c
c=======================================================================
c
       subroutine lorentz_d (w1, w2, w3, u1, u2, u3)

c
c  PURPOSE
c    Driver for lorentz update. Updates the velocities due to the
c    transverse magnetic forces (the longitudinal forces are handled 
c    in FORCES).  Adapted from FORCES_D; but LORENTZ needs two slabs 
c    of velocity both above and below
c
c  AUTHOR
c    Mordecai-Mark Mac Low
c
c  LAST MODIFIED
c    3 Mar 98 by M-M Mac Low
c    5/03/2006 jcv
c    added changes from zeusmpv2 (from John Hayes) so only a single call
c    to lorentz is made as the previous way (several blocks) doesn't
c    work correctly in parallel
c    5/26/2006 jcv
c    added fix from zeusmpv2 (from John Hayes) to put velocities in 
c    a temporary array to fix an old bug
c
c  INPUT
c    w[1,2,3]    velocity components before acceleration
c
c  OUTPUT
c    u[1,2,3]    velocity components after update
c
c  EXTERNALS:
c    BVALV1  , BVALV2  , BVALV3
c    BVALD   , BVALE  
c    LORENTZ
c
c.......................................................................
c
       implicit NONE
#include "mod_names.h"
#include "param.h"
#include "grid.h"
#include "field.h"

       REAL u1(in,jn,kn), u2(in,jn,kn), u3(in,jn,kn)
       REAL w1(in,jn,kn), w2(in,jn,kn), w3(in,jn,kn)



#include "root.h"

#include "bndry.h"
#include "scratch.h"
#include "mpicom.h"

c
       integer i,j,k, k1, k2
c
c      External statements
c
       external      lorentz
     1             , bvalv1  , bvalv2  , bvalv3  , bvale   , bvald

c
c copy old velocities to new arrays]

       do k=ks,ke
          do j=js,je
             do i=is,ie
                u1(i,j,k)=w1(i,j,k)
                u2(i,j,k)=w2(i,j,k)
                u3(i,j,k)=w3(i,j,k)
             enddo
          enddo
       enddo

#ifndef NO_FORCES

c
c-----------------------------------------------------------------------
c
c Divide the computational volume into three equal pieces.  We must
c have at least 5 active zones in the 3-direction.
c
       k1 = int( real( ke - ks + 1 ) / 3.0 ) + ks
       k2 = int( real( ke - ks + 1 ) / 3.0 ) + k1
c
c
#ifdef DEBUG
       write(*,"(/'LORENTZ_D: ',i2,' calling LORENTZ.')") myid
#endif /* DEBUG */

c......................................................................
c
c i boundaries
c

       nreq = 0
       nsub = nsub + 1
       call bvald  (3,1,0,0,0,0,d )

       call bvalv1 (3,3,0,0,0,0,w1)
       call bvalv2 (3,3,0,0,0,0,w2)
       call bvalv3 (3,3,0,0,0,0,w3)

#ifdef MPI_USED
c
c    3) Wait for communications to complete.
c
       if (nreq .ne. 0) call MPI_WAITALL ( nreq, req, stat, ierr )

#endif /* MPI_USED */
c......................................................................
c
c j boundaries
c
c    1) Post sends and receives. 
c
       nreq = 0
       nsub = nsub + 1

       call bvald  (0,0,3,1,0,0,d )

       call bvalv1 (0,0,3,3,0,0,w1)
       call bvalv2 (0,0,3,3,0,0,w2)
       call bvalv3 (0,0,3,3,0,0,w3)

#ifdef MPI_USED
c
c    3) Wait for communications to complete.
c
       if (nreq .ne. 0) call MPI_WAITALL ( nreq, req, stat, ierr )

#endif /* MPI_USED */
c......................................................................
c
c k boundaries
c
c    1) Post sends and receives.  
c
       nreq = 0
       nsub = nsub + 1
       call bvald  (0,0,0,0,3,1,d )
       call bvalv1 (0,0,0,0,3,3,w1)
       call bvalv2 (0,0,0,0,3,3,w2)
       call bvalv3 (0,0,0,0,3,3,w3)

#ifdef MPI_USED
c
c    3) Wait for communications to complete. 
c
       if (nreq .ne. 0) call MPI_WAITALL ( nreq, req, stat, ierr )

#endif /* MPI_USED */

c     Make single call to lorentz

       call lorentz (is,ie,js,je,ks,ke,w1,w2,w3,u1,u2,u3)

c
c      Mark the velocity boundary values out of date.
c
       do i = 1,6
c Need to reset flag for d
          bvstat(i,0) = 0       !  d
          bvstat(i,3) = 0       !  w1
          bvstat(i,4) = 0       !  w2
          bvstat(i,5) = 0       !  w3
      enddo

#endif /* not NO_FORCES */

       return
       end
c
c=======================================================================
c
c    \\\\\\\\\\        E N D   S U B R O U T I N E        //////////
c    //////////           L O R E N T Z _ D               \\\\\\\\\\
c
c=======================================================================
c
