#include "zeusmp.def"
c=======================================================================
c
c    \\\\\\\\\\      B E G I N   S U B R O U T I N E      //////////
c    //////////                 M O M X 2                 \\\\\\\\\\
c
c                            Developed by
c                Laboratory of Computational Astrophysics
c               University of Illinois at Urbana-Champaign
c
c=======================================================================
c
       subroutine momx2 (ibeg,iend,jbeg,jend,kbeg,kend
     &                  ,s1,s2,s3,mflx)
c
c    dac:zeus3d.momx2 <--------------- transports momenta in 2-direction
c    from jms:zeus2d.momx2, mln:zeus04.momr                    may, 1990
c
c    written by: David Clarke
c    modified 1: November, 1992 by David Clarke; momenta are now updated
c                between and including i=is,ie, j=js,je, and k=ks,ke to
c                allow for proper treatment of periodic boundaries.
c    modified 2: Feb. 20, 1996 by Robert Fiedler; completely rewritten
c                for ZEUS-MP.
c    modified 3: 4/11/2006 John Vernaleo Added fix taken for zeusmpv2
c
c    NOTE:  01/19/2007 jcv
c           There is a big related to loop unrolling for reflecting 
c           x2 boundaries in spherical in that was fixed in 
c           zeusmp v2.1.1  I haven't seen evidence of it, but cannot
c           haven't totally proved that it isn't present.
c
c  PURPOSE:  Transports the three components of the momentum density in
c  the 2-direction using the consistent transport algorithm, including
c  the effects of grid compression.  The transported fluxes are thus
c  given by the mass fluxes times the time centred area of the control
c  volume faces times the interpolated velocities.  Interpolations are
c  performed in-line.
c
c  INPUT VARIABLES:
c    mflx    mass flux in 2-direction; computed by TRANX2
c    s1      momentum density in 1-direction
c    s2      momentum density in 2-direction
c    s3      momentum density in 3-direction
c
c BOUNDARY VALUES USED:
c
c    Macro defined  var   ii    oi    ij    oj    ik    ok
c    -------------  ---  ----  ----  ----  ----  ----  ----
c                  mflx  is-1        js-1  je+1  ks-1
c                    u1              js-2  je+2
c                    u2  is-1        js-2  je+2  ks-1
c                    u3              js-2  je+2
c
c  OUTPUT VARIABLES:
c    s1      momentum density in 1-direction updated in the 2-direction
c    s2      momentum density in 2-direction updated in the 2-direction
c    s3      momentum density in 3-direction updated in the 2-direction
c
c  LOCAL VARIABLES:
c
c  EXTERNALS:
c
c-----------------------------------------------------------------------
c
       implicit NONE
#include "param.h"
#include "root.h"
#include "field.h"
#include "grid.h"
#include "scratch.h"
c
       integer i,j,k,ibeg,iend,jbeg,jend,kbeg,kend
c
       REAL  dqm, dqp, vel, xi, q1
c
       REAL        atwid1 (ijkn), atwid2 (ijkn), atwid3 (ijkn)
     &           , atwidj (ijkn)
#ifndef UNROLL_J
#define UNROLL_J 1
#endif /* UNROLL_J */
       REAL        sflx  (ijkn,UNROLL_J), dq   (ijkn,UNROLL_J)
#if UNROLL_J>1
       integer  iriter
#endif
c
       REAL        mflx    (  in,  jn,  kn), s1      (  in,  jn,  kn)
     1           , s2      (  in,  jn,  kn), s3      (  in,  jn,  kn)
c
       equivalence   ( atwid1  , w1da     ), ( atwid2  , w1db     )
     1             , ( atwid3  , w1dc     ), ( atwidj  , w1dd     )
#if UNROLL_J==1
       equivalence   ( sflx    , w1de     ), ( dq      , w1df     )
#endif
#if UNROLL_J==2
       equivalence   ( sflx    , w1de     ), ( dq      , w1dg     )
#endif
#if UNROLL_J==3
       equivalence   ( sflx    , w1de     ), ( dq      , w1dh     )
#endif
c.......................................................................
c
c      Compute time-centred area factors.
c
       do 10 i=ibeg,iend
         atwid1(i) = 0.5 * g31a(i)           * dx1b(i) * dvl1bi(i)
         atwid2(i) = 0.5 * g2b (i) * g31b(i) * dx1a(i) * dvl1ai(i)
         atwid3(i) = 0.5 * g31b(i) * g31b(i) * dx1a(i) * dvl1ai(i)
10     continue
c
       do 20 j=jbeg,jend+1
#ifdef MOVING_GRID
         atwidj(j) = g32a(j) * g32ah(j)
#else
         atwidj(j) = g32a(j) * g32a (j)
#endif
20     continue
c
c.......................................................................
c
       do 3000 k=kbeg,kend
#if UNROLL_J>1
c Do the iterations which will not be done in unrolled loops first.
c
         iriter = mod(iend-ibeg+1,UNROLL_J)
         do 2000 i=ibeg,ibeg+iriter-1
#else /* UNROLL_J>1 */
         do 2000 i=ibeg,iend
#endif /* UNROLL_J>1 */
c
c---------------------------- TRANSPORT S1 -----------------------------
c
           do 1070 j=jbeg-1,jend+1
c
c      Interpolate "v1" at the 2-interfaces.
c
c       call x2zc3d ( v1, vel2, ks, is, ke, ie, iords1, istps1
c     1             , g2a, g2ai, vtwid2, p      )
c
c  1.  Evaluate monotonised, van Leer difference in "q" across the zone.
c
             dqm      = (v1 (i  ,j  ,k) - v1 (i  ,j-1,k)) * dx2bi  (j  )
             dqp      = (v1 (i  ,j+1,k) - v1 (i  ,j  ,k)) * dx2bi  (j+1)
             dq(j,1)  = max ( dqm * dqp, zro )
     1                * sign ( one, dqm + dqp )
     2                / max ( abs ( dqm + dqp ), tiny )
c
1070       continue
c
c  2.  Choose time averaged, upwinded interface value.
c
           do 1100 j=jbeg,jend+1
c
c      Construct an i-average of "v2" to be used for interpolation.
c
c      Construct the 1-momentum flux at the 2-interfaces and perform
c  1-momentum advection.  Note that the timestep "dt" is hidden in the
c  mass flux.
c
              vel        = 0.5 * ( v2(i-1,j  ,k) + v2(i  ,j  ,k) )
              xi         = ( vel         - vg2(j  ) ) * dt * g2ai(i  )
              q1         = sign ( haf, xi )
              sflx (j,1) = ( 0.5 + q1 ) * ( v1 (i  ,j-1,k)
     1                   + ( dx2a(j-1) - xi ) * dq (j-1,1) )
     2                   + ( 0.5 - q1 ) * ( v1 (i  ,j  ,k)
     3                   - ( dx2a(j  ) + xi ) * dq (j  ,1) )
              sflx (j,1) = ( mflx (i-1,j  ,k) + mflx (i  ,j  ,k) )
#ifdef MOVING_GRID
     1                   * sflx (j,1)   * g32ah(j  ) * atwid1(i  )
#else
     1                   * sflx (j,1)   * g32a (j  ) * atwid1(i  )
#endif
c
1100       continue
c
           do 1170 j=jbeg,jend
             s1(i  ,j,k)  = ( s1(i  ,j,k) * dvl2a(j)
#ifdef MOVING_GRID
     1                    - sflx (j+1,1)   + sflx (j,1)   ) * dvl2ani(j)
#else
     1                    - sflx (j+1,1)   + sflx (j,1)   ) * dvl2a i(j)
#endif
c
1170       continue
c
c---------------------------- TRANSPORT S2 -----------------------------
c
c      Interpolate "v2" to the zone centers.
c
c       call x2fc3d ( v2, vel2, ks, is, ke, ie, iords2, sflx2 )
c
c  1.  Evaluate monotonised, van Leer difference in "q" across the zone.
c
           do 1270 j=jbeg-1,jend+1
             dqm        = (v2 (i  ,j  ,k) - v2 (i  ,j-1,k)) * dx2ai(j-1)
             dqp        = (v2 (i  ,j+1,k) - v2 (i  ,j  ,k)) * dx2ai(j  )
             dq(j,1)    = max ( dqm * dqp, zro )
     1                  * sign ( one, dqm + dqp )
     2                  / max ( abs ( dqm + dqp ), tiny )
c
1270       continue
c
c  2.  Choose time averaged, upwinded interface value.
c
c      Construct a j-average of "v2-vg2" to be used for interpolation.
c
c      Construct the 2-momentum flux at the zone centres and perform
c  2-momentum advection.  Note that the timestep "dt" is hidden in the
c  mass flux.
c
           do 1300 j=jbeg-1,jend
              vel        = 0.5 * ( v2(i  ,j  ,k) - vg2(j  )
     1                           + v2(i  ,j+1,k) - vg2(j+1) )
              xi         = vel         * dt * g2bi(i  )
              q1         = sign ( haf, xi )
              sflx (j,1) = ( 0.5 + q1 ) * ( v2 (i  ,j  ,k)
     1                   + ( dx2b(j  ) - xi ) * dq (j  ,1) )
     2                   + ( 0.5 - q1 ) * ( v2 (i  ,j+1,k) 
     3                   - ( dx2b(j+1) + xi ) * dq (j+1,1) )
              sflx (j,1) = ( mflx (i  ,j  ,k) + mflx (i  ,j+1,k) )
c Bugfix from zeusmp v2 atwid2(i+1) to i
#ifdef MOVING_GRID
     1                   * sflx (j,1)   * g32bh(j) * atwid2(i)
#else
     1                   * sflx (j,1)   * g32b (j) * atwid2(i)
#endif
c
1300       continue
c
           do 1370 j=jbeg,jend
             s2(i  ,j,k)  = ( s2(i  ,j,k) * dvl2b(j)
     1                    - sflx (j,1)   + sflx (j-1,1) )
#ifdef MOVING_GRID
     2                    * dvl2bni(j)
#else
     2                    * dvl2b i(j)
#endif
c
1370       continue
c
c---------------------------- TRANSPORT S3 -----------------------------
c
c      Interpolate "v3" at the 2-interfaces.
c
c       call x2zc3d ( v3, vel2, ks, is, ke, ie, iords3, istps3
c     1             , g2b, g2bi, sflx2, p      )
c
c  1.  Evaluate monotonised, van Leer difference in "q" across the zone.
c
           do 1470 j=jbeg-1,jend+1
            dqm         = (v3 (i  ,j  ,k) - v3 (i  ,j-1,k)) * dx2bi(j  )
            dqp         = (v3 (i  ,j+1,k) - v3 (i  ,j  ,k)) * dx2bi(j+1)
            dq(j,1)     = max ( dqm * dqp, zro )
     1                  * sign ( one, dqm + dqp )
     2                  / max ( abs ( dqm + dqp ), tiny )
c
1470       continue
c
c  2.  Choose time averaged, upwinded interface value.
c
           do 1500 j=jbeg,jend+1
c
c      Construct a  k-average of "v2" to be used for interpolation.
c
c      Construct the 3-momentum flux at the 2-interfaces and perform
c  3-momentum advection.  Note that the timestep "dt" is hidden in the
c  mass flux.
c
              vel        = 0.5 * ( v2(i  ,j  ,k-1) + v2(i  ,j  ,k) )
              xi         = ( vel         - vg2(j  ) ) * dt * g2bi(i  )
              q1         = sign ( haf, xi )
              sflx (j,1) = ( 0.5 + q1 ) * ( v3 (i  ,j-1,k)
     1                   + ( dx2a(j-1) - xi ) * dq (j-1,1) )
     2                   + ( 0.5 - q1 ) * ( v3 (i  ,j  ,k)
     3                   - ( dx2a(j  ) + xi ) * dq (j  ,1) )
              sflx (j,1) = ( mflx (i  ,j  ,k-1) + mflx (i  ,j  ,k  ) )
     1                   * sflx (j  ,1) * atwid3(i  ) * atwidj(j  )
c
1500       continue
c
           do 1570 j=jbeg,jend
             s3(i  ,j,k)    = ( s3(i  ,j,k) * dvl2a(j)
     1                      - sflx (j+1,1)     + sflx (j  ,1)   )
#ifdef MOVING_GRID
     2                    *  dvl2ani(j)
#else
     2                    *  dvl2a i(j)
#endif
c
1570       continue
2000     continue
#if UNROLL_J>1
c
c Do the bulk of the iterations in unrolled loops.
c
         do 2900 i=ibeg+iriter,iend,UNROLL_J
           do 2070 j=jbeg-1,jend+1
c
c---------------------------- TRANSPORT S1 -----------------------------
c
c      Interpolate "v1" at the 2-interfaces.
c
c       call x2zc3d ( v1, vel2, ks, is, ke, ie, iords1, istps1
c     1             , g2a, g2ai, vtwid2, p      )
c
c  1.  Evaluate monotonised, van Leer difference in "q" across the zone.
c
             dqm      = (v1 (i  ,j  ,k) - v1 (i  ,j-1,k)) * dx2bi  (j  )
             dqp      = (v1 (i  ,j+1,k) - v1 (i  ,j  ,k)) * dx2bi  (j+1)
             dq(j,1)  = max ( dqm * dqp, zro )
     1                * sign ( one, dqm + dqp )
     2                / max ( abs ( dqm + dqp ), tiny )
c
             dqm      = (v1 (i+1,j  ,k) - v1 (i+1,j-1,k)) * dx2bi  (j  )
             dqp      = (v1 (i+1,j+1,k) - v1 (i+1,j  ,k)) * dx2bi  (j+1)
             dq(j,2)  = max ( dqm * dqp, zro )
     1                * sign ( one, dqm + dqp )
     2                / max ( abs ( dqm + dqp ), tiny )
c
#if UNROLL_J>2
             dqm      = (v1 (i+2,j  ,k) - v1 (i+2,j-1,k)) * dx2bi  (j  )
             dqp      = (v1 (i+2,j+1,k) - v1 (i+2,j  ,k)) * dx2bi  (j+1)
             dq(j,3)  = max ( dqm * dqp, zro )
     1                * sign ( one, dqm + dqp )
     2                / max ( abs ( dqm + dqp ), tiny )
c
#endif
#if UNROLL_J>3
             dqm      = (v1 (i+3,j  ,k) - v1 (i+3,j-1,k)) * dx2bi  (j  )
             dqp      = (v1 (i+3,j+1,k) - v1 (i+3,j  ,k)) * dx2bi  (j+1)
             dq(j,4)  = max ( dqm * dqp, zro )
     1                * sign ( one, dqm + dqp )
     2                / max ( abs ( dqm + dqp ), tiny )
c
#endif
2070       continue
c
c  2.  Choose time averaged, upwinded interface value.
c
           do 2100 j=jbeg,jend+1
c
c      Construct an i-average of "v2" to be used for interpolation.
c
c      Construct the 1-momentum flux at the 2-interfaces and perform
c  1-momentum advection.  Note that the timestep "dt" is hidden in the
c  mass flux.
c
              vel        = 0.5 * ( v2(i-1,j  ,k) + v2(i  ,j  ,k) )
              xi         = ( vel         - vg2(j  ) ) * dt * g2ai(i  )
              q1         = sign ( haf, xi )
              sflx (j,1) = ( 0.5 + q1 ) * ( v1 (i  ,j-1,k)
     1                   + ( dx2a(j-1) - xi ) * dq (j-1,1) )
     2                   + ( 0.5 - q1 ) * ( v1 (i  ,j  ,k)
     3                   - ( dx2a(j  ) + xi ) * dq (j  ,1) )
              sflx (j,1) = ( mflx (i-1,j  ,k) + mflx (i  ,j  ,k) )
#ifdef MOVING_GRID
     1                   * sflx (j,1)   * g32ah(j  ) * atwid1(i  )
#else
     1                   * sflx (j,1)   * g32a (j  ) * atwid1(i  )
#endif
c
              vel        = 0.5 * ( v2(i  ,j  ,k) + v2(i+1,j  ,k) )
              xi         = ( vel         - vg2(j  ) ) * dt * g2ai(i+1)
              q1         = sign ( haf, xi )
              sflx (j,2) = ( 0.5 + q1 ) * ( v1 (i+1,j-1,k)
     1                   + ( dx2a(j-1) - xi ) * dq (j-1,2) )
     2                   + ( 0.5 - q1 ) * ( v1 (i+1,j  ,k)
     3                   - ( dx2a(j  ) + xi ) * dq (j  ,2) )
              sflx (j,2) = ( mflx (i  ,j  ,k) + mflx (i+1,j  ,k) )
#ifdef MOVING_GRID
     1                   * sflx (j,2)   * g32ah(j  ) * atwid1(i+1)
#else
     1                   * sflx (j,2)   * g32a (j  ) * atwid1(i+1)
#endif
c
#if UNROLL_J>2
              vel        = 0.5 * ( v2(i+1,j  ,k) + v2(i+2,j  ,k) )
              xi         = ( vel         - vg2(j  ) ) * dt * g2ai(i+2)
              q1         = sign ( haf, xi )
              sflx (j,3) = ( 0.5 + q1 ) * ( v1 (i+2,j-1,k)
     1                   + ( dx2a(j-1) - xi ) * dq (j-1,3) )
     2                   + ( 0.5 - q1 ) * ( v1 (i+2,j  ,k)
     3                   - ( dx2a(j  ) + xi ) * dq (j  ,3) )
              sflx (j,3) = ( mflx (i+1,j  ,k) + mflx (i+2,j  ,k) )
#ifdef MOVING_GRID
     1                   * sflx (j,3)   * g32ah(j  ) * atwid1(i+2)
#else
     1                   * sflx (j,3)   * g32a (j  ) * atwid1(i+2)
#endif
c
#endif
#if UNROLL_J>3
              vel        = 0.5 * ( v2(i+2,j  ,k) + v2(i+3,j  ,k) )
              xi         = ( vel         - vg2(j  ) ) * dt * g2ai(i+3)
              q1         = sign ( haf, xi )
              sflx (j,4) = ( 0.5 + q1 ) * ( v1 (i+3,j-1,k)
     1                   + ( dx2a(j-1) - xi ) * dq (j-1,4) )
     2                   + ( 0.5 - q1 ) * ( v1 (i+3,j  ,k)
     3                   - ( dx2a(j  ) + xi ) * dq (j  ,4) )
              sflx (j,4) = ( mflx (i+2,j  ,k) + mflx (i+3,j  ,k) )
#ifdef MOVING_GRID
     1                   * sflx (j,4)   * g32ah(j  ) * atwid1(i+3)
#else
     1                   * sflx (j,4)   * g32a (j  ) * atwid1(i+3)
#endif
c
#endif
2100       continue
c
           do 2170 j=jbeg,jend
             s1(i  ,j,k)  = ( s1(i  ,j,k) * dvl2a(j)
#ifdef MOVING_GRID
     1                    - sflx (j+1,1)   + sflx (j,1)   ) * dvl2ani(j)
#else
     1                    - sflx (j+1,1)   + sflx (j,1)   ) * dvl2a i(j)
#endif
c
             s1(i+1,j,k)  = ( s1(i+1,j,k) * dvl2a(j)
#ifdef MOVING_GRID
     1                    - sflx (j+1,2)   + sflx (j,2)   ) * dvl2ani(j)
#else
     1                    - sflx (j+1,2)   + sflx (j,2)   ) * dvl2a i(j)
#endif
c
#if UNROLL_J>2
             s1(i+2,j,k)  = ( s1(i+2,j,k) * dvl2a(j)
#ifdef MOVING_GRID
     1                    - sflx (j+1,3)   + sflx (j,3)   ) * dvl2ani(j)
#else
     1                    - sflx (j+1,3)   + sflx (j,3)   ) * dvl2a i(j)
#endif
c
#endif
#if UNROLL_J>3
             s1(i+3,j,k)  = ( s1(i+3,j,k) * dvl2a(j)
#ifdef MOVING_GRID
     1                    - sflx (j+1,4)   + sflx (j,4)   ) * dvl2ani(j)
#else
     1                    - sflx (j+1,4)   + sflx (j,4)   ) * dvl2a i(j)
#endif
c
#endif
2170       continue
c
c---------------------------- TRANSPORT S2 -----------------------------
c
c      Interpolate "v2" to the zone centers.
c
c       call x2fc3d ( v2, vel2, ks, is, ke, ie, iords2, sflx2 )
c
c  1.  Evaluate monotonised, van Leer difference in "q" across the zone.
c
           do 2270 j=jbeg-1,jend+1
             dqm        = (v2 (i  ,j  ,k) - v2 (i  ,j-1,k)) * dx2ai(j-1)
             dqp        = (v2 (i  ,j+1,k) - v2 (i  ,j  ,k)) * dx2ai(j  )
             dq(j,1)    = max ( dqm * dqp, zro )
     1                  * sign ( one, dqm + dqp )
     2                  / max ( abs ( dqm + dqp ), tiny )
c
             dqm        = (v2 (i+1,j  ,k) - v2 (i+1,j-1,k)) * dx2ai(j-1)
             dqp        = (v2 (i+1,j+1,k) - v2 (i+1,j  ,k)) * dx2ai(j  )
             dq(j,2)    = max ( dqm * dqp, zro )
     1                  * sign ( one, dqm + dqp )
     2                  / max ( abs ( dqm + dqp ), tiny )
c
#if UNROLL_J>2
             dqm        = (v2 (i+2,j  ,k) - v2 (i+2,j-1,k)) * dx2ai(j-1)
             dqp        = (v2 (i+2,j+1,k) - v2 (i+2,j  ,k)) * dx2ai(j  )
             dq(j,3)    = max ( dqm * dqp, zro )
     1                  * sign ( one, dqm + dqp )
     2                  / max ( abs ( dqm + dqp ), tiny )
c
#endif
#if UNROLL_J>3
             dqm        = (v2 (i+3,j  ,k) - v2 (i+3,j-1,k)) * dx2ai(j-1)
             dqp        = (v2 (i+3,j+1,k) - v2 (i+3,j  ,k)) * dx2ai(j  )
             dq(j,4)    = max ( dqm * dqp, zro )
     1                  * sign ( one, dqm + dqp )
     2                  / max ( abs ( dqm + dqp ), tiny )
c
#endif
2270       continue
c
c  2.  Choose time averaged, upwinded interface value.
c
c      Construct a j-average of "v2-vg2" to be used for interpolation.
c
c      Construct the 2-momentum flux at the zone centres and perform
c  2-momentum advection.  Note that the timestep "dt" is hidden in the
c  mass flux.
c
           do 2300 j=jbeg-1,jend
              vel        = 0.5 * ( v2(i  ,j  ,k) - vg2(j  )
     1                           + v2(i  ,j+1,k) - vg2(j+1) )
              xi         = vel         * dt * g2bi(i  )
              q1         = sign ( haf, xi )
              sflx (j,1) = ( 0.5 + q1 ) * ( v2 (i  ,j  ,k)
     1                   + ( dx2b(j  ) - xi ) * dq (j  ,1) )
     2                   + ( 0.5 - q1 ) * ( v2 (i  ,j+1,k) 
     3                   - ( dx2b(j+1) + xi ) * dq (j+1,1) )
              sflx (j,1) = ( mflx (i  ,j  ,k) + mflx (i  ,j+1,k) )
c Bugfix from zeusmp v2 atwid2(i+1) changed to i
#ifdef MOVING_GRID
     1                   * sflx (j,1)   * g32bh(j) * atwid2(i)
#else
     1                   * sflx (j,1)   * g32b (j) * atwid2(i)
#endif
c
              vel        = 0.5 * ( v2(i+1,j  ,k) - vg2(j  )
     1                           + v2(i+1,j+1,k) - vg2(j+1) )
              xi         = vel         * dt * g2bi(i+1)
              q1         = sign ( haf, xi )
              sflx (j,2) = ( 0.5 + q1 ) * ( v2 (i+1,j  ,k)
     1                   + ( dx2b(j  ) - xi ) * dq (j  ,2) )
     2                   + ( 0.5 - q1 ) * ( v2 (i+1,j+1,k) 
     3                   - ( dx2b(j+1) + xi ) * dq (j+1,2) )
              sflx (j,2) = ( mflx (i+1,j  ,k) + mflx (i+1,j+1,k) )
#ifdef MOVING_GRID
     1                   * sflx (j,2)   * g32bh(j) * atwid2(i+1)
#else
     1                   * sflx (j,2)   * g32b (j) * atwid2(i+1)
#endif
c
#if UNROLL_J>2
              vel        = 0.5 * ( v2(i+2,j  ,k) - vg2(j  )
     1                           + v2(i+2,j+1,k) - vg2(j+1) )
              xi         = vel         * dt * g2bi(i+2)
              q1         = sign ( haf, xi )
              sflx (j,3) = ( 0.5 + q1 ) * ( v2 (i+2,j  ,k)
     1                   + ( dx2b(j  ) - xi ) * dq (j  ,3) )
     2                   + ( 0.5 - q1 ) * ( v2 (i+2,j+1,k) 
     3                   - ( dx2b(j+1) + xi ) * dq (j+1,3) )
              sflx (j,3) = ( mflx (i+2,j  ,k) + mflx (i+2,j+1,k) )
#ifdef MOVING_GRID
     1                   * sflx (j,3)   * g32bh(j) * atwid2(i+2)
#else
     1                   * sflx (j,3)   * g32b (j) * atwid2(i+2)
#endif
c
#endif
#if UNROLL_J>3
              vel        = 0.5 * ( v2(i+3,j  ,k) - vg2(j  )
     1                           + v2(i+3,j+1,k) - vg2(j+1) )
              xi         = vel         * dt * g2bi(i+3)
              q1         = sign ( haf, xi )
              sflx (j,4) = ( 0.5 + q1 ) * ( v2 (i+3,j  ,k)
     1                   + ( dx2b(j  ) - xi ) * dq (j  ,4) )
     2                   + ( 0.5 - q1 ) * ( v2 (i+3,j+1,k) 
     3                   - ( dx2b(j+1) + xi ) * dq (j+1,4) )
              sflx (j,4) = ( mflx (i+3,j  ,k) + mflx (i+3,j+1,k) )
#ifdef MOVING_GRID
     1                   * sflx (j,4)   * g32bh(j) * atwid2(i+3)
#else
     1                   * sflx (j,4)   * g32b (j) * atwid2(i+3)
#endif
c
#endif
2300       continue
c
           do 2370 j=jbeg,jend
             s2(i  ,j,k)  = ( s2(i  ,j,k) * dvl2b(j)
     1                    - sflx (j,1)   + sflx (j-1,1) )
#ifdef MOVING_GRID
     2                    * dvl2bni(j)
#else
     2                    * dvl2b i(j)
#endif
c
             s2(i+1,j,k)  = ( s2(i+1,j,k) * dvl2b(j)
     1                    - sflx (j,2)   + sflx (j-1,2) )
#ifdef MOVING_GRID
     2                    * dvl2bni(j)
#else
     2                    * dvl2b i(j)
#endif
c
#if UNROLL_J>2
             s2(i+2,j,k)  = ( s2(i+2,j,k) * dvl2b(j)
     1                    - sflx (j,3)   + sflx (j-1,3) )
#ifdef MOVING_GRID
     2                    * dvl2bni(j)
#else
     2                    * dvl2b i(j)
#endif
c
#endif
#if UNROLL_J>3
             s2(i+3,j,k)  = ( s2(i+3,j,k) * dvl2b(j)
     1                    - sflx (j,4)   + sflx (j-1,4) )
#ifdef MOVING_GRID
     2                    * dvl2bni(j)
#else
     2                    * dvl2b i(j)
#endif
c
#endif
2370       continue
c
c---------------------------- TRANSPORT S3 -----------------------------
c
c      Interpolate "v3" at the 2-interfaces.
c
c       call x2zc3d ( v3, vel2, ks, is, ke, ie, iords3, istps3
c     1             , g2b, g2bi, sflx2, p      )
c
c  1.  Evaluate monotonised, van Leer difference in "q" across the zone.
c
           do 2470 j=jbeg-1,jend+1
            dqm         = (v3 (i  ,j  ,k) - v3 (i  ,j-1,k)) * dx2bi(j  )
            dqp         = (v3 (i  ,j+1,k) - v3 (i  ,j  ,k)) * dx2bi(j+1)
            dq(j,1)     = max ( dqm * dqp, zro )
     1                  * sign ( one, dqm + dqp )
     2                  / max ( abs ( dqm + dqp ), tiny )
c
            dqm         = (v3 (i+1,j  ,k) - v3 (i+1,j-1,k)) * dx2bi(j  )
            dqp         = (v3 (i+1,j+1,k) - v3 (i+1,j  ,k)) * dx2bi(j+1)
            dq(j,2)     = max ( dqm * dqp, zro )
     1                  * sign ( one, dqm + dqp )
     2                  / max ( abs ( dqm + dqp ), tiny )
c
#if UNROLL_J>2
            dqm         = (v3 (i+2,j  ,k) - v3 (i+2,j-1,k)) * dx2bi(j  )
            dqp         = (v3 (i+2,j+1,k) - v3 (i+2,j  ,k)) * dx2bi(j+1)
            dq(j,3)     = max ( dqm * dqp, zro )
     1                  * sign ( one, dqm + dqp )
     2                  / max ( abs ( dqm + dqp ), tiny )
c
#endif
#if UNROLL_J>3
            dqm         = (v3 (i+3,j  ,k) - v3 (i+3,j-1,k)) * dx2bi(j  )
            dqp         = (v3 (i+3,j+1,k) - v3 (i+3,j  ,k)) * dx2bi(j+1)
            dq(j,4)     = max ( dqm * dqp, zro )
     1                  * sign ( one, dqm + dqp )
     2                  / max ( abs ( dqm + dqp ), tiny )
c
#endif
2470       continue
c
c  2.  Choose time averaged, upwinded interface value.
c
           do 2500 j=jbeg,jend+1
c
c      Construct a  k-average of "v2" to be used for interpolation.
c
c      Construct the 3-momentum flux at the 2-interfaces and perform
c  3-momentum advection.  Note that the timestep "dt" is hidden in the
c  mass flux.
c
              vel        = 0.5 * ( v2(i  ,j  ,k-1) + v2(i  ,j  ,k) )
              xi         = ( vel         - vg2(j  ) ) * dt * g2bi(i  )
              q1         = sign ( haf, xi )
              sflx (j,1) = ( 0.5 + q1 ) * ( v3 (i  ,j-1,k)
     1                   + ( dx2a(j-1) - xi ) * dq (j-1,1) )
     2                   + ( 0.5 - q1 ) * ( v3 (i  ,j  ,k)
     3                   - ( dx2a(j  ) + xi ) * dq (j  ,1) )
              sflx (j,1) = ( mflx (i  ,j  ,k-1) + mflx (i  ,j  ,k  ) )
     1                   * sflx (j  ,1) * atwid3(i  ) * atwidj(j  )
c
              vel        = 0.5 * ( v2(i+1,j  ,k-1) + v2(i+1,j  ,k) )
              xi         = ( vel         - vg2(j  ) ) * dt * g2bi(i+1)
              q1         = sign ( haf, xi )
              sflx (j,2) = ( 0.5 + q1 ) * ( v3 (i+1,j-1,k)
     1                   + ( dx2a(j-1) - xi ) * dq (j-1,2) )
     2                   + ( 0.5 - q1 ) * ( v3 (i+1,j  ,k)
     3                   - ( dx2a(j  ) + xi ) * dq (j  ,2) )
              sflx (j,2) = ( mflx (i+1,j  ,k-1) + mflx (i+1,j  ,k  ) )
     1                   * sflx (j  ,2) * atwid3(i+1) * atwidj(j  )
c
#if UNROLL_J>2
              vel        = 0.5 * ( v2(i+2,j  ,k-1) + v2(i+2,j  ,k) )
              xi         = ( vel         - vg2(j  ) ) * dt * g2bi(i+2)
              q1         = sign ( haf, xi )
              sflx (j,3) = ( 0.5 + q1 ) * ( v3 (i+2,j-1,k)
     1                   + ( dx2a(j-1) - xi ) * dq (j-1,3) )
     2                   + ( 0.5 - q1 ) * ( v3 (i+2,j  ,k)
     3                   - ( dx2a(j  ) + xi ) * dq (j  ,3) )
              sflx (j,3) = ( mflx (i+2,j  ,k-1) + mflx (i+2,j  ,k  ) )
     1                   * sflx (j  ,3) * atwid3(i+2) * atwidj(j  )
c
#endif
#if UNROLL_J>3
              vel        = 0.5 * ( v2(i+3,j  ,k-1) + v2(i+3,j  ,k) )
              xi         = ( vel         - vg2(j  ) ) * dt * g2bi(i+3)
              q1         = sign ( haf, xi )
              sflx (j,4) = ( 0.5 + q1 ) * ( v3 (i+3,j-1,k)
     1                   + ( dx2a(j-1) - xi ) * dq (j-1,4) )
     2                   + ( 0.5 - q1 ) * ( v3 (i+3,j  ,k)
     3                   - ( dx2a(j  ) + xi ) * dq (j  ,4) )
              sflx (j,4) = ( mflx (i+3,j  ,k-1) + mflx (i+3,j  ,k  ) )
     1                   * sflx (j  ,4) * atwid3(i+3) * atwidj(j  )
c
#endif
2500       continue
c
           do 2570 j=jbeg,jend
             s3(i  ,j,k)    = ( s3(i  ,j,k) * dvl2a(j)
     1                      - sflx (j+1,1)     + sflx (j  ,1)   )
#ifdef MOVING_GRID
     2                    *  dvl2ani(j)
#else
     2                    *  dvl2a i(j)
#endif
c
             s3(i+1,j,k)    = ( s3(i+1,j,k) * dvl2a(j)
     1                      - sflx (j+1,2)     + sflx (j  ,2)   )
#ifdef MOVING_GRID
     2                    *  dvl2ani(j)
#else
     2                    *  dvl2a i(j)
#endif
c
#if UNROLL_J>2
             s3(i+2,j,k)    = ( s3(i+2,j,k) * dvl2a(j)
     1                      - sflx (j+1,3)     + sflx (j  ,3)   )
#ifdef MOVING_GRID
     2                    *  dvl2ani(j)
#else
     2                    *  dvl2a i(j)
#endif
c
#endif
#if UNROLL_J>3
             s3(i+3,j,k)    = ( s3(i+3,j,k) * dvl2a(j)
     1                      - sflx (j+1,4)     + sflx (j  ,4)   )
#ifdef MOVING_GRID
     2                    *  dvl2ani(j)
#else
     2                    *  dvl2a i(j)
#endif
c
#endif
2570       continue
2900     continue
c
#endif /* UNROLL_J>1 */
3000   continue
c
       return
       end
c
c=======================================================================
c
c    \\\\\\\\\\        E N D   S U B R O U T I N E        //////////
c    //////////                 M O M X 2                 \\\\\\\\\\
c
c=======================================================================
c
c
