#include "zeusmp.def"
c=======================================================================
c
c    \\\\\\\\\\      B E G I N   S U B R O U T I N E      //////////
c    //////////                  N U D T                  \\\\\\\\\\
c
c                            Developed by
c                Laboratory of Computational Astrophysics
c               University of Illinois at Urbana-Champaign
c
c=======================================================================
c
       subroutine nudt
c
c    mln:zeus3d.nudt <------------------------ mhd time step calculation
c                                                              may, 1986
c
c    written by: Mike Norman
c    modified 1: May, 1986 by David Clarke; adapted for mhd.
c    modified 2: October, 1987 by Mike Norman; reworked for covariant
c                formulation.
c    modified 3: June, 1988 by Jim Stone; incorporated into ZEUS2D.
c    modified 4: February, 1990 by David Clarke; incorporated into
c                ZEUS3D.
c    modified 5: September, 1990 by David Clarke; moved magnetic fields
c                to face-centres.
c    modified 6: March 5, 1996 by Robert Fiedler; completely rewritten
c                for ZEUS-MP.
c    modified 7: Aug. 23, 1996 by Robert Fiedler; minor change to 
c                substep counter.
c    modified 8: Dec. 19, 1996 by Robert Fiedler; added radiation 
c                diffusion time step.
c    modified 9: Feb. 14, 1997 by RAF; corrected for non-hydro and
c                lower dimensionality options.
c    modified 10: kluge for Garching T3E (search on M-MML) 26 Feb 98
c    modified 11: December, 1999 by PSLi; add in dt calculation in case
c                 of subcycle of artificial viscosity.
c    modified 12: 9/17/06 CSR tslice fix
c
c  PURPOSE:  Computes the new timestep for explicit calculations from
c  the values of the field variables updated by the source and transport
c  steps.
c
c  In explicit calculations, the timestep is given by:
c
c     dt = courno * sqrt [ min ( dtcs**2 + dtv1**2 + dtv2**2 + dtv3**2
c                              + dtal**2 + dtqq**2 + dtrd**2 ) ]
c
c  where the variable names are described below.  The timestep can be
c  reduced in size by any amount, but can be larger than the old timstep
c  by no more than a factor of 1.26.
c
c  LOCAL VARIABLES:
c
c  i-sweep
c  dr*i      inverse of distance across zone in 1-, 2-, or 3-direction
c  drimax    maximum of dr1i, dr2i, and dr3i
c  dt**i2i   square of the inverse time step of the ** physical process
c            gathered during the i-sweep.  Possible values of ** are:
c                cs = sound speed
c                v1 = fluid motion in x1 direction
c                v2 = fluid motion in x2 direction
c                v3 = fluid motion in x2 direction
c                al = Alfven speed
c                qq = artificial viscosity
c                rd = radiation diffusion
c            The first five are vectors; the next to last is a scalar
c            which has been computed in ARTIFICIALVISC (passed in root).
c            Last is a scalar computed in rad_pfl (passed in radexp).
c  dttoi2i   sum of dt**i2i (without artificial viscosity contribution)
c
c  j-sweep
c  dttoi2j   vector of maximum values of dttoi2i from each i-sweep.
c            This vector is filled during a j-sweep.
c  imaxj     i-index of each dttoi2j
c  dt**i2j   values of above at zone i=imaxj
c
c  k-sweep
c  dttoi2k   vector of maximum values of dttoi2j from each j-sweep.
c  imaxk     i-index of each dttoi2k.
c  jmaxk     j-index of each dttoi2k.
c  dt**i2j   values of above at zone i=imaxk, j=jmaxk
c
c  grand maximum inverse time
c  imin      i-index where dttoi2k is a maximum
c  jmin      j-index where dttoi2k is a maximum
c  kmin      k-index where dttoi2k is a maximum
c  dt**      time step of the ** physical process at i=imax, j=jmax,
c            and k=kmax
c  dtnew     new timestep, which is limited to be no greater than 1.26
c            times the previous timestep.
c
c  EXTERNALS:
c
c-----------------------------------------------------------------------
c
       implicit NONE
#include "mod_names.h"
#include "param.h"
#include "field.h"
#include "bndry.h"
#include "root.h"
#include "scratch.h"
#include "grid.h"
#include "mpicom.h"
#if defined RAD || defined RAD_EXP
#include "radexp.h"
#endif /* RAD || RAD_EXP */
#ifdef RAD
#include "radiation.h"
#endif /* RAD */

#ifdef PROFILING
#ifdef MPE_USED
#else
#include "clocks.h"
#endif /* MPE_USED */
#endif /* PROFILING */

c
       integer       i       ,  j      , k
     1             , imax    ,  jmax   , kmax
     &             , imin    ,  jmin   , kmin
     &             , k0      ,  k1     , k2
     &             , j0
#ifdef PROFILING
#ifdef MPE_USED
#else
     &             , nt
#endif /* MPE_USED */
#endif /* PROFILING */
c
c Removed these redundant specifications (also in newdt.F).  In portland 
c group compiler they gave a warning.  They cause the intel compiler
c to fail.  Everything seems okay without them.
c 2/04/04 jcv
c       REAL        dtcs , dtv1 , dtv2 , dtv3 , dtal , dtqq
       REAL        dtcsm, dtv1m, dtv2m, dtv3m, dtalm, dttoi2m
c       REAL        dtnew
#ifdef RAD_EXP
       REAL        dtrd
#endif /* RAD_EXP */
#ifdef PHYS_VISC
       REAL        dtvisci2, dtvisci2_max, mu
#endif /* PHYS_VISC */
#ifdef RAD_LOSS
       REAL        dtcool
       common / radc / dtcool
#endif /* RAD_LOSS */
#ifdef T_CONDUCT
       REAL        dtcond
       common / tcondr / dtcond
#endif /* T_CONDUCT */
c
c      External statements
c
       external newdt
c
c-----------------------------------------------------------------------
c
c Find the minimum time step required by the Courant condition for
c this tile.  We will first compute 1/dt**2 required by each of the
c various physical processes in the calculation, and save the maximum
c value of their sum at each zone.  In the process, we will compute
c the updated velocity.
c
#ifdef DEBUG
       write(*,"(/'NUDT: Process ',i2,' entering.')") myid
#endif /* DEBUG */
       dttoi2m = 0.0
       dtcsm   = 0.0
       dtv1m   = 0.0
       dtv2m   = 0.0
       dtv3m   = 0.0
       dtalm   = 0.0
       imin    = is
       jmin    = js
       kmin    = ks
       if (nx2z .eq. 1) then
c
c For one-dimensional problems only
c
         j0 = js
       else
         j0 = js + 1
       endif
       if (nx3z .eq. 1) then
c
c For one-dimensional or two-dimensional problems only
c
         k0 = ks
       else
         k0 = ks + 1
       endif
c
c Divide the computational volume into three equal pieces.  We must
c have at least 3 active zones in the 3-direction.
c
       k1 = int( real( ke - ks + 1 ) / 3.0 ) + ks
       k2 = int( real( ke - ks + 1 ) / 3.0 ) + k1

#ifdef PROFILING
#ifdef MPE_USED
       call MPE ...
#else
       nt = nsubstps - 1  !  Newdt always next to last substep.
       call clocks (cputime, wclock)
       wal1 = real(wclock)
       cpu1 = real(cputime)
#endif /* MPE_USED */
#endif /* PROFILING */

c......................................................................
c
c i boundaries
c
c    1) Post sends and receives. 
c
       nreq = 0
       nsub = nsub + 1
#ifdef TOTAL_ENERGY
       imax = ie - 1
       if (nx2z .eq. 1) then
c
c One-dimensional problems only
c
         jmax = js
       else
         jmax = je - 1
       endif
       kmax = ke - 1
       call bvald  (1,1,0,0,0,0,d   )
       call bvalv1 (0,1,0,0,0,0,w3da)
#else  /* TOTAL_ENERGY */
       imax = ie
       jmax = je
       kmax = ke
       call bvald  (1,0,0,0,0,0,d   )
#endif /* TOTAL_ENERGY */

#ifdef PROFILING
#ifdef MPE_USED
       call MPE ...
#else
       call clocks (cputime, wclock)
       wal0 = wal1
       cpu0 = cpu1
       wal1 = real(wclock)
       cpu1 = real(cputime)
       timers(nsubstps,1,2,2) = timers(nsubstps,1,2,2) + wal1  - wal0
       timers(nsubstps,1,2,1) = timers(nsubstps,1,2,1) + cpu1  - cpu0
#endif /* MPE_USED */
#endif /* PROFILING */

c
c    2) Do first 1/3 of the interior points.
c
       call newdt (is+1,imax,j0  ,jmax,k0  ,k1
     &                ,  imin,jmin,kmin
     &                ,  dtcsm,dtv1m,dtv2m,dtv3m,dtalm,dttoi2m)
c
c       subroutine newdt (ibeg,iend,jbeg,jend,kbeg,kend
c     &                ,  imin,jmin,kmin
c     &                ,  dtcsm,dtv1m,dtv2m,dtv3m,dtalm,dttoi2m)
c

#ifdef PROFILING
#ifdef MPE_USED
       call MPE ...
#else
       call clocks (cputime, wclock)
       wal0 = wal1
       cpu0 = cpu1
       wal1 = real(wclock)
       cpu1 = real(cputime)
       timers(nt,1,2,2) = timers(nt,1,2,2) + wal1  - wal0
       timers(nt,1,2,1) = timers(nt,1,2,1) + cpu1  - cpu0
#endif /* MPE_USED */
#endif /* PROFILING */

#ifdef MPI_USED
c
c    3) Wait for communications to complete.
c
       if (nreq .ne. 0) call MPI_WAITALL ( nreq, req, stat, ierr )

#ifdef PROFILING
#ifdef MPE_USED
       call MPE ...
#else
       call clocks (cputime, wclock)
       wal0 = wal1
       cpu0 = cpu1
       wal1 = real(wclock)
       cpu1 = real(cputime)
       timers(nt,1,1,2) = timers(nt,1,1,2) + wal1  - wal0
       timers(nt,1,1,1) = timers(nt,1,1,1) + cpu1  - cpu0
#endif /* MPE_USED */
#endif /* PROFILING */

#endif
c......................................................................
c
c j boundaries
c
c    1) Post sends and receives.
c
       nreq = 0
       nsub = nsub + 1
#ifdef TOTAL_ENERGY
       call bvald  (0,0,1,1,0,0,d   )
       call bvalv2 (0,0,0,1,0,0,w3db)
#else /* TOTAL_ENERGY */
       call bvald  (0,0,1,0,0,0,d   )
#endif /* TOTAL_ENERGY */

#ifdef PROFILING
#ifdef MPE_USED
       call MPE ...
#else
       call clocks (cputime, wclock)
       wal0 = wal1
       cpu0 = cpu1
       wal1 = real(wclock)
       cpu1 = real(cputime)
       timers(nsubstps,2,2,2) = timers(nsubstps,2,2,2) + wal1  - wal0
       timers(nsubstps,2,2,1) = timers(nsubstps,2,2,1) + cpu1  - cpu0
#endif /* MPE_USED */
#endif /* PROFILING */

c
c    2) Do middle 1/3 of the interior points, plus some on borders.
c
       call newdt (is  ,is  ,j0  ,jmax,k0  ,k1
     &                ,  imin,jmin,kmin
     &                ,  dtcsm,dtv1m,dtv2m,dtv3m,dtalm,dttoi2m)
#if defined TOTAL_ENERGY
       call newdt (ie  ,ie  ,j0  ,jmax,k0  ,k1
     &                ,  imin,jmin,kmin
     &                ,  dtcsm,dtv1m,dtv2m,dtv3m,dtalm,dttoi2m)
#endif /* TOTAL_ENERGY */
       if (nx3z .gt. 1) then
         call newdt (is  ,ie  ,j0  ,jmax,k1+1,k2
     &                  ,  imin,jmin,kmin
     &                  ,  dtcsm,dtv1m,dtv2m,dtv3m,dtalm,dttoi2m)
       endif

#ifdef PROFILING
#ifdef MPE_USED
       call MPE ...
#else
       call clocks (cputime, wclock)
       wal0 = wal1
       cpu0 = cpu1
       wal1 = real(wclock)
       cpu1 = real(cputime)
       timers(nt,2,2,2) = timers(nt,2,2,2) + wal1  - wal0
       timers(nt,2,2,1) = timers(nt,2,2,1) + cpu1  - cpu0
#endif /* MPE_USED */
#endif /* PROFILING */

#ifdef MPI_USED
c
c    3) Wait for communications to complete.
c
       if (nreq .ne. 0) call MPI_WAITALL ( nreq, req, stat, ierr )

#ifdef PROFILING
#ifdef MPE_USED
       call MPE ...
#else
       call clocks (cputime, wclock)
       wal0 = wal1
       cpu0 = cpu1
       wal1 = real(wclock)
       cpu1 = real(cputime)
       timers(nt,2,1,2) = timers(nt,2,1,2) + wal1  - wal0
       timers(nt,2,1,1) = timers(nt,2,1,1) + cpu1  - cpu0
#endif /* MPE_USED */
#endif /* PROFILING */

#endif
c......................................................................
c
c k boundaries
c
c    1) Post sends and receives.
c
       nreq = 0
       nsub = nsub + 1
#ifdef TOTAL_ENERGY
       call bvald  (0,0,0,0,1,1,d   )
       call bvalv3 (0,0,0,0,0,1,w3dc)
#else  /* TOTAL_ENERGY */
       call bvald  (0,0,0,0,1,0,d   )
#endif /* TOTAL_ENERGY */

#ifdef PROFILING
#ifdef MPE_USED
       call MPE ...
#else
       call clocks (cputime, wclock)
       wal0 = wal1
       cpu0 = cpu1
       wal1 = real(wclock)
       cpu1 = real(cputime)
       timers(nsubstps,3,2,2) = timers(nsubstps,3,2,2) + wal1  - wal0
       timers(nsubstps,3,2,1) = timers(nsubstps,3,2,1) + cpu1  - cpu0
#endif /* MPE_USED */
#endif /* PROFILING */

c
c    2) Do last 1/3 of the interior points, plus some on borders.
c
       if (nx2z .gt. 1) then
         call newdt (is  ,ie  ,js  ,js  ,k0  ,k2
     &                  ,  imin,jmin,kmin
     &                  ,  dtcsm,dtv1m,dtv2m,dtv3m,dtalm,dttoi2m)
#ifdef TOTAL_ENERGY
         call newdt (is  ,ie  ,je  ,je  ,k0  ,k2
     &                  ,  imin,jmin,kmin
     &                  ,  dtcsm,dtv1m,dtv2m,dtv3m,dtalm,dttoi2m)
#endif /* TOTAL_ENERGY */
       endif
       if (nx3z .gt. 1) then
         call newdt (is  ,ie  ,js  ,je  ,k2+1,kmax
     &                  ,  imin,jmin,kmin
     &                  ,  dtcsm,dtv1m,dtv2m,dtv3m,dtalm,dttoi2m)
       endif
c
c      Mark the velocity boundary values out of date.
c
       do 10 i = 1,6
         bvstat(i,3) = 0      !  v1
         bvstat(i,4) = 0      !  v2
         bvstat(i,5) = 0      !  v3
10     continue

#ifdef PROFILING
#ifdef MPE_USED
       call MPE ...
#else
       call clocks (cputime, wclock)
       wal0 = wal1
       cpu0 = cpu1
       wal1 = real(wclock)
       cpu1 = real(cputime)
       timers(nt,3,2,2) = timers(nt,3,2,2) + wal1  - wal0
       timers(nt,3,2,1) = timers(nt,3,2,1) + cpu1  - cpu0
#endif /* MPE_USED */
#endif /* PROFILING */

#ifdef MPI_USED
c
c    3) Wait for communications to complete. 
c
       if (nreq .ne. 0) call MPI_WAITALL ( nreq, req, stat, ierr )

#ifdef PROFILING
#ifdef MPE_USED
       call MPE ...
#else
       call clocks (cputime, wclock)
       wal0 = wal1
       cpu0 = cpu1
       wal1 = real(wclock)
       cpu1 = real(cputime)
       timers(nt,3,1,2) = timers(nt,3,1,2) + wal1  - wal0
       timers(nt,3,1,1) = timers(nt,3,1,1) + cpu1  - cpu0
#endif /* MPE_USED */
#endif /* PROFILING */

#endif
c......................................................................
c
c Finally, do the remaining border zones.
c
       if (ke .gt. ks) then
         call newdt (is  ,ie  ,js  ,je  ,ks  ,ks  
     &                  ,  imin,jmin,kmin
     &                  ,  dtcsm,dtv1m,dtv2m,dtv3m,dtalm,dttoi2m)
#ifdef TOTAL_ENERGY
         call newdt (is  ,ie  ,js  ,je  ,ke  ,ke  
     &                  ,  imin,jmin,kmin
     &                  ,  dtcsm,dtv1m,dtv2m,dtv3m,dtalm,dttoi2m)
#endif /* TOTAL_ENERGY */
       endif

#ifdef PROFILING
#ifdef MPE_USED
       call MPE ...
#else
       call clocks (cputime, wclock)
       wal0 = wal1
       cpu0 = cpu1
       wal1 = real(wclock)
       cpu1 = real(cputime)
       timers(nt,4,2,2) = timers(nt,4,2,2) + wal1  - wal0
       timers(nt,4,2,1) = timers(nt,4,2,1) + cpu1  - cpu0
#endif /* MPE_USED */
#endif /* PROFILING */
c
c     CSR - viscous timestep
c
#ifdef PHYS_VISC
       dtvisci2_max=0.0
       mu=0.01
c       namelist /visc/ mu
       do i=is,ie
          dtvisci2 = 100.0*mu**2.0/dx1a(i)**4.0
          if (dtvisci2 .gt. dtvisci2_max) dtvisci2_max=dtvisci2
       enddo
       do j=js,je
          dtvisci2 = 100.0*mu**2.0/dx2a(j)**4.0
          if (dtvisci2 .gt. dtvisci2_max) dtvisci2_max=dtvisci2
       enddo
       do k=ks,ke
          dtvisci2 = 100.0*mu**2.0/dx3a(k)**4.0
          if (dtvisci2 .gt. dtvisci2_max) dtvisci2_max=dtvisci2
       enddo
       dtvisci2 = dtvisci2_max
#endif /* PHYS_VISC */

c
c-----------------------------------------------------------------------
c
c Compute preliminary new time step.
c
#ifdef RAD_EXP
#ifdef NO_FORCES
       dtnew  = courno / ( sqrt ( dtrdi2  + dtnri2 ) + tiny )
c      write(*,"('dtrdi2, dtnri2 = ',1p2d12.4)")
c    .            dtrdi2, dtnri2
#else /* NO_FORCES */
       dtnew  = courno / ( sqrt ( dttoi2m + dtqqi2 + dtrdi2
     1        + dtnri2 ) + tiny )
c      write(*,"('dttoi2m, dtqqi2, dtrdi2 = ',1p3d12.4)")
c    .            dttoi2m, dtqqi2, dtrdi2
#endif /* NO_FORCES */
#endif /* RAD_EXP */
#ifdef RAD
#ifdef NO_FORCES
 666   format('tile, dtimrdi2 = ',i3,1pd13.5)
c      dtnew = courno / ( sqrt ( dtimrdi2 ) + tiny )
       dtnew = courno * dtimrdi
c      dtnew  = courno / ( sqrt ( dttoi2m + dtqqi2 ) + tiny )
 667   format('before ALLREDUCE, dtnew = ',1pd13.5)
#else /* NO_FORCES */
c      dtnew = courno / ( sqrt ( dttoi2m + dtqqi2 + dtimrdi2) + tiny )
       dtnew = courno / ( sqrt ( dttoi2m + dtqqi2 ) + tiny )
       dtnew = min(dtnew, courno * dtimrdi)
#endif /* NO_FORCES */
#endif /* RAD */
#ifndef RAD
#ifndef RAD_EXP
CPS
#ifdef SUB_AV
       dtnew  = courno / ( sqrt ( dttoi2m ) + tiny )
#else
       dtnew  = courno / ( sqrt (dttoi2m + dtqqi2
#ifdef PHYS_VISC
     &      + dtvisci2
#endif /* PHYS_VISC */
     &      ) + tiny )

#ifdef RAD_LOSS
       

       if(dtcool.ne.0.0)then
          dtnew = min(dtcool,dtnew)
       endif
#endif /* RAD_LOSS */

#ifdef T_CONDUCT
       if(dtcond.ne.0.0)then
          dtnew = min(dtcond,dtnew)
       endif
#endif /* T_CONDUCT */

cccccccccccccccccccccccccccc
c#ifndef PHYS_VISC
c       dtnew  = courno / ( sqrt ( dttoi2m + dtqqi2 ) + tiny )
c#endif /* NOT PHYS_VISC */
c#ifdef PHYS_VISC
c       dtnew  = courno / 
c     &      ( sqrt ( dttoi2m + dtqqi2 + dtvisci2 ) 
c     &      + tiny )
c#endif /* PHYS_VISC */
ccccccccccccccccc

#endif /* SUB_AV */
C
#endif /* NOT RAD_EXP */
#endif /* NOT RAD */

#ifdef MPI_USED
c
c Now find the smallest dtnew among all tiles, and send the result
c to all in buf_out.  We need the info for the print statement below 
c if the time step is too small, so use the MPI_MINLOC operation
c to pass values AND RANKS.
c
       buf_in(1) = dtnew
       buf_in(2) = real( myid )
#ifdef DEBUG
       write(*,"(/'NUDT: Process ',i2,' calling ALLREDUCE.')") myid
#endif /* DEBUG */
c  M-MML: on advice of Jakob Pichlmeier (Cray Munich) changed MPI_2FLOAT
c  to MPI_2DOUBLE_PRECISION...  26.2.98
       call MPI_ALLREDUCE( buf_in(1), buf_out(1), 1
     &                   , MPI_2DOUBLE_PRECISION
     &                   , MPI_MINLOC, comm3d, ierr)
       dtnew  =   buf_out(1)
#ifdef DEBUG
       write(*,"(/'NUDT: Process ',i2,' completed ALLREDUCE'
     &   ,', dtnew is ',1pe12.5)") myid, dtnew
#endif /* DEBUG */

#ifdef PROFILING
#ifdef MPE_USED
       call MPE ...
#else
       call clocks (cputime, wclock)
       wal0 = wal1
       cpu0 = cpu1
       wal1 = real(wclock)
       cpu1 = real(cputime)
       timers(nt,4,1,2) = timers(nt,4,1,2) + wal1  - wal0
       timers(nt,4,1,1) = timers(nt,4,1,1) + cpu1  - cpu0
#endif /* MPE_USED */
#endif /* PROFILING */

#endif /* MPI_USED */
#if defined RAD && defined NO_FORCES && defined NO_COUPLING
       dt     =   min ( dtnew, 1.26*dt )
       dt     =   min ( dt   , tlim/3.0D2)
#else
       dt     =   min ( dtnew, 1.26*dt )
c      dt     =   min ( dt   , tlim/3.0D2)
#endif /* RAD && NO_FORCES && NO_COUPLING */
c
c      if (dt .le. 5.0*dtmin) then
c
c  CSR - moved this "if" statement to later
c
c       if (dt .le. dtmin) then
c
c Determine which tile requires this short time step and report its
c coordinates and other info.  Convert 1/dt**2 to dt for each 
c physical process.
c


#ifdef MPI_USED
c
c CSR added the initialization of the time steps
c
         dtcs = 0.0
         dtv1 = 0.0
         dtv2 = 0.0
         dtv3 = 0.0
         dtal = 0.0
         dtqq = 0.0
c
c CSR's additions end here
c
         if ( myid .eq. int( buf_out(2) ) ) then
#endif
c
           dtcs = 1.0 / ( sqrt ( dtcsm ) + tiny )
           dtv1 = 1.0 / ( sqrt ( dtv1m ) + tiny )
           dtv2 = 1.0 / ( sqrt ( dtv2m ) + tiny )
           dtv3 = 1.0 / ( sqrt ( dtv3m ) + tiny )
           dtal = 1.0 / ( sqrt ( dtalm ) + tiny )
           dtqq = 1.0 / ( sqrt ( dtqqi2) + tiny )
#ifdef RAD_EXP
           dtrd = 1.0 / ( sqrt ( dtrdi2 ) + tiny )
#endif /* RAD_EXP */
c
c  CSR - new position for "if"
c
        if (dt .le. dtmin) then
           write (2, 2010) coords(1), coords(2), coords(3)
     &                   , imin, jmin, kmin, nhy, dt, dtmin, dtcs
     &                   , dtv1, dtv2, dtv3, dtqq, dtal
#ifdef RAD_EXP
     &                   , dtrd
#endif /* RAD_EXP */
#ifdef MPI_USED
         endif
#endif
ccc   don't bother adding to warning count if time step is too short
ccc   although it looks as though this may be a real problem...
ccc         nwarn = nwarn + 1
       endif
       return
c
c-----------------------------------------------------------------------
c----------------------- Write format statements -----------------------
c-----------------------------------------------------------------------
c
2010   format('NUDT    : **** WARNING **** Hot zone on tile (',i4,','
     &       ,i4,',',i4,')',/
     &       ,'NUDT    : at i=',i4,' j=',i4,' k=',i4,' (dt < 5*dtmin)'
     &       ,', nhy  = ',i6,',',/
     2       ,'NUDT    : dt   = ',1pe12.5,', dtmin= ',1e12.5,', dtcs = '
     3       ,1e12.5,',',/
     4       ,'NUDT    : dtv1 = ',1e12.5,', dtv2 = ',1e12.5,', dtv3 = '
     5       ,1e12.5,',',/
#ifdef RAD_EXP
     6       ,'NUDT    : dtqq = ',1e12.5,', dtal = ',1e12.5,', dtrd = '
#else /* RAD_EXP */
     6       ,'NUDT    : dtqq = ',1e12.5,', dtal = '
#endif /* RAD_EXP */
     7       ,1e12.5,'.')
c
       end
c
c=======================================================================
c
c    \\\\\\\\\\        E N D   S U B R O U T I N E        //////////
c    //////////                  N U D T                  \\\\\\\\\\
c
c=======================================================================
c
c

