#include "zeusmp.def"
c=======================================================================
c
c                            Developed by
c                Laboratory of Computational Astrophysics
c               University of Illinois at Urbana-Champaign
c
      subroutine phibv
#ifdef GRAV
c
c  Modified1: 11/24/99 by PSLi, rewritten for 3-D problem and for ZeusMP
c
c  PURPOSE: Uses a multipole expansion formula to calculate the outer
c  boundary values of the gravitational potential.  The n'th term in the
c  expansion is given by:
c            PHI(n) = Pn(u)*Mn*R**-(n+1)
c  where  Pn(u) = Legengre polynomial of order n
c         u     = direction cosine to grid point
c         R     = position  vector to grid point
c         Mn    = n'th multipole moment
c               = INTEGRAL[dVOL*d(R)*R**n*Pn(u)]
c  Up to nmax terms are added; execution terminates when
c            PHI(n)/[SUM OVER N-1(PHI(n))] < errphi
c  Here, "outer" boundaries are those far from the mass distribution.
c  Outer boundary potentals are calculated if the flags igriib,igroib,
c  igrijb,and igrojb=1 respectively.  The variable "isymm" is used to
c  denote problems with odd symmetry (isymm = 1) (no symmetry across
c  the equatorial plane).  For the even symmetry case (isymm = 0), the
c  odd moments are identically zero and thus never calculated.
c
c  EXTERNALS: [none]
c
c  LOCALS:
c  dm(i,j,k)  is mass contained in zone i,j,k
c  q(i,j,k,n) is the n'th Legendre polynomial at point i,j,k
c  qm(n)      is the n'th moment
c  sqm(n)     is the total n'th moment
c  r(i,j,k,n) is the position vector to the n'th power at i,j,k
c  r**bi(*,n) is the position vector along the ** boundary to the
c               -ve n'th power
c  u       is the direction cosine to grid point
c  Note that distances are scaled by x2b(je) [XYZ geometry] or x1b(ie)
c  [RTP geometry] to prevent floating point undeflows/overflows for the
c  higher moments.  Be aware this may limit the maximum number of
c  moments possible if the length of the position vector to any
c  boundary point is greatly different from xscale.
c-----------------------------------------------------------------------
      implicit NONE
#include "param.h"
#include "grid.h"
#include "field.h"
#include "bndry.h"
#include "root.h"
#include "scratch.h"
#include "gravity.h"
#include "mpicom.h"
      REAL     errmax
      integer  nmax,nm2,nm1,n,np1,np2
      parameter(nm2 = 1, nm1 = 2, n = 1, np1 = 2, np2 = 1)
c
      REAL  q  (in,jn,kn,2), r(in,jn,kn,2), dm(in,jn,kn)
     &     , r1, r2, r1i, r2i, u
     &     , piib(jn,kn,2), poib(jn,kn,2), pijb(in,kn,2)
     &     , pojb(in,kn,2), pikb(in,jn,2), pokb(in,jn,2)
     &     , riibi(jn,kn,2), roibi(jn,kn,2), rijbi(in,kn,2)
     &     , rojbi(in,kn,2), rikbi(in,jn,2), rokbi(in,jn,2)
c
      REAL  xscale,xscale2,fpg,tpg,rn,qm(2),sqm(2),qa,ephimx
     &     ,deltphi,errphi
      integer  i,j,k,iter,ism1,jsm1,ksm1,iep1,jep1,kep1
c
      equivalence
     &  (q(1,1,1,1),w3da), (q(1,1,1,2),w3db), (r(1,1,1,1),w3dc)
     &, (r(1,1,1,2),w3dd)
     &, (piib(1,1,1),w3de(1,1,1)), (poib(1,1,1),w3de(1,1,3))
     &, (pijb(1,1,1),w3de(1,1,5)), (pojb(1,1,1),w3de(1,1,7))
     &, (pikb(1,1,1),w3de(1,1,9)), (pokb(1,1,1),w3de(1,1,11))
     &, (riibi(1,1,1),w3df(1,1,1)), (roibi(1,1,1),w3df(1,1,3))
     &, (rijbi(1,1,1),w3df(1,1,5)), (rojbi(1,1,1),w3df(1,1,7))
     &, (rikbi(1,1,1),w3df(1,1,9)), (rokbi(1,1,1),w3df(1,1,11))
      equivalence (dm,w3dg)

c\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\\////////////////////////////////////
c=======================================================================
c  This routine uses a somewhat twisted way of referencing variables at
c  each iteration to save computations and assignments.  Thus, at each
c  iteration we calculate the n'th (even) and n+1'th (odd) terms using
c  values from the n-2 and n-1 iteration.  Note the n;n-2 are stored in
c  the same array (overwritten), and the n+1;n-1 in the same array.
c  The variable "rn" keeps track of the real n value we are working on.
c
      errmax = phibverr
#ifdef XYZ
      call MPI_Allreduce(x3b(ke+1), xscale, 1, 
     &                MPI_FLOAT, MPI_MAX, comm3d, ierr)
      xscale=1.0
      xscale2=xscale*xscale
#endif
#ifdef RTP
      call MPI_Allreduce(x1b(je+1), xscale, 1, 
     &                MPI_FLOAT, MPI_MAX, comm3d, ierr)
      xscale2=xscale*xscale
#endif
c
c  Negative tpg and fpg for the sign convention of ZeusMP.
c
      tpg = 2.0*pi*g
      fpg = 2.0*tpg
      rn  = 0.0
      qm(n  ) = 0.0
      qm(np1) = 0.0
      sqm(n  ) = 0.0
      sqm(np1) = 0.0
      ism1=is-1
      jsm1=js-1
      ksm1=ks-1
      iep1=ie+1
      jep1=je+1
      kep1=ke+1
c
c  Compute the zero order terms in the expansion (always done).  Begin
c  by getting the zeroth [and first, if needed] multipole moments. 
c  Note p0(u)=1, p1(u)=u
c
      do k=ks,ke
	do j=js,je
          do i=is,ie
#ifdef XYZ
            r2 = (x1b(i)*x1b(i) + x2b(j)*x2b(j)+x3b(k)*x3b(k))
     &           /xscale2
	    r1 = sqrt(r2)
            u = (x1b(i))/(xscale*r1)
#endif
#ifdef RTP
            r1(i) = x1b(i)/xscale
            u (i) = cos(x2b(j))
#endif
            q (i,j,k,n  ) = 1.0
            q (i,j,k,np1) = u
            dm(i,j,k    ) = g2b(i)*g31b(i)*g32b(j)*dx1a(i)*
     &                      dx2a(j)*dx3a(k)*d(i,j,k)
            r (i,j,k,n  ) = 1.0
            qm(n) = qm(n) + dm(i,j,k)
            if (isymm .eq. 1) then
               r(i,j,k,np1) = r1
               qm(np1) = qm(np1) + dm(i,j,k)*u*r1
            endif
          enddo
        enddo
      enddo
c
      if (isymm .eq. 0) then
        qm(n  ) = fpg*qm(n  )
      endif
      if (isymm .eq. 1) then
        qm(n  ) = tpg*qm(n  )
        qm(np1) = tpg*qm(np1)
      endif
c
c  Sum up qm from all processors.
c
      call MPI_Allreduce(qm,sqm,2,MPI_FLOAT,MPI_SUM,comm3d,
     &                   ierr)
c
c  Now compute the zero order terms (n=0 [and n=1 for odd symmetery])
c  along each boundary surface, starting with the inner i surface.
c
      if (niis(3) .eq. 3) then
        do k=ks,ke           
          do j=js,je
#ifdef XYZ
            r2 = (x1b(ism1)*x1b(ism1) + x2b(j)*x2b(j) + x3b(k)*x3b(k))
     &	         /xscale2
            r2i= 1.0/r2
            r1i = sqrt(r2i)
            u  = x1b(ism1)*r1i/xscale
#endif
#ifdef RTP
            r1i = xscale/x1b(ism1)
            r2i  = r1i*r1i
            u   = cos(x2b(j))
#endif
            piib (j,k,n  ) = 1.0
            piib (j,k,np1) = u
            riibi(j,k,np1) = r1i
            gpiib(j,k,1  ) = sqm(n)*r1i/xscale

            if (isymm .eq. 1) then
              riibi(j,k,np2) = r2i
              gpiib(j,k,1) = gpiib(j,k,1) + sqm(np1)*u*r2i/xscale
            endif
          enddo
        enddo
      endif
c
c  Initial solution along outer i surface
c
      if (nois(3) .eq. 3) then
        do k=ks,ke
          do j=js,je
#ifdef XYZ
            r2 = (x1b(iep1)*x1b(iep1) + x2b(j)*x2b(j) + x3b(k)*x3b(k))
     &           /xscale2
            r2i= 1.0/r2
            r1i= sqrt(r2i)
            u  = x1b(iep1)*r1i/xscale
#endif
#ifdef RTP
            r1i= xscale/x1b(iep1)
            r2i= r1i*r1i
            u  = cos(x2b(j))
#endif
            poib (j,k,n  ) = 1.0
            poib (j,k,np1) = u
            roibi(j,k,np1) = r1i
            gpoib(j,k,1  ) = sqm(n)*r1i/xscale

            if (isymm .eq. 1) then
	      roibi(j,k,np2)= r2i
              gpoib(j,k,1)   = gpoib(j,k,1) + sqm(np1)*u*r2i/xscale
            endif
	  enddo
        enddo
      endif
c
c  Initial solution along inner j surface
c
      if (nijs(3) .eq. 3) then
	do k=ks,ke
          do i=is,ie
#ifdef XYZ
            r2i = xscale2/
     &	          (x1b(i)*x1b(i) + x2b(jsm1)*x2b(jsm1) + x3b(k)*x3b(k))
            r1i = sqrt(r2i)
            u   = x1b(i)*r1i/xscale
#endif
#ifdef RTP
            r1i = xscale/x1b(i)
            r2i = r1i*r1i
            u   = cos(x2b(jsm1))
#endif
            pijb (i,k,n  ) = 1.0
            pijb (i,k,np1) = u
            rijbi(i,k,np1) = r1i
            gpijb(i,k,1  ) = sqm(n)*r1i/xscale

            if (isymm .eq. 1) then
              rijbi(i,k,np2)= r2i
              gpijb(i,k,1)   = gpijb(i,k,1) + sqm(np1)*u*r2i/xscale
            endif
          enddo
        enddo
      endif
c
c  Initial solution along outer j surface
c
      if (nojs(3) .eq. 3) then
	do k=ks,ke
          do i=is,ie
#ifdef XYZ
            r2i = xscale2
     &            /(x1b(i)*x1b(i) + x2b(jep1)*x2b(jep1) + x3b(k)*x3b(k))
            r1i = sqrt(r2i)
            u  = x1b(i)*r1i/xscale
#endif
#ifdef RTP
            r1i = xscale/x1b(i)
            r2i = r1i*r1i
            u   = cos(x2b(jep1))
#endif
            pojb (i,k,n  ) = 1.0
            pojb (i,k,np1) = u
            rojbi(i,k,np1) = r1i
            gpojb(i,k,1  ) = sqm(n)*r1i/xscale

            if (isymm .eq. 1) then
              rojbi(i,k,np2)= r2i
              gpojb(i,k,1)   = gpojb(i,k,1) + sqm(np1)*u*r2i/xscale
            endif
	  enddo
	enddo
      endif
c
c  Initial solution along inner k surface
c
      if (niks(3) .eq. 3) then
        do j=js,je
          do i=is,ie
#ifdef XYZ
            r2i= xscale
     &           /(x1b(i)*x1b(i) + x2b(j)*x2b(j) + x3b(ksm1)*x3b(ksm1))
            r1i= sqrt(r2i)
            u  = x1b(i)*r1i/xscale
#endif
#ifdef RTP
            r1i = xscale/x1b(i)
            r2i = r1i*r1i
            u   = cos(x2b(j))
#endif
            pikb (i,j,n  ) = 1.0
            pikb (i,j,np1) = u
            rikbi(i,j,np1) = r1i
            gpikb(i,j,1  ) = sqm(n)*r1i/xscale

            if (isymm .eq. 1) then
              rikbi(i,j,np2)= r2i
              gpikb(i,j,1)   = gpikb(i,j,1) + sqm(np1)*u*r2i/xscale
            endif
          enddo
        enddo
      endif
c
c  Initial solution along outer k surface
c
      if (noks(3) .eq. 3) then
	do j=js,je
          do i=is,ie
#ifdef XYZ
            r2i= xscale2
     &           /(x1b(i)*x1b(i) + x2b(j)*x2b(j) + x3b(kep1)*x3b(kep1))
            r1i= sqrt(r2i)
            u  = x1b(i)*r1i/xscale
#endif
#ifdef RTP
            r1i= xscale/x1b(i)
            r2i= r1i*r1i
            u  = cos(x2b(j))
#endif
            pokb (i,j,n  ) = 1.0
            pokb (i,j,np1) = u
            rokbi(i,j,np1) = r1i
            gpokb(i,j,  1) = sqm(n)*r1i/xscale

            if (isymm .eq. 1) then
              rokbi(i,j,np2)= r2i
              gpokb(i,j,1) = gpokb(i,j,1) + sqm(np1)*u*r2i/xscale
            endif
          enddo
        enddo
      endif
c
c  Begin iterations up to nmax terms.  Once the maximum error over all
c  the boundary grid points (ephimx) drops below errphi, then we break.
c  Start by calculating the multipole moment for this iteration.  A
c  recursion relation is used to generate the succesive Legendre polys:
c          Pn(u) = [(2n-1)*u*Pn-1(u) - (n-1)Pn-2(u)]/n
c
      if (pnmax .le. 2) goto 2000
      do 1000 iter=2,pnmax
        ephimx  = 0.0
        qm(n  ) = 0.0
        qm(np1) = 0.0
        rn = rn + 2.0
	do k=ks,ke
          do j=js,je
            do i=is,ie
#ifdef XYZ
              r2 = (x1b(i)*x1b(i) + x2b(j)*x2b(j) + x3b(k)*x3b(k))
     &             /xscale2
              u  = x1b(i)/(xscale*sqrt(r2))
#endif
#ifdef RTP
              r2 = x1b(i)*x1b(i)/xscale2
              u  = cos(x2b(j))
#endif
              q(i,j,k,n  ) = ((2.0*rn - 1.0)*q(i,j,k,nm1) *u
     &                      - (rn - 1.0)*q(i,j,k,nm2))/ rn 
              q(i,j,k,np1) = ((2.0*rn + 1.0)*q(i,j,k,n  ) *u
     &                      -  rn       *q(i,j,k,nm1))/(rn + 1.0)
              r(i,j,k,n  ) = r(i,j,k,nm2)*r2
              qm(n)        = qm(n) + dm(i,j,k)*r(i,j,k,n)*q(i,j,k,n)

              if (isymm .eq. 1) then
                r(i,j,k,np1) = r(i,j,k,nm1)*r2
                qm(np1) = qm(np1) + dm(i,j,k)*r(i,j,k,np1)*q(i,j,k,np1)
              endif
            enddo
          enddo
        enddo
c
        if (isymm .eq. 0) then
          qm(n  ) = fpg*qm(n  )
        endif
        if (isymm .eq. 1) then
          qm(n  ) = tpg*qm(n  )
          qm(np1) = tpg*qm(np1)
        endif
c
c  Sum up qm from all processors.
c
      call MPI_Allreduce(qm,sqm,2,MPI_FLOAT,MPI_SUM,comm3d,
     &                   ierr)
c
c  Calculate gpiib along the inner i surface
c
        if (niis(3) .eq. 3) then
          do k=ks,ke
            do j=js,je
#ifdef XYZ
              r2i=xscale2/
     &            (x1b(ism1)*x1b(ism1) + x2b(j)*x2b(j) + x3b(k)*x3b(k))
              u  =x1b(ism1)*sqrt(r2i)/xscale
#endif
#ifdef RTP
              r2i = scale2/(x1b(ism1)*x1b(ism1))
              u  = cos(x2b(j))
#endif
              piib (j,k,n  ) = ((2.0*rn - 1.0)*piib(j,k,nm1) *u
     &                       - (    rn - 1.0)*piib(j,k,nm2))/ rn
              piib (j,k,np1) = ((2.0*rn + 1.0)*piib(j,k,n  ) *u
     &                       -      rn       *piib(j,k,nm1))/(rn + 1.0) 
              riibi(j,k,np1) = riibi(j,k,nm1)*r2i
              deltphi = sqm(n)*piib(j,k,n)*riibi(j,k,np1)/xscale
              errphi  = abs(deltphi/gpiib(j,k,1))
              if (errphi .gt. ephimx) ephimx = errphi
              gpiib(j,k,1) = gpiib(j,k,1) + deltphi
 
              if (isymm .eq. 1) then
                riibi(j,k,np2) = riibi(j,k,n)*r2i
                deltphi = sqm(np1)*piib(j,k,np1)*riibi(j,k,np2)/xscale
                errphi  = abs(deltphi/gpiib(j,k,1))
                if (errphi .gt. ephimx) ephimx = errphi
                gpiib(j,k,1) = gpiib(j,k,1) + deltphi
              endif
            enddo
	  enddo
        endif
c
c  Calculate gpoib along the outer i boundary
c
        if (nois(3) .eq. 3) then
	  do k=ks,ke
            do j=js,je
#ifdef XYZ
              r2i = xscale2/(x1b(iep1)*x1b(iep1) + x2b(j)*x2b(j) 
     &            + x2b(j)*x2b(j) + x3b(k)*x3b(k))
              u   = x1b(iep1)*sqrt(r2i)/xscale
#endif
#ifdef RTP
              r2i = xscale2/(x1b(iep1)*x1b(iep1))
              u   = cos(x2b(j))
#endif
              poib (j,k,n  ) = ((2.0*rn - 1.0)*poib(j,k,nm1) *u
     &                       - (     rn - 1.0)*poib(j,k,nm2))/ rn
              poib (j,k,np1) = ((2.0*rn + 1.0)*poib(j,k,n  ) *u
     &                       -       rn       *poib(j,k,nm1))/(rn + 1.0) 
              roibi(j,k,np1) = roibi(j,k,nm1)*r2i
              deltphi = sqm(n)*poib(j,k,n)*roibi(j,k,np1)/xscale
              errphi  = abs(deltphi/gpoib(j,k,1))
              if (errphi .gt. ephimx) ephimx = errphi
              gpoib(j,k,1) = gpoib(j,k,1) + deltphi

              if (isymm .eq. 1) then
                roibi(j,k,np2) = roibi(j,k,n)*r2i
                deltphi = sqm(np1)*poib(j,k,np1)*roibi(j,k,np2)/xscale
                errphi  = abs(deltphi/gpoib(j,k,1))
                if (errphi .gt. ephimx) ephimx = errphi
                gpoib(j,k,1) = gpoib(j,k,1) + deltphi
              endif
            enddo
	  enddo
        endif
c
c  Calculate gpijb along the inner j surface
c
        if (nijs(3) .eq. 3) then
	  do k=ks,ke
            do i=is,ie
#ifdef XYZ
              r2i= xscale2/
     &             (x1b(i)*x1b(i) + x2b(jsm1)*x2b(jsm1) + x3b(k)*x3b(k))
              u  = x1b(i)*sqrt(r2i)/xscale
#endif
#ifdef RTP
              r2i= xscale2/(x1b(i)*x1b(i))
              u  = cos(x2b(jsm1))
#endif
              pijb (i,k,n  ) = ((2.0*rn - 1.0)*pijb(i,k,nm1) *u
     &                       - (    rn - 1.0)*pijb(i,k,nm2))/ rn
              pijb (i,k,np1) = ((2.0*rn + 1.0)*pijb(i,k,n  ) *u
     &                       - (    rn      )*pijb(i,k,nm1))/(rn + 1.0)
              rijbi(i,k,np1) = rijbi(i,k,nm1)*r2i
              deltphi = sqm(n)*pijb(i,k,n)*rijbi(i,k,np1)/xscale
              errphi  = abs(deltphi/gpijb(i,k,1))
              if (errphi .gt. ephimx) ephimx = errphi
              gpijb(i,k,1) = gpijb(i,k,1) + deltphi

              if (isymm .eq. 1) then
                rijbi(i,k,np2) = rijbi(i,k,n)*r2i
                deltphi = sqm(np1)*pijb(i,k,np1)*rijbi(i,k,np2)/xscale
                errphi  = abs(deltphi/gpijb(i,k,1))
                if (errphi .gt. ephimx) ephimx = errphi
                gpijb(i,k,1) = gpijb(i,k,1) + deltphi
              endif
	    enddo
	  enddo
        endif
c
c  Calculate gpojb along the outer j surface
c
        if (nojs(3) .eq. 3) then
	  do k=ks,ke
            do i=is,ie
#ifdef XYZ
              r2i= xscale2/(x1b(i)*x1b(i) + x2b(jep1)*x2b(jep1)
     &           + x3b(k)*x3b(k))
              u  = x1b(i)*sqrt(r2i)/xscale
#endif
#ifdef RTP
              r2i= xscale2/(x1bi(i)*x1bi(i))
              u  = cos(x2b(je))
#endif
              pojb (i,k,n  ) = ((2.0*rn - 1.0)*pojb(i,k,nm1)*u
     &                       - (    rn - 1.0)*pojb(i,k,nm2))/ rn
              pojb (i,k,np1) = ((2.0*rn + 1.0)*pojb(i,k,n  )*u
     &                       -      rn       *pojb(i,k,nm1))/(rn + 1.0)
              rojbi(i,k,np1) = rojbi(i,k,nm1)*r2i
              deltphi = sqm(n)*pojb(i,k,n)*rojbi(i,k,np1)/xscale
              errphi  = abs(deltphi/gpojb(i,k,1))
              if (errphi .gt. ephimx) ephimx = errphi
              gpojb(i,k,1) = gpojb(i,k,1) + deltphi
              if (isymm .eq. 1) then
                rojbi(i,k,np2) = rojbi(i,k,n)*r2i
                deltphi = sqm(np1)*pojb(i,k,np1)*rojbi(i,k,np2)/xscale
                errphi  = abs(deltphi/gpojb(i,k,1))
                if (errphi .gt. ephimx) ephimx = errphi
                gpojb(i,k,1) = gpojb(i,k,1) + deltphi
              endif
	    enddo
	  enddo
        endif
c
c  Calculate gpikb along the inner k surface
c
        if (niks(3) .eq. 3) then
	  do j=js,je
            do i=is,ie
#ifdef XYZ
              r2i= xscale2
     &		   /(x1b(i)*x1b(i) + x2b(j)*x2b(j) + x3b(ksm1)*x3b(ksm1))
              u  = x1b(i)*sqrt(r2i)/xscale
#endif
#ifdef RTP
              r2i= xscale2/(x1b(i)*x1b(i))
              u  = cos(x2b(j))
#endif
              pijb (i,j,n  ) = ((2.0*rn - 1.0)*pijb(i,j,nm1) *u
     &                       - (    rn - 1.0)*pijb(i,j,nm2))/ rn
              pijb (i,j,np1) = ((2.0*rn + 1.0)*pijb(i,j,n  ) *u
     &                       - (    rn      )*pijb(i,j,nm1))/(rn + 1.0)
              rijbi(i,j,np1) = rijbi(i,j,nm1)*r2i
              deltphi = sqm(n)*pijb(i,j,n)*rijbi(i,j,np1)/xscale
              errphi  = abs(deltphi/gpikb(i,j,1))
              if (errphi .gt. ephimx) ephimx = errphi
              gpikb(i,j,1) = gpikb(i,j,1) + deltphi

              if (isymm .eq. 1) then
                rijbi(i,j,np2) = rijbi(i,j,n)*r2i
                deltphi = sqm(np1)*pijb(i,j,np1)*rijbi(i,j,np2)/xscale
                errphi  = abs(deltphi/gpikb(i,j,1))
                if (errphi .gt. ephimx) ephimx = errphi
                gpikb(i,j,1) = gpikb(i,j,1) + deltphi
              endif
	    enddo
	  enddo
        endif
c
c  Calculate gpokb along the outer k surface
c
        if (noks(3) .eq. 3) then
	  do j=js,je
            do i=is,ie
#ifdef XYZ
              r2i= xscale2/(x1b(i)*x1b(i) + x2b(j)*x2b(j) 
     &             + x3b(kep1)*x3b(kep1))
              u  = x1b(i)*sqrt(r2i)/xscale
#endif
#ifdef RTP
              r2i= xscale2/(x1bi(i)*x1bi(i))
              u  = cos(x2b(j))
#endif
              pojb (i,j,n  ) = ((2.0*rn - 1.0)*pojb(i,j,nm1)*u
     &                       - (    rn - 1.0)*pojb(i,j,nm2))/ rn
              pojb (i,j,np1) = ((2.0*rn + 1.0)*pojb(i,j,n  )*u
     &                       -      rn       *pojb(i,j,nm1))/(rn + 1.0)
              rojbi(i,j,np1) = rojbi(i,j,nm1)*r2i
              deltphi = sqm(n)*pojb(i,j,n)*rojbi(i,j,np1)/xscale
              errphi  = abs(deltphi/gpokb(i,j,1))
              if (errphi .gt. ephimx) ephimx = errphi
              gpokb(i,j,1) = gpokb(i,j,1) + deltphi
              if (isymm .eq. 1) then
                rojbi(i,j,np2) = rojbi(i,j,n)*r2i
                deltphi = sqm(np1)*pojb(i,j,np1)*rojbi(i,j,np2)/xscale
                errphi  = abs(deltphi/gpokb(i,j,1))
                if (errphi .gt. ephimx) ephimx = errphi
                gpokb(i,j,1) = gpokb(i,j,1) + deltphi
              endif
	    enddo
	  enddo
        endif
c
c  Break if error criterion satisfied
        if (ephimx .le. errmax) goto 2000
1000  continue
c
c  If this point is reached, then boundary values did not drop below
c  errphi within nmax iterations; issue warning and continue.
c
C      write(2,"(/1x,'WARNING from PHIBV: error in gravitational ',
C     & 'potential boundary value',/1x,'exceeds limit; max err = ',
C     & 1pe12.5)") ephimx
C      nwarn = nwarn + 1 
C      return
c
2000  continue
C      write(2,"(1x,'nhy=',i5,' PHIBV: boundary pots converged in ',i3,
C     & ' iterations, eps=',1pe12.5)") nhy,iter,ephimx
#endif
      return
      end
