#include "zeusmp.def"
c
c=======================================================================
c
c    \\\\\\\\\\      B E G I N   S U B R O U T I N E      //////////
c    //////////              P R E S S U R E              \\\\\\\\\\
c
c                            Developed by
c                Laboratory of Computational Astrophysics
c               University of Illinois at Urbana-Champaign
c
c=======================================================================
c
       subroutine pressure ( p, iprs )
c
c    dac:zeus3d.pressure <---------------------------- computes pressure
c    from mln:zeus3d.pgas                                    march, 1991
c
c    written by: David Clarke
c    modified 1: June 1992, by David Clarke; added the total energy
c                option originally designed by Byung-IL Jun.
c    modified 2: May 1999, by Fabian Heitsch; adapted for use in
c                ZEUS-MP
c
c  PURPOSE: Calculates the gas pressure (thermal, magnetic, or both) at
c  each zone centred point on the grid.  If gam(1) < 0, then the
c  pressure is set to 0 (allows for pressure free (dust) calculations).
c
c  INPUT VARIABLES:
c    iprs     =1 => thermal pressure only
c             =2 => magnetic pressure only
c             =3 => total (thermal plus magnetic) pressure
cFH taken out 99/05/16
c    iall     =0 => compute block bounded by "ism2"   to "iep2"  ,
c                                            "jsm2"   to "jep2"  ,
c                                            "ksm2"   to "kep2"
c             =1 => compute block bounded by "ismnm2" to "iemxp2",
c                                            "jsmnm2" to "jemxp2",
c                                            "ksmnm2" to "kemxp2"
c
c  OUTPUT VARIABLES:
c    p        array of pressure.
c
c  LOCAL VARIABLES:
c
c  EXTERNALS: [NONE]
c
c-----------------------------------------------------------------------
c
#include "root.h"
#include "param.h"
#include "field.h"
#include "grid.h"

#ifdef MPI_USED
#include "mpicom.h"
#endif /* MPI_USED */

       integer       iprs    , iall
       integer       i       , imin    , imax
     1             , j       , jmin    , jmax
     2             , k       , kmin    , kmax
       REAL        q1      , q2      , q3
c
       REAL        p       (  in,  jn,  kn)
c
c-----------------------------------------------------------------------
#ifdef DBGZMP_PRESSURE
       if (idebug .gt. 0) then
         write(*,"('PRESSURE: Process ',i2,' entering')") myid
         call flush(flstat, 101)
       endif
#endif /* DGBZMP_PRESSURE */
c      if (iall .eq. 1) then
c         kmin = ksmnm2
c         kmax = kemxp2
c         jmin = jsmnm2
c         jmax = jemxp2
c         imin = ismnm2
c         imax = iemxp2
c       else
         kmin = ks-2
         kmax = ke+2
         jmin = js-2
         jmax = je+2

         imin = is-2
         imax = ie+2
c       endif
       q1 =   max ( gamma, zro )
c
#ifdef TOTAL_ENERGY

#ifndef ISO
c
c-----------------------------------------------------------------------
c---------------------- For total energy equation ----------------------
c-----------------------------------------------------------------------
c
         k = ks
         j = js
         i = is
#ifndef KSYM
         do k=kmin,kmax
#endif /* !KSYM */
#ifndef JSYM
           do j=jmin,jmax
#endif /* !JSYM*/
#ifndef ISYM
             do i=imin,imax
#endif /* !ISYM */
               p(i,j,k) = e(i,j,k) + efloor
     1                  - ( ( v1(i,j,k) + v1(i+1,j  ,k  ) )**2
     2                    + ( v2(i,j,k) + v2(i  ,j+1,k  ) )**2
     3                    + ( v3(i,j,k) + v3(i  ,j  ,k+1) )**2 ) * 0.125
     4                  * d(i,j,k)
#if defined GRAV || defined GRAV_FFT
     5                  - gp(i,j,k) * d(i,j,k)
#endif /* GRAV || GRAV_FFT */
               p(i,j,k) = q1 * p(i,j,k)
             enddo
           enddo
         enddo

#endif /* !ISO */
#else /* TOTAL_ENERGY */
c
c-----------------------------------------------------------------------
c-------------------- For internal energy equation ---------------------
c-----------------------------------------------------------------------
c
c      Contribution from the internal energy per unit volume (e).
c
         if ( (iprs .eq. 1) .or. (iprs .eq. 3) ) then
           k = ks
           j = js
           i = is
#ifndef KSYM
           do k=kmin,kmax
#endif /* !KSYM */
#ifndef JSYM
             do j=jmin,jmax
#endif /* !JSYM */
#ifndef ISYM
               do i=imin,imax
#endif /* !ISYM */
#ifdef ISO
                 p(i,j,k) = ciso**2 * ( d(i,j,k) + dfloor )
#else /* ISO */
                 p(i,j,k) = q1 * ( e(i,j,k) + efloor )
#endif /* ISO */
               enddo
             enddo
#ifndef KSYM
           enddo
#endif /*!KSYM */
         endif
#ifdef MHD
c
c      Contribution from the magnetic pressure.
c
         if ( (iprs .eq. 2) .or. (iprs .eq. 3) ) then
           q2 =  real( 3 - iprs ) * q1
           q3 =  real( iprs - 2 )
           k  = ks
           j  = js
           i  = is
#ifndef KSYM
           do k=kmin,kmax
#endif /* !KSYM */
#ifndef JSYM
             do j=jmin,jmax
#endif /* !JSYM */
#ifndef ISYM
               do i=imin,imax
#endif /* !ISYM */
                 p(i,j,k) = q2 * efloor + q3 * p(i,j,k)
     1                    + ( ( b1(i,j,k) + b1(i+1,j  ,k  ) )**2
     2                      + ( b2(i,j,k) + b2(i  ,j+1,k  ) )**2
     3                      + ( b3(i,j,k) + b3(i  ,j  ,k+1) )**2 )
     4                    * 0.125
               enddo
             enddo
           enddo
         endif
#endif /* MHD */

#endif /* TOTAL_ENERGY */

#ifdef DBGZMP_PRESSURE
       if (idebug .gt. 0) then
         write(*,"('PRESSURE: Process ',i2,' finished.')") myid
         call flush(flstat, 101)
       endif
#endif /* DGBZMP_PRESSURE */

       return
       end
c
c=======================================================================
c
c    \\\\\\\\\\        E N D   S U B R O U T I N E        //////////
c    //////////              P R E S S U R E              \\\\\\\\\\
c
c=======================================================================
c


