#include "zeusmp.def"
c=======================================================================
c
c    \\\\\\\\\\      B E G I N   S U B R O U T I N E      //////////
c    //////////                T R A N X 1                \\\\\\\\\\
c
c                            Developed by
c                Laboratory of Computational Astrophysics
c               University of Illinois at Urbana-Champaign
c
c=======================================================================
c
       subroutine tranx1 (ibeg,iend,jbeg,jend,kbeg,kend
     &                   ,dlo,den
#ifndef ISO
     &                   ,eod,edn
#endif
#if defined RAD || defined RAD_EXP
     &                   ,ero,ern
#endif /* RAD || RAD_EXP */
     &                   ,mflx)
c
c    dac:zeus3d.tranx1 <----- transports zone-centred variables along x1
c    from jms:zeus2d.tranx1, mln:zeus04.tranz                  may, 1990
c
c    written by: David Clarke
c    modified 1: June 1992, by David Clarke; added the total energy
c                option originally designed by Byung-IL Jun.
c    modified 2: Feb. 20, 1996 by Robert Fiedler; completely rewritten
c                for ZEUS-MP.
c    modified 3: Dec. 19, 1996 by Robert Fiedler; added radiation
c
c  PURPOSE:  Transports all zone centred variables in the 1-direction
c  only.  Currently transported are:
c
c                      mass   density
c                      energy density
c
c  The consistent advection algorithm, in which mass fluxes are used to
c  construct the fluxes of all variables across the interfaces, is used
c  for all hydrodynamical variables.  Thus, the mass fluxes are passed
c  to MOMX1 on order to transport the momenta as well.  The magnetic
c  field components are treated separately from the mass fluxes in CT.
c  Interpolations are done in-line.
c
c  INPUT VARIABLES: 
c    ibeg,iend,jbeg,jend,kbeg,kend  index ranges to cover.
c    dlo      mass            density at previous substep.
c    eod      specific energy density at previous substep; equals
c             (e+p)/d  if TOTAL_ENERGY is defined.
c
c BOUNDARY VALUES USED:
c
c    Macro defined  var   ii    oi    ij    oj    ik    ok
c    -------------  ---  ----  ----  ----  ----  ----  ----
c                    d   is-3  ie+2  js-1        ks-1
c                   e/d  is-2  ie+2
c                    u1  is-1  ie+1  js-1        ks-1
c    TOTAL_ENERGY    u2                    je+1
c    TOTAL_ENERGY    u3                                ke+1
c
c  OUTPUT VARIABLES:
c    den      updated mass            density.
c    edn      updated specific energy density.
c    mflx     mass flux (in the 1-direction)
c
c  LOCAL VARIABLES:
c    atwid    effective cross sectional area of the 1-interfaces
c    etwid    interpolated specific energy densities (e/d) at all
c             1-interfaces
c    eflx     energy density flux across all 1-interfaces  (reuse etwid)
c    dtwid    interpolated mass densities at all 1-interfaces
c    dflx     mass density flux across all 1-interfaces    (reuse dtwid)
c
c  EXTERNALS:
c
c-----------------------------------------------------------------------
c
       implicit NONE
#include "param.h"
#include "root.h"
#include "field.h"
#include "grid.h"
#include "scratch.h"
#include "bndry.h"
#include "mpicom.h"
c
       integer i, j, k, ibeg,iend,jbeg,jend,kbeg,kend
c
       REAL    dqm,dqp, xi,q1
       REAL        atwid (ijkn),  mflux (ijkn)
     &           , dtwid (ijkn),  dd    (ijkn)
#ifndef ISO
     &           , etwid (ijkn),  deod  (ijkn)
#endif
#if defined RAD || defined RAD_EXP
     &           , rtwid (ijkn),  dero  (ijkn)
#endif /* RAD || RAD_EXP */
c
       REAL        mflx    (  in,  jn,  kn)
       REAL  dlo(in,jn,kn), den(in,jn,kn)
#ifndef ISO
     &      ,eod(in,jn,kn), edn(in,jn,kn)
#endif
#if defined RAD || defined RAD_EXP
     &      ,ero(in,jn,kn), ern(in,jn,kn)
#endif /* RAD || RAD_EXP */
c
       equivalence (atwid , w1da), (mflux , w1db)
     &           , (dtwid , w1dc), (dd    , w1dd)
#ifndef ISO
     &           , (etwid , w1de), (deod  , w1df)
#endif
#ifndef UNROLL_I
#define UNROLL_I
#endif
c
c-----------------------------------------------------------------------
c
c Compute time-centered area factors.
c
CDIR$ UNROLL (UNROLL_I)
       do 10 i=ibeg-1,iend+1
#ifdef MOVING_GRID
         atwid (i)  =       g2ah(i) * g31ah(i)
#else
         atwid (i)  =       g2a (i) * g31a (i)
#endif
10     continue
c
c Transport all zone-centered quantities in the 1 direction.
c Note that transporting v1 in MOMX1 will require the mass flux at 
c x1a(is-1).  To get it from the field variables, we need d at is-3.
c We also need mflx at js-1 and ks-1 for i=is,ie+1 for v2 and v3.
c Extend loop indices to compute mass fluxes beyond inner borders.
c Be careful to assign values to mflx, (and den, edn) only
c within the range (ibeg:iend,jbeg:jend,kbeg:kend) when these
c indices are not on the borders, so that they can't get overwritten
c when this routine is called with various index ranges.
c
         do 100 k=kbeg-1,kend
           do 90 j=jbeg-1,jend
c
c   Interpolate to obtain zone-centered quantities at zone faces.
c
c     1.  Evaluate monotonised, van Leer differences across the zone.
c
             if (ibeg .eq. is) then  !  Need d(is-3) from neighbor.
               i         = is - 2
               dqm       = (dlo(i  ,j,k) - diib (j,k,3)) * dx1bi(i  )
               dqp       = (dlo(i+1,j,k) - dlo(i  ,j,k)) * dx1bi(i+1)
               dd(i)     = max ( dqm * dqp, zro )
     1                   * sign ( one, dqm + dqp )
     2                   / max ( abs ( dqm + dqp ), tiny )
#ifndef ISO
               deod  (i) = zro  ! Not valid, but we don't use it.
#endif /* not ISO  */
#if defined RAD || defined RAD_EXP
               dero  (i) = zro  ! Not valid, but we don't use it.
#endif /* RAD || RAD_EXP */
             endif
CDIR$ UNROLL (UNROLL_I)
             do 30 i=max(ibeg-2,is-1),iend+1
               dqm       = (dlo(i  ,j,k) - dlo(i-1,j,k)) * dx1bi(i  )
               dqp       = (dlo(i+1,j,k) - dlo(i  ,j,k)) * dx1bi(i+1)
               dd(i)     = max ( dqm * dqp, zro )
     1                   * sign ( one, dqm + dqp )
     2                   / max ( abs ( dqm + dqp ), tiny )
#ifndef ISO
               dqm       = (eod(i  ,j,k) - eod(i-1,j,k)) * dx1bi(i  )
               dqp       = (eod(i+1,j,k) - eod(i  ,j,k)) * dx1bi(i+1)
               deod(i)   = max ( dqm * dqp, zro )
     1                   * sign ( one, dqm + dqp )
     2                   / max ( abs ( dqm + dqp ), tiny )
#endif /* not ISO  */
#if defined RAD || defined RAD_EXP
               dqm       = (ero(i  ,j,k) - ero(i-1,j,k)) * dx1bi(i  )
               dqp       = (ero(i+1,j,k) - ero(i  ,j,k)) * dx1bi(i+1)
               dero(i)   = max ( dqm * dqp, zro )
     1                   * sign ( one, dqm + dqp )
     2                   / max ( abs ( dqm + dqp ), tiny )
#endif /* RAD || RAD_EXP */
30           continue
c
c     2.  Choose time averaged, upwinded interface values.
c
CDIR$ UNROLL (UNROLL_I)
             do 40 i=ibeg-1,iend+1
               xi        = ( v1  (i,j,k) - vg1(i) ) * dt
               q1        = sign ( haf, xi )
               dtwid (i) = ( 0.5 + q1 ) * ( dlo(i-1,j,k)
     1                   + ( dx1a(i-1) - xi ) * dd   (i-1) )
     2                   + ( 0.5 - q1 ) * ( dlo(i  ,j,k)
     3                   - ( dx1a(i  ) + xi ) * dd   (i  ) )
#ifndef ISO
               etwid (i) = ( 0.5 + q1 ) * ( eod(i-1,j,k)
     1                   + ( dx1a(i-1) - xi ) * deod (i-1) )
     2                   + ( 0.5 - q1 ) * ( eod(i  ,j,k)
     3                   - ( dx1a(i  ) + xi ) * deod (i  ) )
#endif /* not ISO */
#if defined RAD || defined RAD_EXP
               rtwid (i) = ( 0.5 + q1 ) * ( ero(i-1,j,k)
     1                   + ( dx1a(i-1) - xi ) * dero (i-1) )
     2                   + ( 0.5 - q1 ) * ( ero(i  ,j,k)
     3                   - ( dx1a(i  ) + xi ) * dero (i  ) )
#endif /* RAD || RAD_EXP */
40           continue
c
c  For the purposes of consistent advection, construct the mass
c  flux across each 1-interface.  The mass flux will be used to create
c  the fluxes of all variables, including the momenta which are updated
c  in MOMX1.
c
CDIR$ UNROLL (UNROLL_I)
           do 50 i=ibeg-1,iend+1
             mflux (i    ) = dtwid (i    ) * ( v1(i,j,k) - vg1(i) ) * dt
             dtwid (i    ) = mflux (i    ) * atwid (i    )
#ifndef ISO
             etwid (i    ) = dtwid (i    ) * etwid (i    )
#endif /* not ISO */
#if defined RAD || defined RAD_EXP
             rtwid (i    ) = dtwid (i    ) * rtwid (i    )
#endif /* RAD || RAD_EXP */
50         continue
c
c  Save the mass flux outside (ibeg:iend,jbeg:jend,kbeg:kend)
c  only for zones next to the inner borders.
c
           if ( (j.eq.js-1 .and. k.ge.kbeg)   .or.
     &          (j.ge.jbeg .and. k.eq.ks-1)   .or.
     &          (j.eq.js-1 .and. k.eq.ks-1) ) then
             if (ibeg.eq.is) mflx(is-1,j,k) = mflux (is-1)
CDIR$ UNROLL (UNROLL_I)
             do 60 i=ibeg,iend
               mflx(i,j,k) = mflux (i      )
60           continue
             if (iend.eq.ie) mflx(ie+1,j,k) = mflux (ie+1)
           endif
c
c  Perform mass density and energy density advection.  Note that
c  the timestep "dt" is hidden the fluxes "dflx" and "eflx".
c  Do only zones inside (ibeg:iend,jbeg:jend,kbeg:kend).
c
           if (j.ge.jbeg .and. k.ge.kbeg) then
             if (ibeg.eq.is) mflx(is-1,j,k) = mflux (is-1)
CDIR$ UNROLL (UNROLL_I)
             do 80 i=ibeg,iend
             mflx(i,j,k)= mflux (i      )
             den(i,j,k) = ( dlo(i,j,k) * dvl1a(i)
#ifdef MOVING_GRID
     1                  - dtwid (i+1    ) + dtwid (i    ) ) * dvl1ani(i)
#else
     1                  - dtwid (i+1    ) + dtwid (i    ) ) * dvl1a i(i)
#endif /* MOVING_GRID */
#ifndef ISO
             e  (i,j,k) = ( e  (i,j,k) * dvl1a(i)
#ifdef MOVING_GRID
     1                  - etwid (i+1    ) + etwid (i    ) ) * dvl1ani(i)
#else
     1                  - etwid (i+1    ) + etwid (i    ) ) * dvl1a i(i)
#endif /* MOVING_GRID */
c
c Compute e/d for the next substep.
c
#ifndef TOTAL_ENERGY
             edn(i,j,k) = e(i,j,k) / den(i,j,k)
#else /* not TOTAL_ENERGY */
             edn(i,j,k) = gamma * e(i,j,k) / den(i,j,k)
     1                  - gamm1 * ( ( v1(i,j,k) + v1(i+1,j  ,k  ) )**2
     2                            + ( v2(i,j,k) + v2(i  ,j+1,k  ) )**2
     3                            + ( v3(i,j,k) + v3(i  ,j  ,k+1) )**2 )
     4                          * 0.125
#if defined GRAV || defined GRAV_FFT
     5                  - gamm1 * gp(i,j,k)
#endif /* GRAV || GRAV_FFT */
#endif /* not TOTAL_ENERGY */
#endif /* not ISO */
#if defined RAD || defined RAD_EXP
             ern(i,j,k) = ( ero(i,j,k) * dlo(i,j,k) * dvl1a(i)
#ifdef MOVING_GRID
     1                  - rtwid (i+1    ) + rtwid (i    ) ) * dvl1ani(i)
#else
     1                  - rtwid (i+1    ) + rtwid (i    ) ) * dvl1a i(i)
#endif /* MOVING_GRID */
c
c Work with er/d for the next sweep.
c
             ern(i,j,k) = ern(i,j,k) / den(i,j,k)
#endif /* RAD || RAD_EXP */
80         continue
           if (iend.eq.ie) mflx(ie+1,j,k) = mflux (ie+1)
         endif
90       continue
100    continue
c
       return
       end
c
c=======================================================================
c
c    \\\\\\\\\\        E N D   S U B R O U T I N E        //////////
c    //////////                T R A N X 1                \\\\\\\\\\
c
c=======================================================================
c
c
