#include "zeusmp.def"
c=======================================================================
c
c    \\\\\\\\\\      B E G I N   S U B R O U T I N E      //////////
c    //////////                T R A N X 2                \\\\\\\\\\
c
c                            Developed by
c                Laboratory of Computational Astrophysics
c               University of Illinois at Urbana-Champaign
c
c=======================================================================
c
       subroutine tranx2 (ibeg,iend,jbeg,jend,kbeg,kend
     &                   ,dlo,den
#ifndef ISO
     &                   ,eod,edn
#endif
#if defined RAD || defined RAD_EXP
     &                   ,ero,ern
#endif /* RAD || RAD_EXP */
     &                   ,mflx)
c
c    dac:zeus3d.tranx2 <----- transports zone-centred variables along x2
c    from jms:zeus2d.tranx2, mln:zeus04.tranz                  may, 1990
c
c    written by: David Clarke
c    modified 1: June 1992, by David Clarke; added the total energy
c                option originally designed by Byung-IL Jun.
c    modified 2: Feb. 20, 1996 by Robert Fiedler; completely rewritten
c                for ZEUS-MP.
c    modified 3: Aug. 8, 1996 by Robert Fiedler; unrolled i-loops.
c    modified 4: Dec. 23, 1996 by Robert Fiedler; added radiation.
c    modified 5: Jul. 15, 2005 by John C. Vernaleo geometry fix 
c                         from John Hayes
c
c  PURPOSE:  Transports all zone centred variables in the 2-direction
c  only.  Currently transported are:
c
c                      mass   density
c                      energy density
c
c  The consistent advection algorithm, in which mass fluxes are used to
c  construct the fluxes of all variables across the interfaces, is used
c  for all hydrodynamical variables.  Thus, the mass fluxes are passed
c  to MOMX2 on order to transport the momenta as well.  The magnetic
c  field components are treated separately from the mass fluxes in CT.
c  Interpolations are done in-line.
c
c  INPUT VARIABLES: 
c    ibeg,iend,jbeg,jend,kbeg,kend  index ranges to cover.
c    dlo      mass            density at previous substep.
c    eod      specific energy density at previous substep; equals
c             (e+p)/d  if TOTAL_ENERGY is defined.
c    ero      specific radiation energy density at previous substep.
c
c BOUNDARY VALUES USED:
c
c    Macro defined  var   ii    oi    ij    oj    ik    ok
c    -------------  ---  ----  ----  ----  ----  ----  ----
c                    d   is-1        js-3  je+2  ks-1
c                   e/d              js-2  je+2
c    TOTAL_ENERGY    u1        ie+1                         
c                    u2  is-1        js-1  je+1  ks-1
c    TOTAL_ENERGY    u3                                ke+1
c
c  OUTPUT VARIABLES:
c    den      updated mass            density.
c    edn      updated specific energy density.
c    ern      updated specific radiation energy density.
c    mflx     mass flux (in the 2-direction)
c
c  LOCAL VARIABLES:
c    atwid    effective cross sectional area of the 1-interfaces
c    etwid    interpolated specific energy densities (e/d) at all
c             2-interfaces
c    rtwid    interpolated specific radiation energy densities at all
c             2-interfaces
c    eflx     energy density flux across all 2-interfaces  (reuse etwid)
c    dtwid    interpolated mass densities at all 2-interfaces
c    dflx     mass density flux across all 2-interfaces    (reuse dtwid)
c
c  EXTERNALS:
c
c-----------------------------------------------------------------------
c
       implicit NONE
#include "param.h"
#include "root.h"
#include "field.h"
#include "grid.h"
#include "scratch.h"
#include "bndry.h"
#include "mpicom.h"
c
       integer i, j, k, ibeg,iend,jbeg,jend,kbeg,kend
c
       REAL    dqm,dqp, xi,q1
       REAL        atwid (ijkn)
       equivalence (atwid , w1da)
c
#ifndef UNROLL_J
#define UNROLL_J 1
#endif
#if UNROLL_J>1
       integer     iriter
#endif
       REAL        mflux (ijkn,UNROLL_J)
     &           , dtwid (ijkn,UNROLL_J),  dd    (ijkn,UNROLL_J)
#if UNROLL_J==1
       equivalence (mflux , w1db)
     &           , (dtwid , w1dc), (dd    , w1dd)
#endif
#if UNROLL_J==2
       equivalence (mflux , w1db)
     &           , (dtwid , w1dd), (dd    , w1df)
#endif
#ifndef ISO
       REAL        etwid (ijkn,UNROLL_J),  deod  (ijkn,UNROLL_J)
#if UNROLL_J==1
       equivalence (etwid , w1de), (deod  , w1df)
#endif
#if UNROLL_J==2
       equivalence (etwid , w1dh)
#endif
#endif
c
       REAL  dlo  (in,jn,kn), den  (in,jn,kn), mflx (in,jn,kn)
#ifndef ISO
     &      ,eod  (in,jn,kn), edn  (in,jn,kn)
#endif
#if defined RAD || defined RAD_EXP
     &      ,ero  (in,jn,kn), ern  (in,jn,kn)
       REAL        rtwid(ijkn,UNROLL_J),  dero  (ijkn,UNROLL_J)
#endif /* RAD || RAD_EXP */
c
c-----------------------------------------------------------------------
c
c Compute time-centered area factors.
c
       do 10 i=ibeg-1,iend+1
         atwid (i)         = g31b(i) * dx1a(i) * dvl1ai(i)
10     continue
c
c Transport all zone-centered quantities in the 2 direction.
c
       do 2100 k=kbeg-1,kend
         if ( (k.ge.kbeg) .or. (k.eq.ks-1) ) then
c.......................................................................
c
c Split off the i=ibeg-1 iteration to ease unrolling.
c
           i = ibeg - 1
c
c   Interpolate to obtain zone-centered quantities at zone faces.
c
c     1.  Evaluate monotonised, van Leer differences across the zone.
c
           if (jbeg .eq. js) then  !  Need d(js-3) from neighbor.
             j         = js - 2
             dqm       = (dlo(i,j  ,k) - dijb (i,k,3)) * dx2bi(j  )
             dqp       = (dlo(i,j+1,k) - dlo(i,j  ,k)) * dx2bi(j+1)
             dd  (j,1) = max ( dqm * dqp, zro )
     1                 * sign ( one, dqm + dqp )
     2                 / max ( abs ( dqm + dqp ), tiny )
#ifndef ISO
             deod(j,1) = zro  ! Not valid, but we don't use it.
#endif
#if defined RAD || defined RAD_EXP
             dero(j,1) = zro  ! Not valid, but we don't use it.
#endif /* RAD || RAD_EXP */
           endif
           do 30 j=max(jbeg-2,js-1),jend+1
             dqm       = (dlo(i,j  ,k) - dlo(i,j-1,k)) * dx2bi(j  )
             dqp       = (dlo(i,j+1,k) - dlo(i,j  ,k)) * dx2bi(j+1)
             dd  (j,1) = max ( dqm * dqp, zro )
     1                 * sign ( one, dqm + dqp )
     2                 / max ( abs ( dqm + dqp ), tiny )
#ifndef ISO
             dqm       = (eod(i,j  ,k) - eod(i,j-1,k)) * dx2bi(j  )
             dqp       = (eod(i,j+1,k) - eod(i,j  ,k)) * dx2bi(j+1)
             deod(j,1) = max ( dqm * dqp, zro )
     1                 * sign ( one, dqm + dqp )
     2                 / max ( abs ( dqm + dqp ), tiny )
#endif /* not ISO  */
#if defined RAD || defined RAD_EXP
             dqm       = (ero(i,j  ,k) - ero(i,j-1,k)) * dx2bi(j  )
             dqp       = (ero(i,j+1,k) - ero(i,j  ,k)) * dx2bi(j+1)
             dero(j,1) = max ( dqm * dqp, zro )
     1                 * sign ( one, dqm + dqp )
     2                 / max ( abs ( dqm + dqp ), tiny )
#endif /* RAD || RAD_EXP */
30         continue
c
c     2.  Choose time averaged, upwinded interface value.
c
c  For the purposes of consistent advection, construct the mass
c  flux across each 1-interface.  The mass flux will be used to create
c  the fluxes of all variables, including the momenta which are updated
c  in MOMX1.
c
             do 40 j=jbeg-1,jend+1
               xi         = ( v2  (i,j  ,k) - vg2(j  ) ) * dt * g2bi(i)
               q1         = sign ( haf, xi )
               dtwid (j,1)= ( 0.5 + q1 ) * ( dlo(i,j-1,k)
     1                    + ( dx2a(j-1) - xi ) * dd   (j-1,1) )
     2                    + ( 0.5 - q1 ) * ( dlo(i,j  ,k)
     3                    - ( dx2a(j  ) + xi ) * dd   (j  ,1) )
c
             mflux (j,1) = dtwid (j,1) * ( v2(i,j  ,k) - vg2(j  ) ) * dt
#ifdef MOVING_GRID
             dtwid (j,1) = mflux (j,1) * atwid (i) * g32ah(j  )
#else
             dtwid (j,1) = mflux (j,1) * atwid (i) * g32a (j  )
#endif /* MOVING_GRID */
#ifndef ISO
               etwid (j,1)= ( 0.5 + q1 ) * ( eod(i,j-1,k)
     1                    + ( dx2a(j-1) - xi ) * deod (j-1,1) )
     2                    + ( 0.5 - q1 ) * ( eod(i,j  ,k)
     3                    - ( dx2a(j  ) + xi ) * deod (j  ,1) )
c
               etwid (j,1)= dtwid (j,1) * etwid (j,1)
#endif /* not ISO */
#if defined RAD || defined RAD_EXP
               rtwid (j,1)= ( 0.5 + q1 ) * ( ero(i,j-1,k)
     1                    + ( dx2a(j-1) - xi ) * dero (j-1,1) )
     2                    + ( 0.5 - q1 ) * ( ero(i,j  ,k)
     3                    - ( dx2a(j  ) + xi ) * dero (j  ,1) )
c
               rtwid (j,1)= dtwid (j,1) * rtwid (j,1)
#endif /* RAD || RAD_EXP */
40           continue
c
c  Save the mass flux outside (ibeg:iend,jbeg:jend,kbeg:kend)
c  only for zones next to the inner orders.
c
               if (i.eq.is-1) then

                 if (jbeg .eq. js) mflx(i,js-1,k) = mflux (js-1,1)
                 do 60 j=jbeg,jend
                   mflx(i,j,k) = mflux(j   ,1)
60               continue
                 if (jend .eq. je) mflx(i,je+1,k) = mflux (je+1,1)
c
               endif
c.......................................................................
c
#if UNROLL_J>1
c Do the iterations which will not be done in unrolled loops next.
c
           iriter = mod(iend-ibeg+1,UNROLL_J)
           do 1090 i=ibeg,ibeg+iriter-1
#else /* UNROLL_J>1 */
           do 1090 i=ibeg,iend
#endif /* UNROLL_J>1 */
c
c   Interpolate to obtain zone-centered quantities at zone faces.
c
c     1.  Evaluate monotonised, van Leer differences across the zone.
c
           if (jbeg .eq. js) then  !  Need d(js-3) from neighbor.
             j         = js - 2
             dqm       = (dlo(i,j  ,k) - dijb (i,k,3)) * dx2bi(j  )
             dqp       = (dlo(i,j+1,k) - dlo(i,j  ,k)) * dx2bi(j+1)
             dd  (j,1) = max ( dqm * dqp, zro )
     1                 * sign ( one, dqm + dqp )
     2                 / max ( abs ( dqm + dqp ), tiny )
#ifndef ISO
             deod(j,1) = zro  ! Not valid, but we don't use it.
#endif /* not ISO */
#if defined RAD || defined RAD_EXP
             dero(j,1) = zro  ! Not valid, but we don't use it.
#endif /* RAD || RAD_EXP */
           endif
           do 1030 j=max(jbeg-2,js-1),jend+1
             dqm       = (dlo(i,j  ,k) - dlo(i,j-1,k)) * dx2bi(j  )
             dqp       = (dlo(i,j+1,k) - dlo(i,j  ,k)) * dx2bi(j+1)
             dd  (j,1) = max ( dqm * dqp, zro )
     1                 * sign ( one, dqm + dqp )
     2                 / max ( abs ( dqm + dqp ), tiny )
#ifndef ISO
             dqm       = (eod(i,j  ,k) - eod(i,j-1,k)) * dx2bi(j  )
             dqp       = (eod(i,j+1,k) - eod(i,j  ,k)) * dx2bi(j+1)
             deod(j,1) = max ( dqm * dqp, zro )
     1                 * sign ( one, dqm + dqp )
     2                 / max ( abs ( dqm + dqp ), tiny )
#endif /* not ISO  */
#if defined RAD || defined RAD_EXP
             dqm       = (ero(i,j  ,k) - ero(i,j-1,k)) * dx2bi(j  )
             dqp       = (ero(i,j+1,k) - ero(i,j  ,k)) * dx2bi(j+1)
             dero(j,1) = max ( dqm * dqp, zro )
     1                 * sign ( one, dqm + dqp )
     2                 / max ( abs ( dqm + dqp ), tiny )
#endif /* RAD || RAD_EXP */
1030       continue
c
c     2.  Choose time averaged, upwinded interface value.
c
c  For the purposes of consistent advection, construct the mass
c  flux across each 1-interface.  The mass flux will be used to create
c  the fluxes of all variables, including the momenta which are updated
c  in MOMX1.
c
             do 1040 j=jbeg-1,jend+1
               xi         = ( v2  (i,j  ,k) - vg2(j  ) ) * dt
               q1         = sign ( haf, xi )
               dtwid (j,1)= ( 0.5 + q1 ) * ( dlo(i,j-1,k)
     1                    + ( dx2a(j-1) - xi ) * dd   (j-1,1) )
     2                    + ( 0.5 - q1 ) * ( dlo(i,j  ,k)
     3                    - ( dx2a(j  ) + xi ) * dd   (j  ,1) )
c
             mflux (j,1) = dtwid (j,1) * ( v2(i,j  ,k) - vg2(j  ) ) * dt
#ifdef MOVING_GRID
             dtwid (j,1) = mflux (j,1) * atwid (i) * g32ah(j  )
#else
             dtwid (j,1) = mflux (j,1) * atwid (i) * g32a (j  )
#endif /* MOVING_GRID */
#ifndef ISO
               etwid (j,1)= ( 0.5 + q1 ) * ( eod(i,j-1,k)
     1                    + ( dx2a(j-1) - xi ) * deod (j-1,1) )
     2                    + ( 0.5 - q1 ) * ( eod(i,j  ,k)
     3                    - ( dx2a(j  ) + xi ) * deod (j  ,1) )
c
               etwid (j,1)= dtwid (j,1) * etwid (j,1)
#endif /* not ISO */
#if defined RAD || defined RAD_EXP
               rtwid (j,1)= ( 0.5 + q1 ) * ( ero(i,j-1,k)
     1                    + ( dx2a(j-1) - xi ) * dero (j-1,1) )
     2                    + ( 0.5 - q1 ) * ( ero(i,j  ,k)
     3                    - ( dx2a(j  ) + xi ) * dero (j  ,1) )
c
               rtwid (j,1)= dtwid (j,1) * rtwid (j,1)
#endif /* RAD || RAD_EXP */
1040         continue
c
c  Save the mass flux outside (ibeg:iend,jbeg:jend,kbeg:kend)
c  only for zones next to the inner orders.
c
               if ( (i.eq.is-1 .and. k.ge.kbeg) .or.
     &              (i.ge.ibeg .and. k.eq.ks-1) .or.
     &              (i.eq.is-1 .and. k.eq.ks-1) ) then
                 if (jbeg .eq. js) mflx(i,js-1,k) = mflux (js-1,1)
                 do 1060 j=jbeg,jend
                   mflx(i,j,k) = mflux(j   ,1)
1060             continue
                 if (jend .eq. je) mflx(i,je+1,k) = mflux (je+1,1)
               endif
c
c  Perform mass density and energy density advection.  Note that
c  the timestep "dt" is hidden the fluxes "dflx" and "eflx".
c
             if (k.ge.kbeg .and. i.ge.ibeg) then
               if (jbeg .eq. js) mflx(i,js-1,k) = mflux (js-1,1)
               if (jend .eq. je) mflx(i,je+1,k) = mflux (je+1,1)
               do 1080 j=jbeg,jend
               mflx(i,j,k) = mflux (j,1)
               den(i,j,k) = ( dlo(i,j,k) * dvl2a(j)
#ifdef MOVING_GRID
     1                    - dtwid (j+1,1)  + dtwid (j,1)  ) * dvl2ani(j)
#else
     1                    - dtwid (j+1,1)  + dtwid (j,1)  ) * dvl2a i(j)
#endif /* MOVING_GRID */
#ifndef ISO
               e  (i,j,k) = ( e  (i,j,k) * dvl2a(j)
#ifdef MOVING_GRID
     1                    - etwid (j+1,1)  + etwid (j,1)  ) * dvl2ani(j)
#else
     1                    - etwid (j+1,1)  + etwid (j,1)  ) * dvl2a i(j)
#endif /* MOVING_GRID */
c
c Compute e/d for the next substep.
c
#ifndef TOTAL_ENERGY
               edn(i,j,k) = e(i,j,k) / den(i,j,k)
#else /* not TOTAL_ENERGY */
               edn(i,j,k) = gamma * e(i,j,k) / den(i,j,k)
     1                    - gamm1 * ( ( v1(i,j,k) + v1(i+1,j  ,k  ) )**2
     2                              + ( v2(i,j,k) + v2(i  ,j+1,k  ) )**2
     3                              + ( v3(i,j,k) + v3(i  ,j  ,k+1) )**2
     4                              ) * 0.125
#if defined GRAV || defined GRAV_FFT 
     5                    - gamm1 * gp(i,j,k)
#endif /* GRAV || GRAV_FFT */
#endif /* not TOTAL_ENERGY */
#endif /* not ISO */
#if defined RAD || defined RAD_EXP
               ern(i,j,k) = ( ero(i,j,k) * dlo(i,j,k) * dvl2a(j)
#ifdef MOVING_GRID
     1                    - rtwid (j+1,1)  + rtwid (j,1)  ) * dvl2ani(j)
#else
     1                    - rtwid (j+1,1)  + rtwid (j,1)  ) * dvl2a i(j)
#endif /* MOVING_GRID */
c
c Work with er/d for the next sweep.
c
               ern(i,j,k) = ern(i,j,k) / den(i,j,k)
#endif /* RAD || RAD_EXP */
1080         continue
           endif
1090       continue
c
c.......................................................................
c
#if UNROLL_J>1
c Do the bulk of the iterations in unrolled loops.
c
           do 2090 i=ibeg+iriter,iend,UNROLL_J
c
c
c   Interpolate to obtain zone-centered quantities at zone faces.
c
c     1.  Evaluate monotonised, van Leer differences across the zone.
c
           if (jbeg .eq. js) then  !  Need d(js-3) from neighbor.
             j         = js - 2
             dqm       = (dlo(i  ,j  ,k) - dijb (i  ,k,3)) * dx2bi(j  )
             dqp       = (dlo(i  ,j+1,k) - dlo(i  ,j  ,k)) * dx2bi(j+1)
             dd  (j,1) = max ( dqm * dqp, zro )
     1                 * sign ( one, dqm + dqp )
     2                 / max ( abs ( dqm + dqp ), tiny )
#ifndef ISO
             deod(j,1) = zro  ! Not valid, but we don't use it.
#endif
#if defined RAD || defined RAD_EXP
             dero(j,1) = zro  ! Not valid, but we don't use it.
#endif /* RAD || RAD_EXP */
c
             dqm       = (dlo(i+1,j  ,k) - dijb (i+1,k,3)) * dx2bi(j  )
             dqp       = (dlo(i+1,j+1,k) - dlo(i+1,j  ,k)) * dx2bi(j+1)
             dd  (j,2) = max ( dqm * dqp, zro )
     1                 * sign ( one, dqm + dqp )
     2                 / max ( abs ( dqm + dqp ), tiny )
#ifndef ISO
             deod(j,2) = zro  ! Not valid, but we don't use it.
#endif
#if defined RAD || defined RAD_EXP
             dero(j,2) = zro  ! Not valid, but we don't use it.
#endif /* RAD || RAD_EXP */
c
#if UNROLL_J>2
             dqm       = (dlo(i  ,j  ,k) - dijb (i  ,k,3)) * dx2bi(j  )
             dqp       = (dlo(i+2,j+1,k) - dlo(i+2,j  ,k)) * dx2bi(j+1)
             dd  (j,3) = max ( dqm * dqp, zro )
     1                 * sign ( one, dqm + dqp )
     2                 / max ( abs ( dqm + dqp ), tiny )
#ifndef ISO
             deod(j,3) = zro  ! Not valid, but we don't use it.
#endif /* ISO */
#if defined RAD || defined RAD_EXP
             dero(j,3) = zro  ! Not valid, but we don't use it.
#endif /* RAD || RAD_EXP */
c
#endif
#if UNROLL_J>3
             dqm       = (dlo(i+3,j  ,k) - dijb (i+3,k,3)) * dx2bi(j  )
             dqp       = (dlo(i+3,j+1,k) - dlo(i+3,j  ,k)) * dx2bi(j+1)
             dd  (j,4) = max ( dqm * dqp, zro )
     1                 * sign ( one, dqm + dqp )
     2                 / max ( abs ( dqm + dqp ), tiny )
#ifndef ISO
             deod(j,4) = zro  ! Not valid, but we don't use it.
#endif
#if defined RAD || defined RAD_EXP
             dero(j,4) = zro  ! Not valid, but we don't use it.
#endif /* RAD || RAD_EXP */
c
#endif
           endif
           do 2030 j=max(jbeg-2,js-1),jend+1
             dqm       = (dlo(i  ,j  ,k) - dlo(i  ,j-1,k)) * dx2bi(j  )
             dqp       = (dlo(i  ,j+1,k) - dlo(i  ,j  ,k)) * dx2bi(j+1)
             dd  (j,1) = max ( dqm * dqp, zro )
     1                 * sign ( one, dqm + dqp )
     2                 / max ( abs ( dqm + dqp ), tiny )
#ifndef ISO
             dqm       = (eod(i  ,j  ,k) - eod(i  ,j-1,k)) * dx2bi(j  )
             dqp       = (eod(i  ,j+1,k) - eod(i  ,j  ,k)) * dx2bi(j+1)
             deod(j,1) = max ( dqm * dqp, zro )
     1                 * sign ( one, dqm + dqp )
     2                 / max ( abs ( dqm + dqp ), tiny )
#endif /* not ISO  */
#if defined RAD || defined RAD_EXP
             dqm       = (ero(i  ,j  ,k) - ero(i  ,j-1,k)) * dx2bi(j  )
             dqp       = (ero(i  ,j+1,k) - ero(i  ,j  ,k)) * dx2bi(j+1)
             dero(j,1) = max ( dqm * dqp, zro )
     1                 * sign ( one, dqm + dqp )
     2                 / max ( abs ( dqm + dqp ), tiny )
#endif /* RAD || RAD_EXP */
c
             dqm       = (dlo(i+1,j  ,k) - dlo(i+1,j-1,k)) * dx2bi(j  )
             dqp       = (dlo(i+1,j+1,k) - dlo(i+1,j  ,k)) * dx2bi(j+1)
             dd  (j,2) = max ( dqm * dqp, zro )
     1                 * sign ( one, dqm + dqp )
     2                 / max ( abs ( dqm + dqp ), tiny )
#ifndef ISO
             dqm       = (eod(i+1,j  ,k) - eod(i+1,j-1,k)) * dx2bi(j  )
             dqp       = (eod(i+1,j+1,k) - eod(i+1,j  ,k)) * dx2bi(j+1)
             deod(j,2) = max ( dqm * dqp, zro )
     1                 * sign ( one, dqm + dqp )
     2                 / max ( abs ( dqm + dqp ), tiny )
#endif /* not ISO  */
#if defined RAD || defined RAD_EXP
             dqm       = (ero(i+1,j  ,k) - ero(i+1,j-1,k)) * dx2bi(j  )
             dqp       = (ero(i+1,j+1,k) - ero(i+1,j  ,k)) * dx2bi(j+1)
             dero(j,2) = max ( dqm * dqp, zro )
     1                 * sign ( one, dqm + dqp )
     2                 / max ( abs ( dqm + dqp ), tiny )
#endif /* RAD || RAD_EXP */
c
#if UNROLL_J>2
             dqm       = (dlo(i+2,j  ,k) - dlo(i+2,j-1,k)) * dx2bi(j  )
             dqp       = (dlo(i+2,j+1,k) - dlo(i+2,j  ,k)) * dx2bi(j+1)
             dd  (j,3) = max ( dqm * dqp, zro )
     1                 * sign ( one, dqm + dqp )
     2                 / max ( abs ( dqm + dqp ), tiny )
#ifndef ISO
             dqm       = (eod(i+2,j  ,k) - eod(i+2,j-1,k)) * dx2bi(j  )
             dqp       = (eod(i+2,j+1,k) - eod(i+2,j  ,k)) * dx2bi(j+1)
             deod(j,3) = max ( dqm * dqp, zro )
     1                 * sign ( one, dqm + dqp )
     2                 / max ( abs ( dqm + dqp ), tiny )
#endif /* not ISO  */
#if defined RAD || defined RAD_EXP
             dqm       = (ero(i+2,j  ,k) - ero(i+2,j-1,k)) * dx2bi(j  )
             dqp       = (ero(i+2,j+1,k) - ero(i+2,j  ,k)) * dx2bi(j+1)
             dero(j,3) = max ( dqm * dqp, zro )
     1                 * sign ( one, dqm + dqp )
     2                 / max ( abs ( dqm + dqp ), tiny )
#endif /* RAD || RAD_EXP */
c
#endif
#if UNROLL_J>3
             dqm       = (dlo(i+3,j  ,k) - dlo(i+3,j-1,k)) * dx2bi(j  )
             dqp       = (dlo(i+3,j+1,k) - dlo(i+3,j  ,k)) * dx2bi(j+1)
             dd  (j,4) = max ( dqm * dqp, zro )
     1                 * sign ( one, dqm + dqp )
     2                 / max ( abs ( dqm + dqp ), tiny )
#ifndef ISO
             dqm       = (eod(i+3,j  ,k) - eod(i+3,j-1,k)) * dx2bi(j  )
             dqp       = (eod(i+3,j+1,k) - eod(i+3,j  ,k)) * dx2bi(j+1)
             deod(j,4) = max ( dqm * dqp, zro )
     1                 * sign ( one, dqm + dqp )
     2                 / max ( abs ( dqm + dqp ), tiny )
#endif /* not ISO  */
#if defined RAD || defined RAD_EXP
             dqm       = (ero(i+3,j  ,k) - ero(i+3,j-1,k)) * dx2bi(j  )
             dqp       = (ero(i+3,j+1,k) - ero(i+3,j  ,k)) * dx2bi(j+1)
             dero(j,4) = max ( dqm * dqp, zro )
     1                 * sign ( one, dqm + dqp )
     2                 / max ( abs ( dqm + dqp ), tiny )
#endif /* RAD || RAD_EXP */
c
#endif
2030       continue
c
c     2.  Choose time averaged, upwinded interface value.
c
c  For the purposes of consistent advection, construct the mass
c  flux across each 1-interface.  The mass flux will be used to create
c  the fluxes of all variables, including the momenta which are updated
c  in MOMX1.
c
             do 2040 j=jbeg-1,jend+1
               xi         = ( v2  (i  ,j  ,k) - vg2(j  ) ) * dt
               q1         = sign ( haf, xi )
               dtwid (j,1)= ( 0.5 + q1 ) * ( dlo(i  ,j-1,k)
     1                    + ( dx2a(j-1) - xi ) * dd   (j-1,1) )
     2                    + ( 0.5 - q1 ) * ( dlo(i  ,j  ,k)
     3                    - ( dx2a(j  ) + xi ) * dd   (j  ,1) )
c
             mflux (j,1) = dtwid (j,1) * ( v2(i  ,j  ,k) - vg2(j) ) * dt
#ifdef MOVING_GRID
             dtwid (j,1) = mflux (j,1) * atwid (i  ) * g32ah(j  )
#else
             dtwid (j,1) = mflux (j,1) * atwid (i  ) * g32a (j  )
#endif /* MOVING_GRID */
#ifndef ISO
               etwid (j,1)= ( 0.5 + q1 ) * ( eod(i  ,j-1,k)
     1                    + ( dx2a(j-1) - xi ) * deod (j-1,1) )
     2                    + ( 0.5 - q1 ) * ( eod(i  ,j  ,k)
     3                    - ( dx2a(j  ) + xi ) * deod (j  ,1) )
c
               etwid (j,1)= dtwid (j,1) * etwid (j,1)
#endif /* not ISO */
#if defined RAD || defined RAD_EXP
               rtwid (j,1)= ( 0.5 + q1 ) * ( ero(i  ,j-1,k)
     1                    + ( dx2a(j-1) - xi ) * dero (j-1,1) )
     2                    + ( 0.5 - q1 ) * ( ero(i  ,j  ,k)
     3                    - ( dx2a(j  ) + xi ) * dero (j  ,1) )
c
               rtwid (j,1)= dtwid (j,1) * rtwid (j,1)
#endif /* RAD || RAD_EXP */
c
               xi         = ( v2  (i+1,j  ,k) - vg2(j  ) ) * dt
               q1         = sign ( haf, xi )
               dtwid (j,2)= ( 0.5 + q1 ) * ( dlo(i+1,j-1,k)
     1                    + ( dx2a(j-1) - xi ) * dd   (j-1,2) )
     2                    + ( 0.5 - q1 ) * ( dlo(i+1,j  ,k)
     3                    - ( dx2a(j  ) + xi ) * dd   (j  ,2) )
c
             mflux (j,2) = dtwid (j,2) * ( v2(i+1,j  ,k) - vg2(j) ) * dt
#ifdef MOVING_GRID
             dtwid (j,2) = mflux (j,2) * atwid (i+1) * g32ah(j  )
#else
             dtwid (j,2) = mflux (j,2) * atwid (i+1) * g32a (j  )
#endif /* MOVING_GRID */
#ifndef ISO
               etwid (j,2)= ( 0.5 + q1 ) * ( eod(i+1,j-1,k)
     1                    + ( dx2a(j-1) - xi ) * deod (j-1,2) )
     2                    + ( 0.5 - q1 ) * ( eod(i+1,j  ,k)
     3                    - ( dx2a(j  ) + xi ) * deod (j  ,2) )
c
               etwid (j,2)= dtwid (j,2) * etwid (j,2)
#endif /* not ISO */
#if defined RAD || defined RAD_EXP
               rtwid (j,2)= ( 0.5 + q1 ) * ( ero(i+1,j-1,k)
     1                    + ( dx2a(j-1) - xi ) * dero (j-1,2) )
     2                    + ( 0.5 - q1 ) * ( ero(i+1,j  ,k)
     3                    - ( dx2a(j  ) + xi ) * dero (j  ,2) )
c
               rtwid (j,2)= dtwid (j,2) * rtwid (j,2)
#endif /* RAD || RAD_EXP */
c
#if UNROLL_J>2
               xi         = ( v2  (i+2,j  ,k) - vg2(j  ) ) * dt
               q1         = sign ( haf, xi )
               dtwid (j,3)= ( 0.5 + q1 ) * ( dlo(i+2,j-1,k)
     1                    + ( dx2a(j-1) - xi ) * dd   (j-1,3) )
     2                    + ( 0.5 - q1 ) * ( dlo(i+2,j  ,k)
     3                    - ( dx2a(j  ) + xi ) * dd   (j  ,3) )
c
             mflux (j,3) = dtwid (j,3) * ( v2(i+2,j  ,k) - vg2(j) ) * dt
#ifdef MOVING_GRID
             dtwid (j,3) = mflux (j,3) * atwid (i+2) * g32ah(j  )
#else
             dtwid (j,3) = mflux (j,3) * atwid (i+2) * g32a (j  )
#endif /* MOVING_GRID */
#ifndef ISO
               etwid (j,3)= ( 0.5 + q1 ) * ( eod(i+2,j-1,k)
     1                    + ( dx2a(j-1) - xi ) * deod (j-1,3) )
     2                    + ( 0.5 - q1 ) * ( eod(i+2,j  ,k)
     3                    - ( dx2a(j  ) + xi ) * deod (j  ,3) )
c
               etwid (j,3)= dtwid (j,3) * etwid (j,3)
#endif /* not ISO */
#if defined RAD || defined RAD_EXP
               rtwid (j,3)= ( 0.5 + q1 ) * ( ero(i+2,j-1,k)
     1                    + ( dx2a(j-1) - xi ) * dero (j-1,3) )
     2                    + ( 0.5 - q1 ) * ( ero(i+2,j  ,k)
     3                    - ( dx2a(j  ) + xi ) * dero (j  ,3) )
c
               rtwid (j,3)= dtwid (j,3) * rtwid (j,3)
#endif /* RAD || RAD_EXP */
c
#endif
#if UNROLL_J>3
               xi         = ( v2  (i+3,j  ,k) - vg2(j  ) ) * dt
               q1         = sign ( haf, xi )
               dtwid (j,4)= ( 0.5 + q1 ) * ( dlo(i+3,j-1,k)
     1                    + ( dx2a(j-1) - xi ) * dd   (j-1,4) )
     2                    + ( 0.5 - q1 ) * ( dlo(i+3,j  ,k)
     3                    - ( dx2a(j  ) + xi ) * dd   (j  ,4) )
c
             mflux (j,4) = dtwid (j,4) * ( v2(i+3,j  ,k) - vg2(j) ) * dt
#ifdef MOVING_GRID
             dtwid (j,4) = mflux (j,4) * atwid (i+3) * g32ah(j  )
#else
             dtwid (j,4) = mflux (j,4) * atwid (i+3) * g32a (j  )
#endif /* MOVING_GRID */
#ifndef ISO
               etwid (j,4)= ( 0.5 + q1 ) * ( eod(i+3,j-1,k)
     1                    + ( dx2a(j-1) - xi ) * deod (j-1,4) )
     2                    + ( 0.5 - q1 ) * ( eod(i+3,j  ,k)
     3                    - ( dx2a(j  ) + xi ) * deod (j  ,4) )
c
               etwid (j,4)= dtwid (j,4) * etwid (j,4)
#endif /* not ISO */
#if defined RAD || defined RAD_EXP
               rtwid (j,4)= ( 0.5 + q1 ) * ( ero(i+3,j-1,k)
     1                    + ( dx2a(j-1) - xi ) * dero (j-1,4) )
     2                    + ( 0.5 - q1 ) * ( ero(i+3,j  ,k)
     3                    - ( dx2a(j  ) + xi ) * dero (j  ,4) )
c
               rtwid (j,4)= dtwid (j,4) * rtwid (j,4)
#endif /* RAD || RAD_EXP */
c
#endif
2040         continue
c
c  Save the mass flux outside (ibeg:iend,jbeg:jend,kbeg:kend)
c  only for zones next to the inner orders.
c
c               if ( (i.eq.is-1 .and. k.ge.kbeg) .or.
c     &              (i.ge.ibeg .and. k.eq.ks-1) .or.
c     &              (i.eq.is-1 .and. k.eq.ks-1) ) then
c
             if (k.eq.ks-1) then
c
               if (jbeg .eq. js) then
                 mflx(i  ,js-1,k) = mflux (js-1,1)
                 mflx(i+1,js-1,k) = mflux (js-1,2)
#if UNROLL_J>2
                 mflx(i+2,js-1,k) = mflux (js-1,3)
#endif
#if UNROLL_J>3
                 mflx(i+3,js-1,k) = mflux (js-1,4)
#endif
               endif
                 do 2060 j=jbeg,jend
                   mflx(i  ,j,k) = mflux(j   ,1)
                   mflx(i+1,j,k) = mflux(j   ,2)
#if UNROLL_J>2
                   mflx(i+2,j,k) = mflux(j   ,3)
#endif
#if UNROLL_J>3
                   mflx(i+3,j,k) = mflux(j   ,4)
#endif
2060             continue
                 if (jend .eq. je) then
                   mflx(i  ,je+1,k) = mflux (je+1,1)
                   mflx(i+1,je+1,k) = mflux (je+1,2)
#if UNROLL_J>2
                   mflx(i+2,je+1,k) = mflux (je+1,3)
#endif
#if UNROLL_J>3
                   mflx(i+3,je+1,k) = mflux (je+1,4)
#endif
                 endif
c
               endif  !  k=ks-1
c
c  Perform mass density and energy density advection.  Note that
c  the timestep "dt" is hidden the fluxes "dflx" and "eflx".
c
             if (k.ge.kbeg) then
c
               if (jbeg .eq. js) then
                 mflx(i  ,js-1,k) = mflux (js-1,1)
                 mflx(i+1,js-1,k) = mflux (js-1,2)
#if UNROLL_J>2
                 mflx(i+2,js-1,k) = mflux (js-1,3)
#endif
#if UNROLL_J>3
                 mflx(i+3,js-1,k) = mflux (js-1,4)
#endif
               endif
               do 2080 j=jbeg,jend
               mflx(i  ,j,k)= mflux (j,1)
               den(i  ,j,k) = ( dlo(i  ,j,k) * dvl2a(j)
#ifdef MOVING_GRID
     1                      - dtwid (j+1,1) + dtwid (j,1) ) * dvl2ani(j)
#else
     1                      - dtwid (j+1,1) + dtwid (j,1) ) * dvl2a i(j)
#endif /* MOVING_GRID */
#ifndef ISO
               e  (i  ,j,k) = ( e  (i  ,j,k) * dvl2a(j)
#ifdef MOVING_GRID
     1                      - etwid (j+1,1) + etwid (j,1) ) * dvl2ani(j)
#else
     1                      - etwid (j+1,1) + etwid (j,1) ) * dvl2a i(j)
#endif /* MOVING_GRID */
c
c Compute e/d for the next substep.
c
#ifndef TOTAL_ENERGY
               edn(i  ,j,k) =         e(i  ,j,k) / den(i  ,j,k)
#else /* not TOTAL_ENERGY */
               edn(i  ,j,k) = gamma * e(i  ,j,k) / den(i  ,j,k) - gamm1
     1                      * ( ( v1(i  ,j,k) + v1(i+1,j  ,k  ) )**2
     2                        + ( v2(i  ,j,k) + v2(i  ,j+1,k  ) )**2
     3                        + ( v3(i  ,j,k) + v3(i  ,j  ,k+1) )**2
     4                        ) * 0.125
#if defined GRAV || defined GRAV_FFT
     5                      - gamm1 * gp(i  ,j,k)
#endif /* GRAV || GRAV_FFT */
#endif /* not TOTAL_ENERGY */
#endif /* not ISO */
#if defined RAD || defined RAD_EXP
               ern(i  ,j,k) = ( ero(i  ,j,k) * dlo(i  ,j,k) * dvl2a(j)
#ifdef MOVING_GRID
     1                      - rtwid (j+1,1) + rtwid (j,1) ) * dvl2ani(j)
#else
     1                      - rtwid (j+1,1) + rtwid (j,1) ) * dvl2a i(j)
#endif /* MOVING_GRID */
c
c Work with er/d for the next sweep.
c
               ern(i  ,j,k) = ern(i  ,j,k) / den(i  ,j,k)
#endif /* RAD || RAD_EXP */
c
               mflx(i+1,j,k)= mflux (j,2)
               den(i+1,j,k) = ( dlo(i+1,j,k) * dvl2a(j)
#ifdef MOVING_GRID
     1                      - dtwid (j+1,2) + dtwid (j,2) ) * dvl2ani(j)
#else
     1                      - dtwid (j+1,2) + dtwid (j,2) ) * dvl2a i(j)
#endif /* MOVING_GRID */
#ifndef ISO
               e  (i+1,j,k) = ( e  (i+1,j,k) * dvl2a(j)
#ifdef MOVING_GRID
     1                      - etwid (j+1,2) + etwid (j,2) ) * dvl2ani(j)
#else
     1                      - etwid (j+1,2) + etwid (j,2) ) * dvl2a i(j)
#endif /* MOVING_GRID */
c
c Compute e/d for the next substep.
c
#ifndef TOTAL_ENERGY
               edn(i+1,j,k) =         e(i+1,j,k) / den(i+1,j,k)
#else /* not TOTAL_ENERGY */
               edn(i+1,j,k) = gamma * e(i+1,j,k) / den(i+1,j,k) - gamm1
     1                      * ( ( v1(i+1,j,k) + v1(i+1,j  ,k  ) )**2
     2                        + ( v2(i+1,j,k) + v2(i+1,j+1,k  ) )**2
     3                        + ( v3(i+1,j,k) + v3(i+1,j  ,k+1) )**2
     4                        ) * 0.125
#if defined GRAV || defined GRAV_FFT
     5                      - gamm1 * gp(i+1,j,k)
#endif /* GRAV || GRAV_FFT */
#endif /* not TOTAL_ENERGY */
#endif /* not ISO */
#if defined RAD || defined RAD_EXP
               ern(i+1,j,k) = ( ero(i+1,j,k) * dlo(i+1,j,k) * dvl2a(j)
#ifdef MOVING_GRID
     1                      - rtwid (j+1,2) + rtwid (j,2) ) * dvl2ani(j)
#else
     1                      - rtwid (j+1,2) + rtwid (j,2) ) * dvl2a i(j)
#endif /* MOVING_GRID */
c
c Work with er/d for the next sweep.
c
               ern(i+1,j,k) = ern(i+1,j,k) / den(i+1,j,k)
#endif /* RAD || RAD_EXP */
c
#if UNROLL_J>2
               mflx(i+2,j,k)= mflux (j,3)
               den(i+2,j,k) = ( dlo(i+2,j,k) * dvl2a(j)
#ifdef MOVING_GRID
     1                      - dtwid (j+1,3) + dtwid (j,3) ) * dvl2ani(j)
#else
     1                      - dtwid (j+1,3) + dtwid (j,3) ) * dvl2a i(j)
#endif /* MOVING_GRID */
#ifndef ISO
               e  (i+2,j,k) = ( e  (i+2,j,k) * dvl2a(j)
#ifdef MOVING_GRID
     1                      - etwid (j+1,3) + etwid (j,3) ) * dvl2ani(j)
#else
     1                      - etwid (j+1,3) + etwid (j,3) ) * dvl2a i(j)
#endif /* MOVING_GRID */
c
c Compute e/d for the next substep.
c
#ifndef TOTAL_ENERGY
               edn(i+2,j,k) =         e(i+2,j,k) / den(i+2,j,k)
#else /* not TOTAL_ENERGY */
               edn(i+2,j,k) = gamma * e(i+2,j,k) / den(i+2,j,k) - gamm1
     1                      * ( ( v1(i+2,j,k) + v1(i+1,j  ,k  ) )**2
     2                        + ( v2(i+2,j,k) + v2(i+2,j+1,k  ) )**2
     3                        + ( v3(i+2,j,k) + v3(i+2,j  ,k+1) )**2
     4                        ) * 0.125
#if defined GRAV || defined GRAV_FFT
     5                      - gamm1 * gp(i+2,j,k)
#endif /* GRAV || GRAV_FFT */
#endif /* not TOTAL_ENERGY */
#endif /* not ISO */
#if defined RAD || defined RAD_EXP
               ern(i+2,j,k) = ( ero(i+2,j,k) * dlo(i+2,j,k) * dvl2a(j)
#ifdef MOVING_GRID
     1                      - rtwid (j+1,3) + rtwid (j,3) ) * dvl2ani(j)
#else
     1                      - rtwid (j+1,3) + rtwid (j,3) ) * dvl2a i(j)
#endif /* MOVING_GRID */
c
c Work with er/d for the next sweep.
c
               ern(i+2,j,k) = ern(i+2,j,k) / den(i+2,j,k)
#endif /* RAD || RAD_EXP */
c
#endif
#if UNROLL_J>3
               mflx(i+3,j,k)= mflux (j,4)
               den(i+3,j,k) = ( dlo(i+3,j,k) * dvl2a(j)
#ifdef MOVING_GRID
     1                      - dtwid (j+1,4) + dtwid (j,4) ) * dvl2ani(j)
#else
     1                      - dtwid (j+1,4) + dtwid (j,4) ) * dvl2a i(j)
#endif /* MOVING_GRID */
#ifndef ISO
               e  (i+3,j,k) = ( e  (i+3,j,k) * dvl2a(j)
#ifdef MOVING_GRID
     1                      - etwid (j+1,4) + etwid (j,4) ) * dvl2ani(j)
#else
     1                      - etwid (j+1,4) + etwid (j,4) ) * dvl2a i(j)
#endif /* MOVING_GRID */
c
c Compute e/d for the next substep.
c
#ifndef TOTAL_ENERGY
               edn(i+3,j,k) =         e(i+3,j,k) / den(i+3,j,k)
#else /* not TOTAL_ENERGY */
               edn(i+3,j,k) = gamma * e(i+3,j,k) / den(i+3,j,k) - gamm1
     1                      * ( ( v1(i+3,j,k) + v1(i+3,j  ,k  ) )**2
     2                        + ( v2(i+3,j,k) + v2(i+3,j+1,k  ) )**2
     3                        + ( v3(i+3,j,k) + v3(i+3,j  ,k+1) )**2
     4                        ) * 0.125
#if defined GRAV || defined GRAV_FFT
     5                      - gamm1 * gp(i+3,j,k)
#endif /* GRAV || GRAV_FFT */
#endif /* not TOTAL_ENERGY */
#endif /* not ISO */
#if defined RAD || defined RAD_EXP
               ern(i+3,j,k) = ( ero(i+3,j,k) * dlo(i+3,j,k) * dvl2a(j)
#ifdef MOVING_GRID
     1                      - rtwid (j+1,4) + rtwid (j,4) ) * dvl2ani(j)
#else
     1                      - rtwid (j+1,4) + rtwid (j,4) ) * dvl2a i(j)
#endif /* MOVING_GRID */
c
c Work with er/d for the next sweep.
c
               ern(i+3,j,k) = ern(i+3,j,k) / den(i+3,j,k)
#endif /* RAD || RAD_EXP */
c
#endif
2080         continue
               if (jend .eq. je) then
                 mflx(i  ,je+1,k) = mflux (je+1,1)
                 mflx(i+1,je+1,k) = mflux (je+1,2)
#if UNROLL_J>2
                 mflx(i+2,je+1,k) = mflux (je+1,3)
#endif
#if UNROLL_J>3
                 mflx(i+3,je+1,k) = mflux (je+1,4)
#endif
               endif  !  jend>=je
             endif  !  k>=kbeg
2090       continue
#endif /* UNROLL_J>1 */
         endif  !  j=js-1 || j>=jbeg
2100   continue
c
       return
       end
c
c=======================================================================
c
c    \\\\\\\\\\        E N D   S U B R O U T I N E        //////////
c    //////////                T R A N X 2                \\\\\\\\\\
c
c=======================================================================
c
c
