#include "zeusmp.def"
c=======================================================================
c
c    \\\\\\\\\\      B E G I N   S U B R O U T I N E      //////////
c    //////////                T R A N X 3                \\\\\\\\\\
c
c                            Developed by
c                Laboratory of Computational Astrophysics
c               University of Illinois at Urbana-Champaign
c
c=======================================================================
c
       subroutine tranx3 (ibeg,iend,jbeg,jend,kbeg,kend
     &                   ,dlo,den
#ifndef ISO
     &                   ,eod,edn
#endif
#if defined RAD || defined RAD_EXP
     &                   ,ero,ern
#endif /* RAD || RAD_EXP */
     &                   ,mflx)
c
c    dac:zeus3d.tranx3 <----- transports zone-centred variables along x3
c    from jms:zeus2d.tranx2, mln:zeus04.tranz                  may, 1990
c
c    written by: David Clarke
c    modified 1: June 1992, by David Clarke; added the total energy
c                option originally designed by Byung-IL Jun.
c    modified 2: Feb. 22, 1996 by Robert Fiedler; completely rewritten
c                for ZEUS-MP.
c    modified 3: Aug. 6,  1996 by Robert Fiedler; rearranged for
c                maximum efficiency, and unrolled i-loops.
c    modified 3: Dec. 18,  1996 by Robert Fiedler; added radiation.
c    modified 4: Jul. 15, 2005 by John C. Vernaleo geometry fix
c                         from John Hayes
c
c  PURPOSE:  Transports all zone centred variables in the 3-direction
c  only.  Currently transported are:
c
c                      mass   density
c                      energy density
c
c  The consistent advection algorithm, in which mass fluxes are used to
c  construct the fluxes of all variables across the interfaces, is used
c  for all hydrodynamical variables.  Thus, the mass fluxes are passed
c  to MOMX3 in order to transport the momenta as well.  The magnetic
c  field components are treated separately from the mass fluxes in CT.
c  Interpolations are done in-line.
c
c  INPUT VARIABLES: 
c    ibeg,iend,jbeg,jend,kbeg,kend  index ranges to cover.
c    dlo      mass            density at previous substep.
c    eod      specific energy density at previous substep; equals
c             (e+p)/d  if TOTAL_ENERGY is defined.
c
c BOUNDARY VALUES USED:
c
c    Macro defined  var   ii    oi    ij    oj    ik    ok
c    -------------  ---  ----  ----  ----  ----  ----  ----
c                    d   is-1        js-1        ks-3  ke+2
c                   e/d                          ks-2  ke+2
c    TOTAL_ENERGY    u1        ie+1
c    TOTAL_ENERGY    u2                    je+1
c                    u3  is-1        js-1        ks-1  ke+1
c
c  OUTPUT VARIABLES:
c    den      updated mass            density.
c    edn      updated specific energy density.
c    mflx     mass flux (in the 3-direction)
c
c  LOCAL VARIABLES:
c    atwid    effective cross sectional area of the 1-interfaces
c    atwid2   effective cross sectional area of the 2-interfaces
c    etwid    interpolated specific energy densities (e/d) at all
c             3-interfaces
c    eflx     energy density flux across all 3-interfaces (reuse etwid)
c    dtwid    interpolated mass densities at all 3-interfaces
c    dflx     mass density flux across all 3-interfaces   (reuse dtwid)
c
c  EXTERNALS:
c
c-----------------------------------------------------------------------
c
       implicit NONE
#include "param.h"
#include "root.h"
#include "field.h"
#include "bndry.h"
#include "grid.h"
#include "scratch.h"
#include "mpicom.h"
c
       integer i, j, k, ibeg,iend,jbeg,jend,kbeg,kend
c
       REAL    dqm,dqp, xi,q1
c
       REAL        atwid2(ijkn),  atwid (ijkn)
#ifndef UNROLL_K
#define UNROLL_K 1
#endif
#if UNROLL_K>1
       integer iriter
#endif
       REAL        mflux (ijkn,UNROLL_K)
     &           , dtwid (ijkn,UNROLL_K),  dd    (ijkn,UNROLL_K)
#ifndef ISO
     &           , etwid (ijkn,UNROLL_K),  deod  (ijkn,UNROLL_K)
#endif
#if defined RAD || defined RAD_EXP
     &           , rtwid (ijkn,UNROLL_K),  dero  (ijkn,UNROLL_K)
#endif /* RAD || RAD_EXP */
c
c
       REAL  mflx (in,jn,kn)
       REAL  dlo  (in,jn,kn), den  (in,jn,kn)
#ifndef ISO
     &      ,eod  (in,jn,kn), edn  (in,jn,kn)
#endif /* ISO */
#if defined RAD || defined RAD_EXP
     &      ,ero  (in,jn,kn), ern  (in,jn,kn)
#endif /* RAD || RAD_EXP */
c
       equivalence (atwid , w1da), (atwid2, w1db)
#if UNROLL_K==1
     &           , (dtwid , w1dc), (dd    , w1dd)
     &           , (mflux , w1de)
#ifndef ISO
     &           , (etwid , w1df), (deod  , w1dg)
#endif
#endif /* UNROLL_K==1 */
#if UNROLL_K==2
     &           , (dtwid , w1dc), (dd    , w1de)
     &           , (mflux , w1dg)
#ifndef ISO
     &           , (etwid , w1di)
#endif
#endif /* UNROLL_K==2 */
c
c
c-----------------------------------------------------------------------
c
c Compute time-centered area factors.
c
       do 10 i=ibeg-1,iend+1
         atwid (i)           = g2b(i) * dx1a(i) * dvl1ai(i)
10     continue
       do 20 j=jbeg-1,jend+1
         atwid2(j)           = dx2a(j) * dvl2ai(j)
20     continue
c
c Transport all zone-centered quantities in the 3 direction.
c
       do 2100 j=jbeg-1,jend
         if ( (j.ge.jbeg) .or. (j.eq.js-1) ) then
c.......................................................................
c
c Split off the i=ibeg-1 iteration to ease unrolling.
c
           i = ibeg - 1
c
c   Interpolate to obtain zone-centered quantities at zone faces.
c
c     1.  Evaluate monotonised, van Leer differences across the zone.
c
             if (kbeg .eq. ks) then  !  Need d(ks-3) from neighbor.
               k        = ks - 2
               dqm      = (dlo(i  ,j,k  ) - dikb (i  ,j,3)) * dx3bi(k  )
               dqp      = (dlo(i  ,j,k+1) - dlo(i  ,j,k  )) * dx3bi(k+1)
               dd(k,1)  = max ( dqm * dqp, zro )
     1                  * sign ( one, dqm + dqp )
     2                  / max ( abs ( dqm + dqp ), tiny )
             endif
             do 30 k=max(kbeg-2,ks-1),kend+1
               dqm      = (dlo(i  ,j,k  ) - dlo(i  ,j,k-1)) * dx3bi(k  )
               dqp      = (dlo(i  ,j,k+1) - dlo(i  ,j,k  )) * dx3bi(k+1)
               dd(k,1)  = max ( dqm * dqp, zro )
     1                  * sign ( one, dqm + dqp )
     2                  / max ( abs ( dqm + dqp ), tiny )
30         continue
c
c     2.  Choose time averaged, upwinded interface value.
c
c  For the purposes of consistent advection, construct the mass
c  flux across each 1-interface.  The mass flux will be used to create
c  the fluxes of all variables, including the momenta which are updated
c  in MOMX1.
c
             do 40 k=kbeg-1,kend+1
               xi          = ( v3  (i  ,j,k  ) - vg3(k  ) ) * dt
     &               * g31bi(i) * g32bi(j)
               q1          = sign ( haf, xi )
               dtwid (k,1) = ( 0.5 + q1 ) * ( dlo(i  ,j,k-1)
     1                     + ( dx3a(k-1) - xi ) * dd   (k-1,1) )
     2                     + ( 0.5 - q1 ) * ( dlo(i  ,j,k  )
     3                     - ( dx3a(k  ) + xi ) * dd   (k  ,1) )
c
               mflux (k,1) = dtwid (k,1) * ( v3(i  ,j,k) - vg3(k) ) * dt
40           continue
c
c  Save the mass flux outside (ibeg:iend,jbeg:jend,kbeg:kend)
c  only for zones along the inner borders.
c
             if (kbeg .eq. ks) mflx(i  ,j,ks-1) = mflux (ks-1,1)
             do 60 k=kbeg,kend
               mflx(i  ,j,k) = mflux (k   ,1)
60           continue
             if (kend .eq. ke) mflx(i  ,j,ke+1) = mflux (ke+1,1)
c
c.......................................................................
c
#if UNROLL_K>1
c Do the iterations which will not be done in unrolled loops next.
c
           iriter = mod(iend-ibeg+1,UNROLL_K)
           do 1090 i=ibeg,ibeg+iriter-1
#else /* UNROLL_K>1 */
           do 1090 i=ibeg,iend
#endif /* UNROLL_K>1 */
c
c   Interpolate to obtain zone-centered quantities at zone faces.
c
c     1.  Evaluate monotonised, van Leer differences across the zone.
c
             if (kbeg .eq. ks) then  !  Need d(ks-3) from neighbor.
               k        = ks - 2
               dqm      = (dlo(i  ,j,k  ) - dikb (i  ,j,3)) * dx3bi(k  )
               dqp      = (dlo(i  ,j,k+1) - dlo(i  ,j,k  )) * dx3bi(k+1)
               dd(k,1)  = max ( dqm * dqp, zro )
     1                  * sign ( one, dqm + dqp )
     2                  / max ( abs ( dqm + dqp ), tiny )
#ifndef ISO
               deod(k,1) = zro  ! Not valid, but we don't use it.
#endif /* not ISO  */
#if defined RAD || defined RAD_EXP
               dero(k,1) = zro  ! Not valid, but we don't use it.
#endif /* RAD || RAD_EXP */
             endif
             do 1030 k=max(kbeg-2,ks-1),kend+1
               dqm      = (dlo(i  ,j,k  ) - dlo(i  ,j,k-1)) * dx3bi(k  )
               dqp      = (dlo(i  ,j,k+1) - dlo(i  ,j,k  )) * dx3bi(k+1)
               dd(k,1)  = max ( dqm * dqp, zro )
     1                  * sign ( one, dqm + dqp )
     2                  / max ( abs ( dqm + dqp ), tiny )
#ifndef ISO
               dqm      = (eod(i  ,j,k  ) - eod(i  ,j,k-1)) * dx3bi(k  )
               dqp      = (eod(i  ,j,k+1) - eod(i  ,j,k  )) * dx3bi(k+1)
               deod(k,1)= max ( dqm * dqp, zro )
     1                  * sign ( one, dqm + dqp )
     2                  / max ( abs ( dqm + dqp ), tiny )
#endif /* not ISO  */
#if defined RAD || defined RAD_EXP
               dqm      = (ero(i  ,j,k  ) - ero(i  ,j,k-1)) * dx3bi(k  )
               dqp      = (ero(i  ,j,k+1) - ero(i  ,j,k  )) * dx3bi(k+1)
               dero(k,1)= max ( dqm * dqp, zro )
     1                  * sign ( one, dqm + dqp )
     2                  / max ( abs ( dqm + dqp ), tiny )
#endif /* RAD || RAD_EXP */
1030       continue
c
c     2.  Choose time averaged, upwinded interface value.
c
c  For the purposes of consistent advection, construct the mass
c  flux across each 1-interface.  The mass flux will be used to create
c  the fluxes of all variables, including the momenta which are updated
c  in MOMX1.
c
             do 1040 k=kbeg-1,kend+1
               xi          = ( v3  (i  ,j,k  ) - vg3(k  ) ) * dt
               q1          = sign ( haf, xi )
               dtwid (k,1) = ( 0.5 + q1 ) * ( dlo(i  ,j,k-1)
     1                     + ( dx3a(k-1) - xi ) * dd   (k-1,1) )
     2                     + ( 0.5 - q1 ) * ( dlo(i  ,j,k  )
     3                     - ( dx3a(k  ) + xi ) * dd   (k  ,1) )
c
               mflux (k,1) = dtwid (k,1) * ( v3(i  ,j,k) - vg3(k) ) * dt
               dtwid (k,1) = mflux (k,1) * atwid (i  ) * atwid2(j)
#ifndef ISO
               etwid (k,1) = ( 0.5 + q1 ) * ( eod(i  ,j,k-1)
     1                     + ( dx3a(k-1) - xi ) * deod (k-1,1) )
     2                     + ( 0.5 - q1 ) * ( eod(i  ,j,k  )
     3                     - ( dx3a(k  ) + xi ) * deod (k  ,1) )
c
               etwid (k,1) = dtwid (k,1) * etwid (k,1)
#endif /* not ISO */
#if defined RAD || defined RAD_EXP
               rtwid (k,1) = ( 0.5 + q1 ) * ( ero(i  ,j,k-1)
     1                     + ( dx3a(k-1) - xi ) * dero (k-1,1) )
     2                     + ( 0.5 - q1 ) * ( ero(i  ,j,k  )
     3                     - ( dx3a(k  ) + xi ) * dero (k  ,1) )
c
               rtwid (k,1) = dtwid (k,1) * rtwid (k,1)
#endif /* RAD || RAD_EXP */
1040         continue
c
c  Save the mass flux outside (ibeg:iend,jbeg:jend,kbeg:kend)
c  only for zones along the inner borders.
c
             if (j.eq.js-1) then
c
               if (kbeg .eq. ks) mflx(i  ,j,ks-1) = mflux (ks-1,1)
               do 1060 k=kbeg,kend
                 mflx(i  ,j,k) = mflux (k   ,1)
1060           continue
               if (kend .eq. ke) mflx(i  ,j,ke+1) = mflux (ke+1,1)
c
             endif
c
c  Perform mass density and energy density advection.  Note that
c  the timestep "dt" is hidden the fluxes "dtwid" and "etwid".
c
             if (j.ge.jbeg) then
c
               if (kbeg .eq. ks) mflx(i  ,j,ks-1) = mflux (ks-1,1)
               do 1080 k=kbeg,kend
               mflx(i  ,j,k)= mflux (k   ,1)
               den(i  ,j,k) = ( dlo(i  ,j,k) * dvl3a(k)
#ifdef MOVING_GRID
     1                      - dtwid (k+1,1) + dtwid (k,1) ) * dvl3ani(k)
#else
     1                      - dtwid (k+1,1) + dtwid (k,1) ) * dvl3a i(k)
#endif /* MOVING GRID */
#ifndef ISO
               e  (i  ,j,k) = ( e  (i  ,j,k) * dvl3a(k)
#ifdef MOVING_GRID
     1                      - etwid (k+1,1) + etwid (k,1) ) * dvl3ani(k)
#else
     1                      - etwid (k+1,1) + etwid (k,1) ) * dvl3a i(k)
#endif /* MOVING GRID */
c
c Compute e/d for the next substep.
c
#ifndef TOTAL_ENERGY
               edn(i  ,j,k) =         e(i  ,j,k) / den(i  ,j,k)
#else /* not TOTAL_ENERGY */
               edn(i  ,j,k) = gamma * e(i  ,j,k) / den(i  ,j,k) - gamm1
     1                      * (
     2                            ( v1(i  ,j,k) + v1(i+1,j  ,k  ) )**2
     3                          + ( v2(i  ,j,k) + v2(i  ,j+1,k  ) )**2
     4                          + ( v3(i  ,j,k) + v3(i  ,j  ,k+1) )**2
     5                        ) * 0.125
#if defined GRAV || defined GRAV_FFT
     5                      - gamm1 * gp(i  ,j,k)
#endif /* GRAV || GRAV_FFT */
#endif /* not TOTAL_ENERGY */
#endif /* not ISO */
#if defined RAD || defined RAD_EXP
               ern(i  ,j,k) = ( ero(i  ,j,k) * dlo(i  ,j,k) * dvl3a(k)
#ifdef MOVING_GRID
     1                      - rtwid (k+1,1) + rtwid (k,1) ) * dvl3ani(k)
#else
     1                      - rtwid (k+1,1) + rtwid (k,1) ) * dvl3a i(k)
#endif /* MOVING GRID */
c
c Work with er/d for the next sweep.
c
               ern(i  ,j,k) = ern(i  ,j,k) / den(i  ,j,k)
#endif /* RAD || RAD_EXP */
1080           continue
               if (kend .eq. ke) mflx(i  ,j,ke+1) = mflux (ke+1,1)
             endif
1090       continue
#if UNROLL_K>1
c.......................................................................
c
c Do the bulk of the iterations in unrolled loops.
c
           do 2090 i=ibeg+iriter,iend,UNROLL_K
c
c   Interpolate to obtain zone-centered quantities at zone faces.
c
c     1.  Evaluate monotonised, van Leer differences across the zone.
c
             if (kbeg .eq. ks) then  !  Need d(ks-3) from neighbor.
               k        = ks - 2
               dqm      = (dlo(i  ,j,k  ) - dikb (i  ,j,3)) * dx3bi(k  )
               dqp      = (dlo(i  ,j,k+1) - dlo(i  ,j,k  )) * dx3bi(k+1)
               dd(k,1)  = max ( dqm * dqp, zro )
     1                  * sign ( one, dqm + dqp )
     2                  / max ( abs ( dqm + dqp ), tiny )
#ifndef ISO
               deod(k,1) = zro  ! Not valid, but we don't use it.
#endif /* not ISO  */
#if defined RAD || defined RAD_EXP
               dero(k,1) = zro  ! Not valid, but we don't use it.
#endif /* RAD || RAD_EXP */
c
               dqm      = (dlo(i+1,j,k  ) - dikb (i+1,j,3)) * dx3bi(k  )
               dqp      = (dlo(i+1,j,k+1) - dlo(i+1,j,k  )) * dx3bi(k+1)
               dd(k,2)  = max ( dqm * dqp, zro )
     1                  * sign ( one, dqm + dqp )
     2                  / max ( abs ( dqm + dqp ), tiny )
#ifndef ISO
               deod(k,2) = zro  ! Not valid, but we don't use it.
#endif /* not ISO  */
#if defined RAD || defined RAD_EXP
               dero(k,2) = zro  ! Not valid, but we don't use it.
#endif /* RAD || RAD_EXP */
c
#if UNROLL_K>2
               dqm      = (dlo(i+2,j,k  ) - dikb (i+2,j,3)) * dx3bi(k  )
               dqp      = (dlo(i+2,j,k+1) - dlo(i+2,j,k  )) * dx3bi(k+1)
               dd(k,3)  = max ( dqm * dqp, zro )
     1                  * sign ( one, dqm + dqp )
     2                  / max ( abs ( dqm + dqp ), tiny )
#ifndef ISO
               deod(k,3) = zro  ! Not valid, but we don't use it.
#endif /* not ISO  */
#if defined RAD || defined RAD_EXP
               dero(k,3) = zro  ! Not valid, but we don't use it.
#endif /* RAD || RAD_EXP */
c
#endif
#if UNROLL_K>3
               dqm      = (dlo(i+3,j,k  ) - dikb (i+3,j,3)) * dx3bi(k  )
               dqp      = (dlo(i+3,j,k+1) - dlo(i+3,j,k  )) * dx3bi(k+1)
               dd(k,4)  = max ( dqm * dqp, zro )
     1                  * sign ( one, dqm + dqp )
     2                  / max ( abs ( dqm + dqp ), tiny )
#ifndef ISO
               deod(k,4) = zro  ! Not valid, but we don't use it.
#endif /* not ISO  */
#if defined RAD || defined RAD_EXP
               dero(k,4) = zro  ! Not valid, but we don't use it.
#endif /* RAD || RAD_EXP */
c
#endif
             endif
             do 2030 k=max(kbeg-2,ks-1),kend+1
               dqm      = (dlo(i  ,j,k  ) - dlo(i  ,j,k-1)) * dx3bi(k  )
               dqp      = (dlo(i  ,j,k+1) - dlo(i  ,j,k  )) * dx3bi(k+1)
               dd(k,1)  = max ( dqm * dqp, zro )
     1                  * sign ( one, dqm + dqp )
     2                  / max ( abs ( dqm + dqp ), tiny )
#ifndef ISO
               dqm      = (eod(i  ,j,k  ) - eod(i  ,j,k-1)) * dx3bi(k  )
               dqp      = (eod(i  ,j,k+1) - eod(i  ,j,k  )) * dx3bi(k+1)
               deod(k,1)= max ( dqm * dqp, zro )
     1                  * sign ( one, dqm + dqp )
     2                  / max ( abs ( dqm + dqp ), tiny )
#endif /* not ISO  */
#if defined RAD || defined RAD_EXP
               dqm      = (ero(i  ,j,k  ) - ero(i  ,j,k-1)) * dx3bi(k  )
               dqp      = (ero(i  ,j,k+1) - ero(i  ,j,k  )) * dx3bi(k+1)
               dero(k,1)= max ( dqm * dqp, zro )
     1                  * sign ( one, dqm + dqp )
     2                  / max ( abs ( dqm + dqp ), tiny )
#endif /* RAD || RAD_EXP */
c
               dqm      = (dlo(i+1,j,k  ) - dlo(i+1,j,k-1)) * dx3bi(k  )
               dqp      = (dlo(i+1,j,k+1) - dlo(i+1,j,k  )) * dx3bi(k+1)
               dd(k,2)  = max ( dqm * dqp, zro )
     1                  * sign ( one, dqm + dqp )
     2                  / max ( abs ( dqm + dqp ), tiny )
#ifndef ISO
               dqm      = (eod(i+1,j,k  ) - eod(i+1,j,k-1)) * dx3bi(k  )
               dqp      = (eod(i+1,j,k+1) - eod(i+1,j,k  )) * dx3bi(k+1)
               deod(k,2)= max ( dqm * dqp, zro )
     1                  * sign ( one, dqm + dqp )
     2                  / max ( abs ( dqm + dqp ), tiny )
#endif /* not ISO  */
#if defined RAD || defined RAD_EXP
               dqm      = (ero(i+1,j,k  ) - ero(i+1,j,k-1)) * dx3bi(k  )
               dqp      = (ero(i+1,j,k+1) - ero(i+1,j,k  )) * dx3bi(k+1)
               dero(k,2)= max ( dqm * dqp, zro )
     1                  * sign ( one, dqm + dqp )
     2                  / max ( abs ( dqm + dqp ), tiny )
#endif /* RAD || RAD_EXP */
c
#if UNROLL_K>2
               dqm      = (dlo(i+2,j,k  ) - dlo(i+2,j,k-1)) * dx3bi(k  )
               dqp      = (dlo(i+2,j,k+1) - dlo(i+2,j,k  )) * dx3bi(k+1)
               dd(k,3)  = max ( dqm * dqp, zro )
     1                  * sign ( one, dqm + dqp )
     2                  / max ( abs ( dqm + dqp ), tiny )
#ifndef ISO
               dqm      = (eod(i+2,j,k  ) - eod(i+2,j,k-1)) * dx3bi(k  )
               dqp      = (eod(i+2,j,k+1) - eod(i+2,j,k  )) * dx3bi(k+1)
               deod(k,3)= max ( dqm * dqp, zro )
     1                  * sign ( one, dqm + dqp )
     2                  / max ( abs ( dqm + dqp ), tiny )
#endif /* not ISO  */
#if defined RAD || defined RAD_EXP
               dqm      = (ero(i+2,j,k  ) - ero(i+2,j,k-1)) * dx3bi(k  )
               dqp      = (ero(i+2,j,k+1) - ero(i+2,j,k  )) * dx3bi(k+1)
               dero(k,3)= max ( dqm * dqp, zro )
     1                  * sign ( one, dqm + dqp )
     2                  / max ( abs ( dqm + dqp ), tiny )
#endif /* RAD || RAD_EXP */
c
#endif
#if UNROLL_K>3
               dqm      = (dlo(i+3,j,k  ) - dlo(i+3,j,k-1)) * dx3bi(k  )
               dqp      = (dlo(i+3,j,k+1) - dlo(i+3,j,k  )) * dx3bi(k+1)
               dd(k,4)  = max ( dqm * dqp, zro )
     1                  * sign ( one, dqm + dqp )
     2                  / max ( abs ( dqm + dqp ), tiny )
#ifndef ISO
               dqm      = (eod(i+3,j,k  ) - eod(i+3,j,k-1)) * dx3bi(k  )
               dqp      = (eod(i+3,j,k+1) - eod(i+3,j,k  )) * dx3bi(k+1)
               deod(k,4)= max ( dqm * dqp, zro )
     1                  * sign ( one, dqm + dqp )
     2                  / max ( abs ( dqm + dqp ), tiny )
#endif /* not ISO  */
#if defined RAD || defined RAD_EXP
               dqm      = (ero(i+3,j,k  ) - ero(i+3,j,k-1)) * dx3bi(k  )
               dqp      = (ero(i+3,j,k+1) - ero(i+3,j,k  )) * dx3bi(k+1)
               dero(k,4)= max ( dqm * dqp, zro )
     1                  * sign ( one, dqm + dqp )
     2                  / max ( abs ( dqm + dqp ), tiny )
#endif /* RAD || RAD_EXP */
c
#endif
2030       continue
c
c     2.  Choose time averaged, upwinded interface value.
c
c  For the purposes of consistent advection, construct the mass
c  flux across each 1-interface.  The mass flux will be used to create
c  the fluxes of all variables, including the momenta which are updated
c  in MOMX1.
c
             do 2040 k=kbeg-1,kend+1
               xi          = ( v3  (i  ,j,k  ) - vg3(k  ) ) * dt
               q1          = sign ( haf, xi )
               dtwid (k,1) = ( 0.5 + q1 ) * ( dlo(i  ,j,k-1)
     1                     + ( dx3a(k-1) - xi ) * dd   (k-1,1) )
     2                     + ( 0.5 - q1 ) * ( dlo(i  ,j,k  )
     3                     - ( dx3a(k  ) + xi ) * dd   (k  ,1) )
c
               mflux (k,1) = dtwid (k,1) * ( v3(i  ,j,k) - vg3(k) ) * dt
               dtwid (k,1) = mflux (k,1) * atwid (i  ) * atwid2(j)
#ifndef ISO
               etwid (k,1) = ( 0.5 + q1 ) * ( eod(i  ,j,k-1)
     1                     + ( dx3a(k-1) - xi ) * deod (k-1,1) )
     2                     + ( 0.5 - q1 ) * ( eod(i  ,j,k  )
     3                     - ( dx3a(k  ) + xi ) * deod (k  ,1) )
c
               etwid (k,1) = dtwid (k,1) * etwid (k,1)
#endif /* not ISO */
#if defined RAD || defined RAD_EXP
               rtwid (k,1) = ( 0.5 + q1 ) * ( ero(i  ,j,k-1)
     1                     + ( dx3a(k-1) - xi ) * dero (k-1,1) )
     2                     + ( 0.5 - q1 ) * ( ero(i  ,j,k  )
     3                     - ( dx3a(k  ) + xi ) * dero (k  ,1) )
c
               rtwid (k,1) = dtwid (k,1) * rtwid (k,1)
#endif /* RAD || RAD_EXP */
c
               xi          = ( v3  (i+1,j,k  ) - vg3(k  ) ) * dt
               q1          = sign ( haf, xi )
               dtwid (k,2) = ( 0.5 + q1 ) * ( dlo(i+1,j,k-1)
     1                     + ( dx3a(k-1) - xi ) * dd   (k-1,2) )
     2                     + ( 0.5 - q1 ) * ( dlo(i+1,j,k  )
     3                     - ( dx3a(k  ) + xi ) * dd   (k  ,2) )
c
               mflux (k,2) = dtwid (k,2) * ( v3(i+1,j,k) - vg3(k) ) * dt
               dtwid (k,2) = mflux (k,2) * atwid (i+1) * atwid2(j)
#ifndef ISO
               etwid (k,2) = ( 0.5 + q1 ) * ( eod(i+1,j,k-1)
     1                     + ( dx3a(k-1) - xi ) * deod (k-1,2) )
     2                     + ( 0.5 - q1 ) * ( eod(i+1,j,k  )
     3                     - ( dx3a(k  ) + xi ) * deod (k  ,2) )
c
               etwid (k,2) = dtwid (k,2) * etwid (k,2)
#endif /* not ISO */
#if defined RAD || defined RAD_EXP
               rtwid (k,2) = ( 0.5 + q1 ) * ( ero(i+1,j,k-1)
     1                     + ( dx3a(k-1) - xi ) * dero (k-1,2) )
     2                     + ( 0.5 - q1 ) * ( ero(i+1,j,k  )
     3                     - ( dx3a(k  ) + xi ) * dero (k  ,2) )
c
               rtwid (k,2) = dtwid (k,2) * rtwid (k,2)
#endif /* RAD || RAD_EXP */
c
#if UNROLL_K>2
               xi          = ( v3  (i+2,j,k  ) - vg3(k  ) ) * dt
               q1          = sign ( haf, xi )
               dtwid (k,3) = ( 0.5 + q1 ) * ( dlo(i+2,j,k-1)
     1                     + ( dx3a(k-1) - xi ) * dd   (k-1,3) )
     2                     + ( 0.5 - q1 ) * ( dlo(i+2,j,k  )
     3                     - ( dx3a(k  ) + xi ) * dd   (k  ,3) )
c
               mflux (k,3) = dtwid (k,3) * ( v3(i+2,j,k) - vg3(k) ) * dt
               dtwid (k,3) = mflux (k,3) * atwid (i+2) * atwid2(j)
#ifndef ISO
               etwid (k,3) = ( 0.5 + q1 ) * ( eod(i+2,j,k-1)
     1                     + ( dx3a(k-1) - xi ) * deod (k-1,3) )
     2                     + ( 0.5 - q1 ) * ( eod(i+2,j,k  )
     3                     - ( dx3a(k  ) + xi ) * deod (k  ,3) )
c
               etwid (k,3) = dtwid (k,3) * etwid (k,3)
#endif /* not ISO */
#if defined RAD || defined RAD_EXP
               rtwid (k,3) = ( 0.5 + q1 ) * ( ero(i+2,j,k-1)
     1                     + ( dx3a(k-1) - xi ) * dero (k-1,3) )
     2                     + ( 0.5 - q1 ) * ( ero(i+2,j,k  )
     3                     - ( dx3a(k  ) + xi ) * dero (k  ,3) )
c
               rtwid (k,3) = dtwid (k,3) * rtwid (k,3)
#endif /* RAD || RAD_EXP */
c
#endif
#if UNROLL_K>3
               xi          = ( v3  (i+3,j,k  ) - vg3(k  ) ) * dt
               q1          = sign ( haf, xi )
               dtwid (k,4) = ( 0.5 + q1 ) * ( dlo(i+3,j,k-1)
     1                     + ( dx3a(k-1) - xi ) * dd   (k-1,4) )
     2                     + ( 0.5 - q1 ) * ( dlo(i+3,j,k  )
     3                     - ( dx3a(k  ) + xi ) * dd   (k  ,4) )
c
               mflux (k,4) = dtwid (k,4) * ( v3(i+3,j,k) - vg3(k) ) * dt
               dtwid (k,4) = mflux (k,4) * atwid (i+3) * atwid2(j)
#ifndef ISO
               etwid (k,4) = ( 0.5 + q1 ) * ( eod(i+3,j,k-1)
     1                     + ( dx3a(k-1) - xi ) * deod (k-1,4) )
     2                     + ( 0.5 - q1 ) * ( eod(i+3,j,k  )
     3                     - ( dx3a(k  ) + xi ) * deod (k  ,4) )
c
               etwid (k,4) = dtwid (k,4) * etwid (k,4)
#endif /* not ISO */
#if defined RAD || defined RAD_EXP
               rtwid (k,4) = ( 0.5 + q1 ) * ( ero(i+3,j,k-1)
     1                     + ( dx3a(k-1) - xi ) * dero (k-1,4) )
     2                     + ( 0.5 - q1 ) * ( ero(i+3,j,k  )
     3                     - ( dx3a(k  ) + xi ) * dero (k  ,4) )
c
               rtwid (k,4) = dtwid (k,4) * rtwid (k,4)
#endif /* RAD || RAD_EXP */
c
#endif
2040         continue
c
c  Save the mass flux outside (ibeg:iend,jbeg:jend,kbeg:kend)
c  only for zones along the inner borders.
c
             if (j.eq.js-1) then
c
               if (kbeg .eq. ks) then
                 mflx(i  ,j,ks-1) = mflux (ks-1,1)
                 mflx(i+1,j,ks-1) = mflux (ks-1,2)
#if UNROLL_K>2
                 mflx(i+2,j,ks-1) = mflux (ks-1,3)
#endif
#if UNROLL_K>3
                 mflx(i+3,j,ks-1) = mflux (ks-1,4)
#endif
               endif
               do 2060 k=kbeg,kend
                 mflx(i  ,j,k) = mflux (k   ,1)
                 mflx(i+1,j,k) = mflux (k   ,2)
#if UNROLL_K>2
                 mflx(i+2,j,k) = mflux (k   ,3)
#endif
#if UNROLL_K>3
                 mflx(i+3,j,k) = mflux (k   ,4)
#endif
2060           continue
               if (kend .eq. ke) then
                 mflx(i  ,j,ke+1) = mflux (ke+1,1)
                 mflx(i+1,j,ke+1) = mflux (ke+1,2)
#if UNROLL_K>2
                 mflx(i+2,j,ke+1) = mflux (ke+1,3)
#endif
#if UNROLL_K>3
                 mflx(i+3,j,ke+1) = mflux (ke+1,4)
#endif
               endif
c
             endif
c
c  Perform mass density and energy density advection.  Note that
c  the timestep "dt" is hidden the fluxes "dtwid" and "etwid".
c
             if (j.ge.jbeg) then
c
               if (kbeg .eq. ks) then
                 mflx(i  ,j,ks-1) = mflux (ks-1,1)
                 mflx(i+1,j,ks-1) = mflux (ks-1,2)
#if UNROLL_K>2
                 mflx(i+2,j,ks-1) = mflux (ks-1,3)
#endif
#if UNROLL_K>3
                 mflx(i+3,j,ks-1) = mflux (ks-1,4)
#endif
               endif
               do 2080 k=kbeg,kend
               mflx(i  ,j,k)= mflux (k   ,1)
               den(i  ,j,k) = ( dlo(i  ,j,k) * dvl3a(k)
#ifdef MOVING_GRID
     1                      - dtwid (k+1,1) + dtwid (k,1) ) * dvl3ani(k)
#else
     1                      - dtwid (k+1,1) + dtwid (k,1) ) * dvl3a i(k)
#endif /* MOVING GRID */
#ifndef ISO
               e  (i  ,j,k) = ( e  (i  ,j,k) * dvl3a(k)
#ifdef MOVING_GRID
     1                      - etwid (k+1,1) + etwid (k,1) ) * dvl3ani(k)
#else
     1                      - etwid (k+1,1) + etwid (k,1) ) * dvl3a i(k)
#endif /* MOVING GRID */
c
c Compute e/d for the next substep.
c
#ifndef TOTAL_ENERGY
               edn(i  ,j,k) =         e(i  ,j,k) / den(i  ,j,k)
#else /* not TOTAL_ENERGY */
               edn(i  ,j,k) = gamma * e(i  ,j,k) / den(i  ,j,k) - gamm1
     1                      * (
     2                            ( v1(i  ,j,k) + v1(i+1,j  ,k  ) )**2
     3                          + ( v2(i  ,j,k) + v2(i  ,j+1,k  ) )**2
     4                          + ( v3(i  ,j,k) + v3(i  ,j  ,k+1) )**2
     5                        ) * 0.125
#if defined GRAV || defined GRAV_FFT
     5                      - gamm1 * gp(i  ,j,k)
#endif /* GRAV || GRAV_FFT */
#endif /* not TOTAL_ENERGY */
#endif /* not ISO */
#if defined RAD || defined RAD_EXP
               ern(i  ,j,k) = ( ero(i  ,j,k) * dlo(i  ,j,k) * dvl3a(k)
#ifdef MOVING_GRID
     1                      - rtwid (k+1,1) + rtwid (k,1) ) * dvl3ani(k)
#else
     1                      - rtwid (k+1,1) + rtwid (k,1) ) * dvl3a i(k)
#endif /* MOVING GRID */
c
c Work with er/d for the next sweep.
c
               ern(i  ,j,k) = ern(i  ,j,k) / den(i  ,j,k)
#endif /* RAD || RAD_EXP */
c
               mflx(i+1,j,k)= mflux (k   ,2)
               den(i+1,j,k) = ( dlo(i+1,j,k) * dvl3a(k)
#ifdef MOVING_GRID
     1                      - dtwid (k+1,2) + dtwid (k,2) ) * dvl3ani(k)
#else
     1                      - dtwid (k+1,2) + dtwid (k,2) ) * dvl3a i(k)
#endif /* MOVING GRID */
#ifndef ISO
               e  (i+1,j,k) = ( e  (i+1,j,k) * dvl3a(k)
#ifdef MOVING_GRID
     1                      - etwid (k+1,2) + etwid (k,2) ) * dvl3ani(k)
#else
     1                      - etwid (k+1,2) + etwid (k,2) ) * dvl3a i(k)
#endif /* MOVING GRID */
c
c Compute e/d for the next substep.
c
#ifndef TOTAL_ENERGY
               edn(i+1,j,k) =         e(i+1,j,k) / den(i+1,j,k)
#else /* not TOTAL_ENERGY */
               edn(i+1,j,k) = gamma * e(i+1,j,k) / den(i+1,j,k) - gamm1
     1                      * (
     2                            ( v1(i+1,j,k) + v1(i+2,j  ,k  ) )**2
     3                          + ( v2(i+1,j,k) + v2(i+1,j+1,k  ) )**2
     4                          + ( v3(i+1,j,k) + v3(i+1,j  ,k+1) )**2
     5                        ) * 0.125
#if defined GRAV || defined GRAV_FFT
     5                      - gamm1 * gp(i+1,j,k)
#endif /* GRAV || GRAV_FFT */
#endif /* not TOTAL_ENERGY */
#endif /* not ISO */
#if defined RAD || defined RAD_EXP
               ern(i+1,j,k) = ( ero(i+1,j,k) * dlo(i+1,j,k) * dvl3a(k)
#ifdef MOVING_GRID
     1                      - rtwid (k+1,2) + rtwid (k,2) ) * dvl3ani(k)
#else
     1                      - rtwid (k+1,2) + rtwid (k,2) ) * dvl3a i(k)
#endif /* MOVING GRID */
c
c Work with er/d for the next sweep.
c
               ern(i+1,j,k) = ern(i+1,j,k) / den(i+1,j,k)
#endif /* RAD || RAD_EXP */
c
#if UNROLL_K>2
               mflx(i+2,j,k)= mflux (k   ,3)
               den(i+2,j,k) = ( dlo(i+2,j,k) * dvl3a(k)
#ifdef MOVING_GRID
     1                      - dtwid (k+1,3) + dtwid (k,3) ) * dvl3ani(k)
#else
     1                      - dtwid (k+1,3) + dtwid (k,3) ) * dvl3a i(k)
#endif /* MOVING GRID */
#ifndef ISO
               e  (i+2,j,k) = ( e  (i+2,j,k) * dvl3a(k)
#ifdef MOVING_GRID
     1                      - etwid (k+1,3) + etwid (k,3) ) * dvl3ani(k)
#else
     1                      - etwid (k+1,3) + etwid (k,3) ) * dvl3a i(k)
#endif /* MOVING GRID */
c
c Compute e/d for the next substep.
c
#ifndef TOTAL_ENERGY
               edn(i+2,j,k) =         e(i+2,j,k) / den(i+2,j,k)
#else /* not TOTAL_ENERGY */
               edn(i+2,j,k) = gamma * e(i+2,j,k) / den(i+2,j,k) - gamm1
     1                      * (
     2                            ( v1(i+2,j,k) + v1(i+3,j  ,k  ) )**2
     3                          + ( v2(i+2,j,k) + v2(i+2,j+1,k  ) )**2
     4                          + ( v3(i+2,j,k) + v3(i+2,j  ,k+1) )**2
     5                        ) * 0.125
#if defined GRAV || defined GRAV_FFT
     5                      - gamm1 * gp(i+2,j,k)
#endif /* GRAV || GRAV_FFT */
#endif /* not TOTAL_ENERGY */
#endif /* not ISO */
#if defined RAD || defined RAD_EXP
               ern(i+2,j,k) = ( ero(i+2,j,k) * dlo(i+2,j,k) * dvl3a(k)
#ifdef MOVING_GRID
     1                      - rtwid (k+1,3) + rtwid (k,3) ) * dvl3ani(k)
#else
     1                      - rtwid (k+1,3) + rtwid (k,3) ) * dvl3a i(k)
#endif /* MOVING GRID */
c
c Work with er/d for the next sweep.
c
               ern(i+2,j,k) = ern(i+2,j,k) / den(i+2,j,k)
#endif /* RAD || RAD_EXP */
c
#endif
#if UNROLL_K>3
               mflx(i+3,j,k)= mflux (k   ,4)
               den(i+3,j,k) = ( dlo(i+3,j,k) * dvl3a(k)
#ifdef MOVING_GRID
     1                      - dtwid (k+1,4) + dtwid (k,4) ) * dvl3ani(k)
#else
     1                      - dtwid (k+1,4) + dtwid (k,4) ) * dvl3a i(k)
#endif /* MOVING GRID */
#ifndef ISO
               e  (i+3,j,k) = ( e  (i+3,j,k) * dvl3a(k)
#ifdef MOVING_GRID
     1                      - etwid (k+1,4) + etwid (k,4) ) * dvl3ani(k)
#else
     1                      - etwid (k+1,4) + etwid (k,4) ) * dvl3a i(k)
#endif /* MOVING GRID */
c
c Compute e/d for the next substep.
c
#ifndef TOTAL_ENERGY
               edn(i+3,j,k) =         e(i+3,j,k) / den(i+3,j,k)
#else /* not TOTAL_ENERGY */
               edn(i+3,j,k) = gamma * e(i+3,j,k) / den(i+3,j,k) - gamm1
     1                      * (
     2                            ( v1(i+3,j,k) + v1(i+4,j  ,k  ) )**2
     3                          + ( v2(i+3,j,k) + v2(i+3,j+1,k  ) )**2
     4                          + ( v3(i+3,j,k) + v3(i+3,j  ,k+1) )**2
     5                        ) * 0.125
#if defined GRAV || defined GRAV_FFT
     5                      - gamm1 * gp(i+3,j,k)
#endif /* GRAV || GRAV_FFT */
#endif /* not TOTAL_ENERGY */
#endif /* not ISO */
#if defined RAD || defined RAD_EXP
               ern(i+3,j,k) = ( ero(i+3,j,k) * dlo(i+3,j,k) * dvl3a(k)
#ifdef MOVING_GRID
     1                      - rtwid (k+1,4) + rtwid (k,4) ) * dvl3ani(k)
#else
     1                      - rtwid (k+1,4) + rtwid (k,4) ) * dvl3a i(k)
#endif /* MOVING GRID */
c
c Work with er/d for the next sweep.
c
               ern(i+3,j,k) = ern(i+3,j,k) / den(i+3,j,k)
#endif /* RAD || RAD_EXP */
c
#endif
2080           continue
               if (kend .eq. ke) then
                 mflx(i  ,j,ke+1) = mflux (ke+1,1)
                 mflx(i+1,j,ke+1) = mflux (ke+1,2)
#if UNROLL_K>2
                 mflx(i+2,j,ke+1) = mflux (ke+1,3)
#endif
#if UNROLL_K>3
                 mflx(i+3,j,ke+1) = mflux (ke+1,4)
#endif
               endif
             endif  !  j>=jbeg
2090       continue
#endif /* UNROLL_K>1 */
         endif  !  j>=jbeg || j=js-1
2100   continue
c
       return
       end
c
c=======================================================================
c
c    \\\\\\\\\\        E N D   S U B R O U T I N E        //////////
c    //////////                T R A N X 3                \\\\\\\\\\
c
c=======================================================================
c
c
